/*
************************************************************************
**	O.S.		: Linux
**	FILE NAME	: osm_linux.c
**	Author		: C.L. Huang
**	Description	: SCSI Device Driver for Areca Host Bus Adapter
************************************************************************
** Copyright (C) 2002 - 2019, Areca Technology Corporation All rights reserved.
**
** Web site: www.areca.com.tw
** E-mail: support@areca.com.tw
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
************************************************************************
** Redistribution and use in source and binary forms,with or without
** modification,are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice,this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice,this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
** IMPLIED WARRANTIES,INCLUDING,BUT NOT LIMITED TO,THE IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
** IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,INDIRECT,
** INCIDENTAL,SPECIAL,EXEMPLARY,OR CONSEQUENTIAL DAMAGES(INCLUDING,BUT
** NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA,OR PROFITS; OR BUSINESS INTERRUPTION)HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY,WHETHER IN CONTRACT,STRICT LIABILITY,OR TORT
**(INCLUDING NEGLIGENCE OR OTHERWISE)ARISING IN ANY WAY OUT OF THE USE OF
** THIS SOFTWARE,EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
************************************************************************
*/
#include "arc13xx.h"
/*
************************************************************************
**
************************************************************************
*/
PDISCOVERY_ELEMENT arcsas_construct_smp_state_element(PDiscover_Extension pDiscovery,
    PDomain_Port pPort, unsigned char nTgt, unsigned char PhyId, unsigned char CurrentState,
    unsigned char context, void *pBuffer);
unsigned long arcsas_post_smp_issue_phy_control_reset(PDiscover_Extension pDiscovery,
    PDomain_Port pPort, unsigned char nTgt, unsigned char PhyId);
void arcsas_free_enclosure_entry_index(PACB pACB, unsigned char entry_index);
void arcsas_set_lba_and_sector_count(struct CommandControlBlock *pCCB);
void arcsas_free_ncq_tag(PDomain_Device pDevice, struct CommandControlBlock *pCCB);
void arcsas_discovery_set_resource(PDiscover_Extension pDiscovery, u32 Size, unsigned char wantedTargetCount);
void arcsas_os_sas_core_ccb_callback(struct _Core_Extension *pCore, PCCB pNewCCB);
void arcsas_get_core_supported_counts(unsigned char *pEventElementCount, unsigned char *pDevCount,
    unsigned char *pExpanderCount, unsigned char *pPMCount, unsigned short *pCoreCCBCount,
    unsigned int *pSGEntryCount, unsigned short *pPRDBufferCount, unsigned short *pSASScratchCount,
    unsigned short *pSMPScratchCount, unsigned short *pSlotCount, unsigned short *pContextCount,
    unsigned short *pPostQSize, unsigned short *pDoneQSize);
u32 arcsas_core_get_cached_memory_quota(PACB pacb);
struct dv_event_work *arcsas_get_dv_event_work(PACB pACB);
Bool arcsas_init_chip(PACB pACB);
irqreturn_t arcsas_interrupt_check(struct AdapterControlBlock *pACB);
irqreturn_t arcsas_interrupt_msix_1330(struct msix_data *msix_data);
void arcsas_sorting_entry_index(PACB pACB);
void arcsas_free_lun_device(PCore_Extension pCore, PDomain_Device pDevice);
ARCSAS_U64 arcsas_u64_add_u32(ARCSAS_U64 v64, u32 v32);
void arcsas_core_ses_set_led(PDomain_Device pDevice, unsigned char flag);
void arcsas_enable_msi(PACB pACB);
void arcsas_disable_msi(PACB pACB);
void arcsas_enable_msix(PACB pACB);
void arcsas_enable_msix_SM(PACB pACB);
void arcsas_disable_msix(PACB pACB);
void arcsas_disable_msix_SM(PACB pACB);
void arcsas_enable_interrupt(PACB pACB);
void arcsas_disable_interrupt(struct AdapterControlBlock *pACB);
Bool arcsas_device_babies_lun_check(PDomain_Device pDevice);
int arcsas_eh_abort_handler(struct scsi_cmnd *pscsi_cmd);
int arcsas_eh_device_reset_handler(struct scsi_cmnd *pscsi_cmd);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)
void arcsas_pm_ccb_timeout(struct timer_list *t);
void arcsas_timer_wait_pm_phy_ready(struct timer_list *t);
#else
void arcsas_pm_ccb_timeout(unsigned long temp);
void arcsas_timer_wait_pm_phy_ready(unsigned long temp);
#endif
Bool arcsas_update_acb_bios_flag_from_bios(PACB pACB);
Bool arcsas_update_acb_bios_flag_from_bios_1330(PACB pACB);
Bool arcsas_core_initialize(struct AdapterControlBlock *pACB);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
	static int arcsas_queuecommand(struct Scsi_Host *h, struct scsi_cmnd *cmd);
#else
	static int arcsas_queuecommand(struct scsi_cmnd *cmd, void (* done)(struct scsi_cmnd *));
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
	static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth, int reason);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11)
	static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth);
#else
	static ssize_t arcsas_change_queue_depth(struct device *dev, const char *buf, size_t count);
#endif
static int arcsas_probe(struct pci_dev *pcidev, const struct pci_device_id *id);
static void arcsas_remove(struct pci_dev *pcidev);
static void arcsas_shutdown(struct pci_dev *pcidev);
int arcsas_pm_error_handling(PDomain_PM pPM, struct CommandControlBlock *pCCB);
int arcsas_sata_error_handling(struct CommandControlBlock *pCCB);
int arcsas_sas_error_handling(struct CommandControlBlock *pCCB);
void arcsas_free_device_sata_register_set(PCore_Extension pCore, PDomain_Device pDevice);
u32 arcsas_disable_hba_isr(PACB pACB);
void arcsas_enable_main_int(PACB pACB);
void arcsas_disable_main_int(PACB pACB);
void arcsas_enable_hba_isr(PACB pACB, u32 int_cause);
void arcsas_update_phy_info(PCore_Extension pCore, unsigned char phy_index);
static int arcsas_request_irq(struct pci_dev *pdev, struct AdapterControlBlock *pACB);
static void arcsas_free_irq(struct pci_dev *pdev, struct AdapterControlBlock *pACB);
static void arcsas_iounmap(struct AdapterControlBlock *pACB);

extern void arcsas_handle_virtual_ses_device(PACB pACB, struct scsi_cmnd *pscsi_cmnd);
extern void arcsas_dump_reg_1330(PCore_Extension pCore);
#if ARCSAS_DBG_CDB
extern void arcsas_print_cdb(struct CommandControlBlock *pCCB);
#endif
extern void arcsas_core_send_request_sense_ccb(PCore_Extension pCore, PDomain_Device pDevice,
    struct CommandControlBlock *pCCBNew, struct CommandControlBlock * pOrgCCB);
PEVENT_ELEMENT arcsas_construct_event_element(PCore_Extension pCore, u16 EntryIndex, u32 ErrorCode, u8 EventType, u8 *pEventScratchPad);
extern void arcsas_clear_all_irq_status(PACB pACB);
extern void arcsas_sata_send_command(PCORE pCore, PDomain_Port pPort, Domain_Device *pDevice, PCCB pCCB, PCmdStruct pCmdStruct);
extern void arcsas_pm_send_command(PCORE pCore, PDomain_Port pPort, Domain_Device *pDevice, PCCB pCCB, PCmdStruct pCmdStruct);
extern void arcsas_smp_send_command(PCORE pCore, PDomain_Port pPort, Domain_Device *pDevice, PCCB pCCB, PCmdStruct pCmdStruct);
extern void arcsas_ssp_send_command(PCORE pCore, PDomain_Port pPort, Domain_Device *pDevice, PCCB pCCB, PCmdStruct pCmdStruct);
//extern void arcsas_sata_process_command(PCORE pCore, PCCB pCCB, PCmdTable pcmd_table);
//extern void arcsas_stp_process_command(PCORE pCore, PCCB pCCB, PCmdTable pcmd_table);
//extern void arcsas_ssp_process_command(PCORE pCore, PCCB pCCB, PCmdTable pcmd_table);
//extern void arcsas_smp_process_command(PCORE pCore, PCCB pCCB, PCmdTable pcmd_table);
//extern void arcsas_pm_process_command(PCORE pCore, PCCB pCCB, PCmdTable pcmd_table);
//extern u8 arcsas_ssp_error_handler(Domain_Device *pDevice, PCCB pCCB);
//extern u8 arcsas_sata_error_handler(Domain_Device *pDevice, PCCB pCCB);
extern void arcsas_interrupt_service_msix_1330(unsigned long msixData);

/* scsi-mid layer global parmeter is max_report_luns, which is 511 */
static struct scsi_host_template	arcsas_scsi_host_template;
static int max_lun = 32;
module_param(max_lun, int, 0);
MODULE_PARM_DESC(max_lun, " max lun(1 ~ 127), default is 32");

static int aspm = 1;
module_param(aspm, int, 0);
MODULE_PARM_DESC(aspm, " enable aspm(0 ~ 1), aspm=0(disable); aspm=1(enable)");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
int msix_enable = 1;
#else
int msix_enable = 0;
#endif
module_param(msix_enable, int, 0644);
MODULE_PARM_DESC(msix_enable, " MSI-X interrupt(0 ~ 1), msix_enable=1(enable), =0(disable)");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
static int msi_enable = 1;
#else
static int msi_enable = 0;
#endif
module_param(msi_enable, int, 0644);
MODULE_PARM_DESC(msi_enable, " Enable MSI interrupt(0 ~ 1), msi_enable=1(enable), =0(disable)");

static int host_can_queue = ARCSAS_OUTSTANDING_CMD;
module_param(host_can_queue, int, 0644);
MODULE_PARM_DESC(host_can_queue, " controller queue depth(1 ~ 256), default is 256");

static int cmd_per_lun = ARCSAS_DEFAULT_CMD_PERDEV;
module_param(cmd_per_lun, int, 0644);
MODULE_PARM_DESC(cmd_per_lun, " device queue depth(1 ~ 64), default is 31");

MODULE_PARM_DESC(Set_parameter_when_installing, " e.g. insmod arcsas.ko cmd_per_lun=8 host_can_queue=64");
MODULE_PARM_DESC(Set_parameter_when_kernel_booting, " e.g. arcsas.host_can_queue=128 arcsas.msix_enable=0");

static u8 log_2_asic[] = {1,0,2,3,5,4,6,7,9,8,10,11,13,12,14,15};
/*
************************************************************************
************************************************************************
*/
MODULE_AUTHOR("Ching Huang <support@areca.com.tw>");
MODULE_DESCRIPTION("Areca SAS 3G/6G/12G HBA driver");
MODULE_VERSION(ARCSAS_DRIVER_VERSION);
MODULE_LICENSE("GPL v2");
#if 1
/*
************************************************************************
* reset nmi watchdog
************************************************************************
*/
void arcsas_touch_watchdog(void)
{
	#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 32)
	//	touch_nmi_watchdog();
	#elif LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)

	#elif  LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 13)
	//	touch_nmi_watchdog();
	#else
	//	touch_all_softlockup_watchdogs();
	#endif
}
#endif
/*
************************************************************************
************************************************************************
*/
#define _WARN(condition, format...) ({		\
	int __ret_warn_on = !!(condition);	\
	unlikely(__ret_warn_on);		\
})
#define _WARN_ONCE(condition, format...)({	\
	static int __warned;			\
	int __ret_warn_once = !!(condition);	\
						\
	if (unlikely(__ret_warn_once))		\
		if (WARN(!__warned, format)) 	\
			__warned = 1;		\
	unlikely(__ret_warn_once);		\
})
/*
************************************************************************
************************************************************************
*/
u32 arcsas_show_time(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0)
	struct tm tm;
#else
	unsigned int days, j, i, a, b, c, d, e, m;
	struct timeval tv;
#endif
	unsigned int year, mon, day, hour, min, sec, secs;


#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0)
	time64_to_tm(ktime_get_real_seconds(), -sys_tz.tz_minuteswest * 60, &tm);
	year = tm.tm_year + 1900; /* base on 1900 */
	mon = tm.tm_mon + 1;
	day = tm.tm_mday;
	hour = tm.tm_hour;
	min = tm.tm_min;
	sec = tm.tm_sec;
	secs = sec + 60 * min;
#else
	do_gettimeofday(&tv);
	secs = (u32)(tv.tv_sec - (sys_tz.tz_minuteswest * 60));
	days = secs / 86400;
	secs = secs - 86400 * days;
	if (secs < 0) {
		days = days - 1;
		secs = secs + 86400;
	}
	j = days / 146097;
	i = days - 146097 * j;
	a = i + 719468;
	b = ( 4 * a + 3 ) / 146097;
	c = a - ( 146097 * b ) / 4;
	d = ( 4 * c + 3 ) / 1461 ;
	e = c - ( 1461 * d ) / 4 ;
	m = ( 5 * e + 2 ) / 153 ;
	year = 400 * j + 100 * b + d + m / 10;
	mon = m + 3 - 12 * ( m /10 );
	day = e - ( 153 * m + 2 ) / 5 + 1;
	hour = secs / 3600;
	secs = secs - 3600 * hour;
	min = secs / 60;
	sec = secs - 60 * min;
#endif
	printk("%02d:%02d:%02d %d/%d/%d \n", hour, min, sec, year, mon, day);
	return (ARCSAS_BIT(secs & 63, 0) | ARCSAS_BIT(min & 63, 6) | ARCSAS_BIT(mon & 15, 12) |
	    ARCSAS_BIT(hour & 63, 16) | ARCSAS_BIT(day & 31, 21) | ARCSAS_BIT((year - 2000) & 31, 26));
}
/*
************************************************************************
************************************************************************
*/
/* register read write: memory io */
void osm_writel(u32 val, void *addr)	{writel(val, addr);}
u32 osm_readl(void *addr)		{return readl(addr);}
/* register read write: port io */
void osm_outl(u32 val, unsigned long addr)	{outl(val, addr);}
u32 osm_inl(unsigned long addr)		{return inl(addr);}
/*
************************************************************************
************************************************************************
*/
void osm_memset(void *buf, int  value, size_t len)	{memset(buf, value, len);}
void osm_memcpy(void *dest, void *source, size_t len)	{memcpy(dest, source, len);}
int osm_memcmp(void *buf0, void *buf1, size_t len)	{return memcmp(buf0, buf1, len);}
int osm_strncmp(const char *cs, const char *ct, size_t count)	{return strncmp(cs, ct, count);}
/*
************************************************************************
************************************************************************
*/
int osm_schedule_work(struct work_struct *work)	{ return schedule_work(work);}			
/*
************************************************************************
************************************************************************
*/
int osm_atomic_inc_return(atomic_t *v)		{return atomic_inc_return(v);}
int osm_atomic_read(const atomic_t *v)		{return atomic_read(v);}
void osm_atomic_inc(atomic_t *v)		{atomic_inc(v);}
void osm_atomic_dec(atomic_t *v)		{atomic_dec(v);}
void osm_atomic_set(atomic_t *addr, int v)	{atomic_set(addr, v);}
/*
************************************************************************
************************************************************************
*/
int osm_pci_read_config_dword(struct pci_dev *dev, int offset, u32 *ptr)	{return pci_read_config_dword(dev, offset, ptr);}
int osm_pci_read_config_word(struct pci_dev *dev, int offset, u16 *ptr)		{return pci_read_config_word(dev, offset, ptr);}
int osm_pci_read_config_byte(struct pci_dev *dev, int offset, u8 *ptr)		{return pci_read_config_byte(dev, offset, ptr);}
int osm_pci_write_config_dword(struct pci_dev *dev, int offset, u32 val)	{return pci_write_config_dword(dev, offset, val);}	
int osm_pci_write_config_word(struct pci_dev *dev, int offset, u16 val)		{return pci_write_config_word(dev, offset, val);}
int osm_pci_write_config_byte(struct pci_dev *dev, int offset, u8 val)		{return pci_write_config_byte(dev, offset, val);}
/*
************************************************************************
************************************************************************
*/
void arcsas_spin_lock_init(spinlock_t *lock)
{
	spin_lock_init(lock);
}
void arcsas_spin_lock_irqsave(spinlock_t *lock, unsigned long *flags)
{
	spin_lock_irqsave(lock, *flags);
}
void arcsas_spin_unlock_irqrestore(spinlock_t *lock, unsigned long *flags)
{
	spin_unlock_irqrestore(lock, *flags);
}
void arcsas_spin_lock(spinlock_t *lock)
{
	spin_lock(lock);
}
void arcsas_spin_unlock(spinlock_t *lock)
{
	spin_unlock(lock);
}
void swap_buf_le16(u16 *buf, unsigned int buf_words)
{
#ifdef _BIG_ENDIAN__
        unsigned int i;

        for (i = 0; i < buf_words; i++)
                buf[i] = ARCSAS_LE16_TO_CPU(buf[i]);
#endif
}
/*
************************************************************************
************************************************************************
*/
struct scsi_device *osm_scsi_device_lookup(struct Scsi_Host *shost, uint channel, uint id, uint lun)
{
	struct scsi_device *sdev;
	sdev = scsi_device_lookup(shost, channel, id, lun);
	if (sdev)
		scsi_device_put(sdev);
	return sdev;
}
/*
************************************************************************
************************************************************************
*/
void __osm_list_del(struct osm_list_head * prev, struct osm_list_head * next)
{
	next->prev = prev;
	prev->next = next;
}
/*
************************************************************************
************************************************************************
*/
void __osm_list_add(struct osm_list_head *new, struct osm_list_head *prev, struct osm_list_head *next)
{
#if 1
	_WARN(next->prev != prev,
		"osm_list_add corruption. next->prev should be "
		"prev (%p), but was %p. (next=%p).\n",
		prev, next->prev, next);
	_WARN(prev->next != next,
		"osm_list_add corruption. prev->next should be "
		"next (%p), but was %p. (prev=%p).\n",
		next, prev->next, prev);
#endif
	next->prev = new;
	new->next = next;
	new->prev = prev;
	prev->next = new;
}
/*
************************************************************************
************************************************************************
*/
void osm_list_del(struct osm_list_head *entry)
{
#if 1
	_WARN(entry->next == LIST_POISON1,
		"osm_list_del corruption, next is LIST_POISON1 (%p)\n",
		LIST_POISON1);
	_WARN(entry->next != LIST_POISON1 && entry->prev == LIST_POISON2,
		"osm_list_del corruption, prev is LIST_POISON2 (%p)\n",
		LIST_POISON2);
	_WARN(entry->prev->next != entry,
		"osm_list_del corruption. prev->next should be %p, "
		"but was %p\n", entry, entry->prev->next);
	_WARN(entry->next->prev != entry,
		"osm_list_del corruption. next->prev should be %p, "
		"but was %p\n", entry, entry->next->prev);
#endif
	__osm_list_del(entry->prev, entry->next);
	entry->next = LIST_POISON1;
	entry->prev = LIST_POISON2;
}
/*
************************************************************************
************************************************************************
*/
void osm_list_add_tail(struct osm_list_head *new, struct osm_list_head *head)
{
	__osm_list_add(new, head->prev, head);
}
/*
************************************************************************
************************************************************************
*/
void OSM_INIT_LIST_HEAD(struct osm_list_head *list)
{
	list->next = list;
	list->prev = list;
}
/*
************************************************************************
************************************************************************
*/
void osm_list_del_init(struct osm_list_head *entry)
{
	__osm_list_del(entry->prev, entry->next);
	OSM_INIT_LIST_HEAD(entry);
}
/*
************************************************************************
************************************************************************
*/
void osm_list_add(struct osm_list_head *new, struct osm_list_head *head)
{
	__osm_list_add(new, head, head->next);
}
/*
************************************************************************
************************************************************************
*/
void osm_list_move(struct osm_list_head *list, struct osm_list_head *head)
{
	__osm_list_del(list->prev, list->next);
	osm_list_add(list, head);
}
/*
************************************************************************
************************************************************************
*/
void osm_list_move_tail(struct osm_list_head *list, struct osm_list_head *head)
{
	__osm_list_del(list->prev, list->next);
	osm_list_add_tail(list, head);
}
/*
************************************************************************
************************************************************************
*/
Bool osm_list_empty(const struct osm_list_head *head)
{
	return head->next == head;
}
/*
************************************************************************
************************************************************************
*/
unsigned long get_jiffies(void)
{
	return jiffies;
}
void *osm_kmalloc(size_t size, gfp_t flags) {	return kmalloc(size, flags);}
void osm_kfree(const void *x) {	return kfree(x);}
/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 17, 0)
#ifndef scsilun_to_int
int scsilun_to_int(struct scsi_lun *scsilun)
{
	int i;
	unsigned int lun;

	lun = 0;
	for (i = 0; i < sizeof(lun); i += 2)
		lun = lun | (((scsilun->scsi_lun[i] << 8) |
			      scsilun->scsi_lun[i + 1]) << (i * 8));
	return lun;
}
#endif
#endif
/*
************************************************************************
************************************************************************
*/
#define SAS_MAX_QD 64

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 16)
static inline void setup_timer(struct timer_list * timer,
				void (*function)(unsigned long),
				unsigned long data)
{
	init_timer(timer);
	timer->function = function;
	timer->data = data;
}
#endif

/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)
void arcsas_setup_timer(struct timer_list *timer, Timer_CBFunc function, unsigned long data)
{
	timer_setup(timer, function, 0);
}
#else
void arcsas_setup_timer(struct timer_list *timer, Timer_CBFunc function, unsigned long data)
{
	setup_timer(timer, function, data);
}
#endif
/*
************************************************************************
************************************************************************
*/
unsigned char arcsas_mod_timer(struct timer_list *timer, unsigned long timeout)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	//mod_timer(timer, jiffies + timeout * HZ);
	return RUNNING;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_hba_smart_cancel_timer(PACB pACB)
{
	del_timer_sync(&pACB->SMART_timer_request);
	return;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_delete_device_timer(PDomain_Device pDevice)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	//del_timer_sync(&pDevice->timer);
	return;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_delete_pm_timer(PDomain_PM pPM)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	del_timer(&pPM->timer);
	return;
}

void arcsas_delete_pm_timers(PDomain_PM pPM)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	del_timer_sync(&pPM->timer);
	del_timer_sync(&pPM->timer1);
	return;
}
/*
************************************************************************
************************************************************************
*/
unsigned char rbuffer[1024];
void arcsas_map_sg_to_buffer(struct CommandControlBlock *pCCB)
{
	struct scsi_cmnd *pscsi_cmnd;
	struct scatterlist *sg;
	unsigned short	sgCnt;

	#if ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB
		osm_printk("%s: pCCB = 0x%p\n", __func__, pCCB);
	#endif
	pscsi_cmnd = (struct scsi_cmnd *)pCCB->Org_Req;
//	pscsi_cmnd = (struct scsi_cmnd *)pCCB->pscsi_cmnd;
	#if ARCSAS_DBG_OS_CCB_CONTEXT
		osm_printk("%s: pscsi_cmnd = 0x%p\n", __func__, pscsi_cmnd);
	#endif
	sgCnt = arcsas_use_sg(pscsi_cmnd);
	if (sgCnt) {
		sg = (struct scatterlist *)arcsas_rq_bf(pscsi_cmnd);
	// !!!do not kmalloc here when kernel version >= 5.19.0
	//	pCCB->Data_Buffer = kmalloc(arcsas_rq_bflen(pscsi_cmnd), GFP_ATOMIC);
		pCCB->Data_Buffer = rbuffer;
		if (pCCB->Data_Buffer) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
			sg_copy_to_buffer(scsi_sglist(pscsi_cmnd), scsi_sg_count(pscsi_cmnd),
				    pCCB->Data_Buffer, scsi_bufflen(pscsi_cmnd));
#else
			void *virt_address;
			unsigned short i;
			pCCB->Data_Transfer_Length = 0;
			for (i = 0; i < sgCnt; i++, sg_next(sg)) {
				virt_address = arcsas_map_sg_page(sg);
				#if ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_ATA_PASSTHROU_CONTEXT
					osm_printk("%s: virt_address = 0x%p, sg->offset = 0x%x\n", __func__, virt_address, sg->offset);
				#endif
				if (pCCB->Cmd_Flag & CMD_FLAG_DMA)
					memcpy(pCCB->Data_Buffer + pCCB->Data_Transfer_Length , virt_address + sg->offset, sg->length);
				arcsas_unmap_buffer(virt_address);
				pCCB->Data_Transfer_Length += sg->length;
			}
#endif
		} else {
		//	_ASSERT();
		}
		#if ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT
			osm_printk("%s: Data_Transfer_Length = 0x%x.\n", __func__, pCCB->Data_Transfer_Length);
		#endif
	} else {
		#if ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT
			osm_printk("%s: no use sg\n", __func__);
		#endif
		pCCB->Data_Buffer = arcsas_rq_bf(pscsi_cmnd);
	}
}
/*
************************************************************************
************************************************************************
*/
void arcsas_unmap_sg_to_buffer(struct CommandControlBlock *pCCB)
{
	struct scsi_cmnd *pscsi_cmnd = NULL;
	struct scatterlist *sg = NULL;
	unsigned short sgCnt;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pscsi_cmnd = (struct scsi_cmnd *)pCCB->Org_Req;
//	pscsi_cmnd = (struct scsi_cmnd *)pCCB->pscsi_cmnd;
	sg = (struct scatterlist *)arcsas_rq_bf(pscsi_cmnd);
	sgCnt = arcsas_use_sg(pscsi_cmnd);
	if (sgCnt) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
		sg_copy_from_buffer(scsi_sglist(pscsi_cmnd), scsi_sg_count(pscsi_cmnd),
				    pCCB->Data_Buffer, scsi_bufflen(pscsi_cmnd));
#else
		void *virt_address;
		unsigned short i;
		pCCB->Data_Transfer_Length = 0;
		for (i = 0; i < sgCnt; i++, sg_next(sg)) {
			virt_address = arcsas_map_sg_page(sg);
			#if ARCSAS_DBG_CONTEXT || ARCSAS_DBG_ATA_PASSTHROU_CONTEXT
				osm_printk("%s: virt_address = 0x%p, sg->offset = 0x%x\n", __func__, virt_address, sg->offset);
			#endif
			memcpy(virt_address + sg->offset, pCCB->Data_Buffer + pCCB->Data_Transfer_Length, sg->length);
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s: kunmap_atomic\n", __func__);
			#endif
			arcsas_unmap_buffer(virt_address);
			pCCB->Data_Transfer_Length += sg->length;
		}
#endif
	//	kfree(pCCB->Data_Buffer);
		pCCB->Data_Buffer = arcsas_rq_bf(pscsi_cmnd);
	}
}
/*
************************************************************************
************************************************************************
*/
/* Release the buffer and set the pointer to NULL. */
void arcsas_core_free_sas_scratch_buffer_to_list(PCore_Extension pCore, PSAS_Scratch_Buffer pSASSB)
{
	unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	if (pSASSB) {
		arcsas_spin_lock_irqsave(&pCore->sassb_list_lock, &flags);
		osm_list_add_tail(&pSASSB->Queue_Pointer, &pCore->SAS_Scratch_List);
		#if ARCSAS_DBG_SES_CONTEXT
			osm_printk("%s: pSASSB = 0x%p, pSASSB->Scratch_Vir = 0x%p\n", __func__, pSASSB, pSASSB->Scratch_Vir);
		#endif
		arcsas_spin_unlock_irqrestore(&pCore->sassb_list_lock, &flags);
	}
}
/*
************************************************************************
************************************************************************
*/
PSAS_Scratch_Buffer arcsas_core_get_sas_scratch_buffer_from_list(PCore_Extension pCore)
{
	PSAS_Scratch_Buffer pSASSB;
	unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock_irqsave(&pCore->sassb_list_lock, &flags);
	if (!osm_list_empty(&pCore->SAS_Scratch_List)) {
		pSASSB = (PSAS_Scratch_Buffer)osm_list_first_entry(&pCore->SAS_Scratch_List, struct _SAS_Scratch_Buffer, Queue_Pointer);
		osm_list_del_init(&pSASSB->Queue_Pointer);
		arcsas_spin_unlock_irqrestore(&pCore->sassb_list_lock, &flags);
		#if ARCSAS_DBG_SES_CONTEXT
			osm_printk("%s: pSASSB = 0x%p, pSASSB->Scratch_Vir = 0x%p\n", __func__, pSASSB, pSASSB->Scratch_Vir);
		#endif
		return pSASSB;
	}
	arcsas_spin_unlock_irqrestore(&pCore->sassb_list_lock, &flags);
	return NULL;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_fill_sense_buffer(struct scsi_cmnd *pscsi_cmnd, unsigned char senseKey, unsigned char adSenseCode)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	((unsigned char *)pscsi_cmnd->sense_buffer)[0] = 0x70;		/* Current */
	((unsigned char *)pscsi_cmnd->sense_buffer)[2] = senseKey;
	((unsigned char *)pscsi_cmnd->sense_buffer)[7] = 0;		/* additional sense length */
	((unsigned char *)pscsi_cmnd->sense_buffer)[12] = adSenseCode;/* additional sense code */
}
/*
************************************************************************
************************************************************************
*/
void arcsas_scsi_device_lookup(struct AdapterControlBlock *pACB)
{
	int i;
	PDomain_Device pDevice;
	struct dv_event_work *dv_event;
//	unsigned long flags;
	#if ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PLUG_UNPLUG
		osm_printk("%s:\n", __func__);
	#endif
	for (i = 0; i < ARCSAS_MAX_DISK_ID; i++) {/*(0..135,136..143,144..151)*/
		if (pACB->DevicesExpandersPMs_EntryIndex[i] == ID_NOT_MAPPED) {
			continue;
		}
		pDevice = &pACB->Devices[Get_DevicesExpandersPMs_EntryIndex(pACB, i)];/*(0..135)*/
		#if ARCSAS_DBG_PLUG_UNPLUG
			osm_printk("%s: pDevice = 0x%p, pDevice->Status = 0x%x, pDevice->arcsas_add_scsi_device = 0x%p\n",
			    __func__, pDevice, pDevice->Status, &pDevice->arcsas_add_scsi_device);
			osm_printk("%s: pPort = 0x%p, &pPort->Device_List= 0x%p\n", __func__, pDevice->pPort,
			    &pDevice->pPort->Device_List);
			osm_printk("%s: pDevice->Port_Queue_Pointer->next = 0x%p, pDevice->Port_Queue_Pointer->prev = 0x%p\n",
			    __func__, pDevice->Port_Queue_Pointer.next, pDevice->Port_Queue_Pointer.prev);
		#endif
		if (!(pDevice->Status & DEVICE_STATE_REPORT2_MLDRIVER) && (pDevice->State == DEVICE_STATE_INIT_DONE)) {
			pDevice->psdev = osm_scsi_device_lookup(pACB->host, 0, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex),
			    DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
			#if ARCSAS_DBG_PLUG_UNPLUG
				osm_printk("%s: pDevice->psdev = 0x%p\n", __func__, pDevice->psdev);
			#endif
			if (!pDevice->psdev) {
				#if ARCSAS_DBG_PLUG_UNPLUG
					osm_printk("%s: No.1\n", __func__);
				#endif
	//	#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 19, 0)
	//			dv_event = arcsas_get_dv_event_work(pACB);
	//			if (!dv_event)
	//				pDevice->Status |= DEVICE_STATE_REPORT2_MLDRIVER;
	//			dv_event->pACB = pACB;
	//			dv_event->pDevice = pDevice;
	//			pDevice->dv_event = dv_event;
	//	#endif
				dv_event = pDevice->dv_event;
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
				INIT_WORK(&dv_event->work, arcsas_add_scsi_device_wqfn);
		#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
				INIT_WORK(&dv_event->work, arcsas_add_scsi_device_wqfn);
		#else
				INIT_WORK(&dv_event->work, arcsas_add_scsi_device_wqfn, pDevice);
		#endif
				queue_work(pACB->scsi_lookup, &dv_event->work);
				#if ARCSAS_DBG_PLUG_UNPLUG
					osm_printk("%s: line %d queue_work returns\n", __func__, __LINE__);
				#endif
			}
		}
	}
}
/*
************************************************************************
************************************************************************
*/
static void
arcsas_free_dv_event_work(struct AdapterControlBlock *pACB, struct dv_event_work *dv_event)

{
	arcsas_spin_lock(&pACB->dv_event_lock);
//	list_del(&dv_event->list);
//	kfree(dv_event);
//	INIT_LIST_HEAD(&dv_event->list);
	osm_list_add_tail(&dv_event->list, &pACB->dv_event_list);
	arcsas_spin_unlock(&pACB->dv_event_lock);
}

struct dv_event_work *arcsas_get_dv_event_work(PACB pACB)
{
	struct dv_event_work *dv_event;

	arcsas_spin_lock(&pACB->dv_event_lock);
	if (osm_list_empty(&pACB->dv_event_list)) {
		arcsas_spin_unlock(&pACB->dv_event_lock);
		return NULL;
	}
	dv_event = ((struct dv_event_work *)osm_list_first_entry(&pACB->dv_event_list, struct dv_event_work, list));
	osm_list_del_init(&dv_event->list);
	arcsas_spin_unlock(&pACB->dv_event_lock);
	return dv_event;
}

/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
void arcsas_add_scsi_device_wqfn(struct work_struct *work)
{
	int rc = -1;
	struct _Domain_Device *pDevice;
	struct AdapterControlBlock *pACB;
	struct dv_event_work *dv_event = container_of(work, struct dv_event_work, work);
	//struct scsi_device *psdev;

	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PLUG_UNPLUG)
		osm_printk("%s:\n", __func__);
	#endif
//	pDevice = container_of(work, struct _Domain_Device, arcsas_add_scsi_device);
	pDevice = dv_event->pDevice;
	pACB = pDevice->pCore->pACB;
	rc = scsi_add_device(pACB->host, 0, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex),
	    DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
	pDevice->Status &= ~DEVICE_STATE_REPORT2_MLDRIVER;
	if (rc) {
		osm_printk("Failed to add SCSI device(id:lun=%d:%d), rc = %d\n",
		    DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex), rc);
	} else {
		u32 ErrorCode = ARCSAS_EVENT_DEVICE_PLUG;
		u8 EventType = ARCSAS_EVENT_TYPE_MESSAGE_ONLY;
		u8 EventScratchPad[16] = {0};
		PCore_Extension pCore = pDevice->pCore;
		//pDevice->psdev = osm_scsi_device_lookup(pACB->host, 0, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex),
		//    DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
		osm_printk("SCSI device(id:lun=%d:%d), added @phy %d\n", DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex),
		    DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex), pDevice->Phy_Id);
		if (pDevice->Dev_Type == DT_ENCLOSURE) {
			u16 ExpEntryIndex;
			EventType = ARCSAS_EVENT_TYPE_ENCLOSURE;
			ExpEntryIndex = pACB->DevicesExpandersPMs_EntryIndex[pDevice->pExpander->EntryIndex];
			snprintf(EventScratchPad, sizeof(EventScratchPad), "Exp#%d Enc#%d", (u32)ExpEntryIndex, (u32)pDevice->EntryIndex);
		} else if ((pDevice->pPM != NULL) && (pDevice->Connection & DC_ATA)) {
			u8 CorePhyNum;
			EventType = ARCSAS_EVENT_TYPE_PM_SLOT;
			CorePhyNum = (pDevice->pPM->CorePhyNum + ((pCore->core_flag & ARCSAS_CORE_FLAG_HBA_CORE_ID1) ? 4 : 0));
			snprintf(EventScratchPad, sizeof(EventScratchPad), "PM#%d Slot#%d", (u32)CorePhyNum, (u32)pDevice->pm_port_no);
		} else if ((pDevice->pExpander != NULL) && (pDevice->Connection & DC_SCSI)) {
			u16 ExpEntryIndex;
			EventType = ARCSAS_EVENT_TYPE_EXPANDER_SLOT;
			ExpEntryIndex = pACB->DevicesExpandersPMs_EntryIndex[pDevice->pExpander->EntryIndex];
			snprintf(EventScratchPad, sizeof(EventScratchPad), "Exp#%d Slot#%d", (u32)ExpEntryIndex,
			    (u32)pDevice->SesSlotNumber);
		}
		if (pDevice->Connection & DC_SGPIO) {
			EventType = ARCSAS_EVENT_TYPE_HBA_SLOT;
			snprintf(EventScratchPad, sizeof(EventScratchPad), "HBA Slot#%d", (u32)pDevice->EntryIndex);
		}
		arcsas_construct_event_element(pCore, pDevice->EntryIndex, ErrorCode, EventType, EventScratchPad);
		//OSM_INIT_WORK((struct work_struct *)&(pDevice->psdev->hostdata), arcsas_remove_scsi_device_wqfn);
	}
//	arcsas_free_dv_event_work(pACB, dv_event);
}
/*
************************************************************************
************************************************************************
*/
void arcsas_remove_scsi_device_wqfn(struct work_struct *work)
{
	u8 EventScratchPad[16] = {0};
	u8 EventType = ARCSAS_EVENT_TYPE_MESSAGE_ONLY;
	u32 ErrorCode = ARCSAS_EVENT_DEVICE_UNPLUG;
	struct _Domain_Device *pDevice;
	PACB pACB;
	PCore_Extension pCore;
	struct dv_event_work *dv_event = container_of(work, struct dv_event_work, work);
	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PLUG_UNPLUG)
		osm_printk("%s:\n", __func__);
	#endif
//	pDevice = container_of(work, struct _Domain_Device, arcsas_remove_scsi_device);
	pDevice = dv_event->pDevice;
	pCore = pDevice->pCore;
	pACB = pCore->pACB;
	#if ARCSAS_DBG_PLUG_UNPLUG
		osm_printk("SCSI device(0x%p, id:lun=%d:%d) going to be removed\n", pDevice->psdev,
		    DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
	#endif
	if (pDevice->psdev != NULL) {
		if (IS_SGPIO(pDevice)) {
	#if ARCSAS_SUPPORT_SGPIO
			arcsas_core_sgpio_set_led(pDevice, LED_FLAG_ALL_OFF);
	#endif
		} else if (IS_SSP(pDevice) || IS_STP(pDevice)) {
	#if ARCSAS_SUPPORT_SES
			arcsas_core_ses_set_led(pDevice, LED_FLAG_ALL_OFF);
	#endif
		}
		#if ARCSAS_DBG_PLUG_UNPLUG
			osm_printk("%s: No.2\n", __func__);
		#endif
		if ((pDevice->pPM != NULL) && (pDevice->Connection & DC_ATA)) {
			u8 CorePhyNum;
			EventType = ARCSAS_EVENT_TYPE_PM_SLOT;
			#if ARCSAS_DBG_PLUG_UNPLUG
				osm_printk("%s: No.3\n", __func__);
			#endif
			CorePhyNum = (pDevice->pPM->CorePhyNum + ((pCore->core_flag & ARCSAS_CORE_FLAG_HBA_CORE_ID1) ? 4 : 0));
			snprintf(EventScratchPad, sizeof(EventScratchPad),"Pm#%d Slot#%d", (u32)CorePhyNum, (u32)pDevice->pm_port_no);
		} else if ((pDevice->pExpander != NULL) && (pDevice->Connection & DC_SCSI)) {
			u16 ExpEntryIndex;
			EventType = ARCSAS_EVENT_TYPE_EXPANDER_SLOT;
			#if ARCSAS_DBG_PLUG_UNPLUG
				osm_printk("%s: No.4\n", __func__);
			#endif
			ExpEntryIndex = pACB->DevicesExpandersPMs_EntryIndex[pDevice->pExpander->EntryIndex];
			snprintf(EventScratchPad, sizeof(EventScratchPad), "Exp#%d Slot#%d", (u32)ExpEntryIndex,
			    (u32)pDevice->SesSlotNumber);
		} else if (pDevice->Connection & DC_SGPIO) {
			EventType = ARCSAS_EVENT_TYPE_HBA_SLOT;
			#if ARCSAS_DBG_PLUG_UNPLUG
				osm_printk("%s: No.5\n", __func__);
			#endif
			snprintf(EventScratchPad, sizeof(EventScratchPad), "HBA Slot#%d",
			    (u32)DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex));
		}
		#if ARCSAS_DBG_PLUG_UNPLUG
			osm_printk("%s: No.6\n", __func__);
		#endif
		arcsas_construct_event_element(pCore, pDevice->EntryIndex, ErrorCode, EventType, EventScratchPad);
		#if ARCSAS_DBG_PLUG_UNPLUG
			osm_printk("%s: No.7\n", __func__);
		#endif
		scsi_remove_device(pDevice->psdev);
		pDevice->psdev = NULL;
		osm_printk("SCSI device scsi:id:lun=%x:%x:%x was removed\n", pACB->host->host_no,
		    DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
	}
	#if ARCSAS_DBG_PLUG_UNPLUG
		osm_printk("%s: No.8\n", __func__);
	#endif
	arcsas_free_dv_event_work(pACB, dv_event);
}	
/*
************************************************************************
************************************************************************
*/
void arcsas_do_smp_state_machine_wqfn(struct work_struct *work)
{
	struct _Discover_Extension *pDiscovery;
	PACB pACB;
	PDISCOVERY_ELEMENT pDiscovery_element;
	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_DISCOVERY)
		osm_printk("%s:\n", __func__);
	#endif
	pDiscovery = container_of(work, struct _Discover_Extension, arcsas_do_smp_state_machine);
	pACB = pDiscovery->Target->pPort->pCore->pACB;
	pDiscovery_element = pACB->pDiscovery_element;
	arcsas_discovery_smp_state_machine(pDiscovery_element->pDiscovery, pDiscovery_element->pPort,
	    pDiscovery_element->nTgt, pDiscovery_element->CurrentState, pDiscovery_element->context, pDiscovery_element->pBuffer);
	pACB->discoveryQ_count--;
	#if ARCSAS_DBG_DISCOVERY_CONTEXT
		osm_printk("%s: No.1\n", __func__);
	#endif
	return;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_do_smp_phy_control_reset_wqfn(struct work_struct *work)
{
	struct _Discover_Extension *pDiscovery;
	PACB pACB;
	PDISCOVERY_ELEMENT pDiscovery_element;
	PDomain_Port pPort;
	unsigned char nTgt;
	unsigned char PhyId;
	unsigned char CurrentState;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
	struct dv_event_work	*dv_event;
#endif
	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_DISCOVERY)
		osm_printk("%s:\n", __func__);
	#endif
	pDiscovery = container_of(work, struct _Discover_Extension, arcsas_do_smp_phy_control_reset);
	pACB = pDiscovery->Target->pPort->pCore->pACB;
	pDiscovery_element = pACB->pDiscovery_element;
	pPort = pDiscovery_element->pPort;
	nTgt = pDiscovery_element->nTgt;
	PhyId = pDiscovery_element->PhyId;
	CurrentState = pDiscovery_element->CurrentState;
	pDiscovery->numOfDiscoveryReq++;
	if (arcsas_post_smp_issue_phy_control_reset(pDiscovery, pPort, nTgt, PhyId) == STATUS_REQUEST_OK) {
	        /*issue PHY_CONTROL SMP function ,let sata oob done*/
		osm_printk("issue phy control .. reset ....................\n");
		pACB->discoveryQ_count--;
	} else {
		osm_printk("arcsas_post_smp_issue_phy_control_reset, we got a queue full....\n");
		osm_printk("issue phy control .. reset ....not ok................\n");
		pACB->discoveryQ_count--;
		pDiscovery->numOfDiscoveryReq--;
		if ((list_empty(&(&pDiscovery->arcsas_do_smp_phy_control_reset)->entry))) {
			arcsas_construct_smp_state_element(pDiscovery, pPort, nTgt, PhyId, CurrentState, 0, NULL);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
			dv_event = pDiscovery->dv_event;
			queue_work(pACB->scsi_lookup, &dv_event->work);
#else
			schedule_work(&pDiscovery->arcsas_do_smp_phy_control_reset);
#endif
		}
	}
	osm_printk("%s: line %d numOfDiscoveryReq = %d\n", __func__, __LINE__, pDiscovery->numOfDiscoveryReq);
	return;
}
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11)
#undef	scsi_add_device
#define MAX_ERRNO       4095

int scsi_add_device(struct Scsi_Host *host, uint channel,
            uint target, uint lun)
{
	struct scsi_device *sdev = 
		__scsi_add_device(host, channel, target, lun, NULL);
	if (unlikely((unsigned long)sdev) >= (unsigned long)-MAX_ERRNO)
        	return ((long) sdev);
	scsi_device_put(sdev);
	return 0;
}
#endif
/*
************************************************************************
************************************************************************
*/
void arcsas_add_scsi_device_wqfn(void *ptr)
{
	int rc = -1;
	struct _Domain_Device *pDevice;
	struct AdapterControlBlock *pACB;
	//struct _Domain_Port *pPort;
	//struct scsi_device *psdev;

	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PLUG_UNPLUG)
		osm_printk("%s:\n", __func__);
	#endif
	pDevice = (struct _Domain_Device *)ptr;
	pACB = pDevice->pCore->pACB;
	rc = scsi_add_device(pACB->host, 0, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex),
	    DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
	pDevice->Status &= ~DEVICE_STATE_REPORT2_MLDRIVER;
	if (rc) {
		osm_printk("Failed to add SCSI device(id:lun=%d:%d), rc = %d\n",
		    DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex), rc);
	} else {
		u8 EventScratchPad[16] = {0};
		u32 ErrorCode = ARCSAS_EVENT_DEVICE_PLUG;
		u8 EventType = ARCSAS_EVENT_TYPE_MESSAGE_ONLY;
		PCore_Extension pCore = pDevice->pCore;
		//pDevice->psdev = osm_scsi_device_lookup(pACB->host, 0, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex),
		//DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
		osm_printk("SCSI device(scsi:id:lun=%d:%d:%d), added @phy %d\n", pACB->host->host_no,
		    DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex),
		        pDevice->Phy_Id);
		if (pDevice->Dev_Type == DT_ENCLOSURE) {
			u16 ExpEntryIndex;
			EventType = ARCSAS_EVENT_TYPE_ENCLOSURE;
			ExpEntryIndex = pACB->DevicesExpandersPMs_EntryIndex[pDevice->pExpander->EntryIndex];
			snprintf(EventScratchPad, sizeof(EventScratchPad), "Exp#%d Enc#%d", (u32)ExpEntryIndex, (u32)pDevice->EntryIndex);
		} else if ((pDevice->pPM!=NULL) && (pDevice->Connection & DC_ATA)) {
			u8 CorePhyNum;
			EventType = ARCSAS_EVENT_TYPE_PM_SLOT;
			CorePhyNum = (pDevice->pPM->CorePhyNum + ((pCore->core_flag & ARCSAS_CORE_FLAG_HBA_CORE_ID1) ? 4 : 0));
			snprintf(EventScratchPad, sizeof(EventScratchPad), "PM#%d Slot#%d", (u32)CorePhyNum, (u32)pDevice->pm_port_no);
		} else if ((pDevice->pExpander!=NULL) && (pDevice->Connection & DC_SCSI)) {
			u16 ExpEntryIndex;
			EventType = ARCSAS_EVENT_TYPE_EXPANDER_SLOT;
			ExpEntryIndex = pACB->DevicesExpandersPMs_EntryIndex[pDevice->pExpander->EntryIndex];
			snprintf(EventScratchPad, sizeof(EventScratchPad), "Exp#%d Slot#%d", (u32)ExpEntryIndex,
			    (u32)pDevice->SesSlotNumber);
		}
		if (pDevice->Connection & DC_SGPIO) {
			EventType = ARCSAS_EVENT_TYPE_HBA_SLOT;
			snprintf(EventScratchPad, sizeof(EventScratchPad), "HBA Slot#%d", (u32)pDevice->EntryIndex);
		}
		arcsas_construct_event_element(pCore, pDevice->EntryIndex, ErrorCode, EventType, EventScratchPad);
		//OSM_INIT_WORK((struct work_struct *)&(pDevice->psdev->hostdata), arcsas_remove_scsi_device_wqfn, pDevice->psdev);
	//	arcsas_free_dv_event_work(pACB, pDevice->dv_event);
	}
}
/*
************************************************************************
************************************************************************
*/
void arcsas_remove_scsi_device_wqfn(void *ptr)
{
	u8 EventScratchPad[16] = {0};
	u8 EventType = ARCSAS_EVENT_TYPE_MESSAGE_ONLY;
	u32 ErrorCode = ARCSAS_EVENT_DEVICE_UNPLUG;
	struct _Domain_Device *pDevice;
	PCore_Extension pCore;
	PACB pACB;
	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PLUG_UNPLUG)
		osm_printk("%s:\n", __func__);
	#endif
	pDevice = (struct _Domain_Device *)ptr;
	pCore = pDevice->pCore;
	pACB = pCore->pACB;
	#if ARCSAS_DBG_PLUG_UNPLUG
		osm_printk("SCSI device(0x%p, id:lun=%d:%d) going to be removed\n", pDevice->psdev,
		    DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
	#endif
	if (pDevice->psdev != NULL) {
		if (IS_SGPIO(pDevice)) {
	#if ARCSAS_SUPPORT_SGPIO
			arcsas_core_sgpio_set_led(pDevice, LED_FLAG_ALL_OFF);
	#endif
		} else if (IS_SSP(pDevice) || IS_STP(pDevice)) {
	#if ARCSAS_SUPPORT_SES
			arcsas_core_ses_set_led(pDevice, LED_FLAG_ALL_OFF);
	#endif
		}
		#if ARCSAS_DBG_PLUG_UNPLUG
			osm_printk("%s: No.2\n", __func__);
		#endif
		if ((pDevice->pPM != NULL) && (pDevice->Connection & DC_ATA)) {
			u8 CorePhyNum;
			EventType = ARCSAS_EVENT_TYPE_PM_SLOT;
			#if ARCSAS_DBG_PLUG_UNPLUG
				osm_printk("%s: No.3\n", __func__);
			#endif
			CorePhyNum = (pDevice->pPM->CorePhyNum + ((pCore->core_flag & ARCSAS_CORE_FLAG_HBA_CORE_ID1) ? 4 : 0));
			snprintf(EventScratchPad, sizeof(EventScratchPad),"Pm#%d Slot#%d", (u32)CorePhyNum, (u32)pDevice->pm_port_no);
		} else if ((pDevice->pExpander != NULL) && (pDevice->Connection & DC_SCSI)) {
			u16 ExpEntryIndex;
			EventType = ARCSAS_EVENT_TYPE_EXPANDER_SLOT;
			#if ARCSAS_DBG_PLUG_UNPLUG
				osm_printk("%s: No.4\n", __func__);
			#endif
			ExpEntryIndex = pACB->DevicesExpandersPMs_EntryIndex[pDevice->pExpander->EntryIndex];
			snprintf(EventScratchPad, sizeof(EventScratchPad), "Exp#%d Slot#%d", (u32)ExpEntryIndex,
			    (u32)pDevice->SesSlotNumber);
		} else if (pDevice->Connection & DC_SGPIO) {
			EventType = ARCSAS_EVENT_TYPE_HBA_SLOT;
			#if ARCSAS_DBG_PLUG_UNPLUG
				osm_printk("%s: No.5\n", __func__);
			#endif
			snprintf(EventScratchPad, sizeof(EventScratchPad), "HBA Slot#%d",
			    (u32)DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex));
		}
		#if ARCSAS_DBG_PLUG_UNPLUG
			osm_printk("%s: No.6\n", __func__);
		#endif
		arcsas_construct_event_element(pCore, pDevice->EntryIndex, ErrorCode, EventType, EventScratchPad);
		scsi_remove_device(pDevice->psdev);
		pDevice->psdev = NULL;
		osm_printk("SCSI device scsi:id:lun=%x:%x:%x was removed\n", pACB->host->host_no,
		    DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
		arcsas_free_dv_event_work(pACB, pDevice->dv_event);
	}
	#if ARCSAS_DBG_PLUG_UNPLUG
		osm_printk("%s: No.9\n", __func__);
	#endif
}
/*
************************************************************************
************************************************************************
*/
void arcsas_do_smp_state_machine_wqfn(void *ptr)
{
	struct _Discover_Extension *pDiscovery;
	PDISCOVERY_ELEMENT pDiscovery_element;
	PACB pACB;
	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_DISCOVERY)
		osm_printk("%s:\n", __func__);
	#endif
	pDiscovery = (struct _Discover_Extension *)ptr;
	pACB = pDiscovery->Target->pPort->pCore->pACB;
	pDiscovery_element = pACB->pDiscovery_element;
	arcsas_discovery_smp_state_machine(pDiscovery_element->pDiscovery, pDiscovery_element->pPort,
	    pDiscovery_element->nTgt, pDiscovery_element->CurrentState, pDiscovery_element->context, pDiscovery_element->pBuffer);
	pACB->discoveryQ_count--;
	#if ARCSAS_DBG_DISCOVERY_CONTEXT
		osm_printk("%s: No.1\n", __func__);
	#endif
	return;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_do_smp_phy_control_reset_wqfn(void *ptr)
{
	struct _Discover_Extension *pDiscovery;
	PACB pACB;
	PDISCOVERY_ELEMENT pDiscovery_element;
	PDomain_Port pPort;
	unsigned char nTgt;
	unsigned char PhyId;
	unsigned char CurrentState;
	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_DISCOVERY)
		osm_printk("%s:\n", __func__);
	#endif
	pDiscovery = (struct _Discover_Extension *)ptr;
	pACB = pDiscovery->Target->pPort->pCore->pACB;
	pDiscovery_element = pACB->pDiscovery_element;
	pPort = pDiscovery_element->pPort;
	nTgt = pDiscovery_element->nTgt;
	PhyId = pDiscovery_element->PhyId;
	CurrentState = pDiscovery_element->CurrentState;
	pDiscovery->numOfDiscoveryReq++;
	if (arcsas_post_smp_issue_phy_control_reset(pDiscovery, pPort, nTgt, PhyId) == STATUS_REQUEST_OK) {
	       /*issue PHY_CONTROL SMP function ,let sata oob done*/
		osm_printk("issue phy control .. reset ....................\n");
		pACB->discoveryQ_count--;
	} else {
		osm_printk("arcsas_post_smp_issue_phy_control_reset, we got a queue full....\n");
		osm_printk("issue phy control .. reset ....not ok................\n");
		pACB->discoveryQ_count--;
		pDiscovery->numOfDiscoveryReq--;
		if ((list_empty(&(&pDiscovery->arcsas_do_smp_phy_control_reset)->entry))) {
			arcsas_construct_smp_state_element(pDiscovery, pPort, nTgt, PhyId, CurrentState, 0, NULL);
			schedule_work(&pDiscovery->arcsas_do_smp_phy_control_reset);
		}
	}
	osm_printk("%s: line %d numOfDiscoveryReq = %d\n", __func__, __LINE__, pDiscovery->numOfDiscoveryReq);
	return;
}
#endif
/*
************************************************************************
************************************************************************
*/
void arcsas_os_ccb_done(struct AdapterControlBlock *pACB, struct CommandControlBlock *pCCB)
{
	struct scsi_cmnd *pscsi_cmnd;
	PARCSAS_Sense_Data senseBuffer;
	PDomain_Device pDevice;
	PCore_Extension pCore;

	pscsi_cmnd = (struct scsi_cmnd *)pCCB->Org_Req;
	if (pscsi_cmnd == NULL)
		BUG();
//	osm_printk("cdb[0] = 0x%x, ccb = %p\n", pscsi_cmnd->cmnd[0], pCCB);
//	osm_printk("ID:LUN=%d:%d, cdb[0] = 0x%x, CCB_STATUS = 0x%x, pcmd = %p, ccb = %p\n",
//	    pscsi_cmnd->device->id, (u16)pscsi_cmnd->device->lun, pscsi_cmnd->cmnd[0], pCCB->CCB_Status, pscsi_cmnd, pCCB);
	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB)
		if (!SCSI_IS_READ(pscsi_cmnd->cmnd[0]) && !SCSI_IS_WRITE(pscsi_cmnd->cmnd[0]))
		osm_printk("%s: ID:LUN=%d:%d, scmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x], buffer_len(0x%x), start_time = , request = 0x%p\n", __func__,
			pscsi_cmnd->device->id,
			(u16)pscsi_cmnd->device->lun,
			pscsi_cmnd,
			pscsi_cmnd->cmnd[0],
			pscsi_cmnd->cmnd[1],
			pscsi_cmnd->cmnd[2],
			pscsi_cmnd->cmnd[3],
			pscsi_cmnd->cmnd[4],
			pscsi_cmnd->cmnd[5],
			pscsi_cmnd->cmnd[6],
			pscsi_cmnd->cmnd[7],
			pscsi_cmnd->cmnd[8],
			pscsi_cmnd->cmnd[9],
			arcsas_rq_bflen(pscsi_cmnd),
		//	pscsi_cmnd->request->start_time,
			pscsi_cmnd->request
		);
	#endif
	senseBuffer = (PARCSAS_Sense_Data)pscsi_cmnd->sense_buffer;
	pDevice = pCCB->pDevice;
	pCore = pDevice->pCore;
	switch (pCCB->CCB_Status) {
		case CCB_STATUS_SUCCESS:
			pscsi_cmnd->result = DID_OK << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_SUCCESS, DID_OK\n", __func__, pscsi_cmnd, pCCB);
			#endif
			break;
		case CCB_STATUS_MEDIA_ERROR:
			pscsi_cmnd->result = DID_PARITY << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_MEDIA_ERROR, DID_ABORT\n",
				    __func__, pscsi_cmnd, pCCB);
			#endif
			break;
		case CCB_STATUS_BUSY:
			pscsi_cmnd->result = DID_BUS_BUSY << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_BUSY, DID_BUS_BUSY\n",
				    __func__, pscsi_cmnd, pCCB);
			#endif
			break;
		case CCB_STATUS_NO_DEVICE:
			pscsi_cmnd->result = DID_NO_CONNECT << 16  | COMMAND_COMPLETE << 8;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_PORT_ABORT_CCB)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_NO_DEVICE, DID_NO_CONNECT\n",
				    __func__, pscsi_cmnd, pCCB);
				osm_printk("%s: pCore->Running_CCB[0x%x] = 0x%p\n", __func__,
				    pCCB->SlotNo, pCore->Running_CCB[pCCB->SlotNo]);
			#endif
			break;
		case CCB_STATUS_HAS_SENSE://The ccb has sense data inside
			#if (ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_HAS_SENSE_ERROR)
				osm_printk("%s: pCCB = 0x%p, CCB_STATUS_HAS_SENSE\n", __func__, pCCB);
				osm_printk("%s: scsi_cmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x], id:lun = %d:%d\n", __func__,
					pscsi_cmnd,
					pscsi_cmnd->cmnd[0],
					pscsi_cmnd->cmnd[1],
					pscsi_cmnd->cmnd[2],
					pscsi_cmnd->cmnd[3],
					pscsi_cmnd->cmnd[4],
					pscsi_cmnd->cmnd[5],
					pscsi_cmnd->cmnd[6],
					pscsi_cmnd->cmnd[7],
					pscsi_cmnd->cmnd[8],
					pscsi_cmnd->cmnd[9],
					pscsi_cmnd->device->id,
					(u16)pscsi_cmnd->device->lun
				);
			#endif
			senseBuffer->ResponseCode = SCSI_SENSE_CURRENT_ERRORS;
			/* Sense buffer data is valid already. */
			//senseBuffer->Valid = 1;
		//	osm_printk("arcsas: sense Valid=0x%x, sense key=0x%x, ASC=0x%x, ASCQ=0x%x\n", senseBuffer->Valid, senseBuffer->SenseKey, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier);
			switch (senseBuffer->SenseKey) {
				case SCSI_SK_NO_SENSE:
				case SCSI_SK_BLANK_CHECK:
				case SCSI_SK_RECOVERED_ERROR:
					pscsi_cmnd->result = DID_OK << 16;
					break;
				case SCSI_SK_MEDIUM_ERROR:
					#if (ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_HAS_SENSE_ERROR)
						if (senseBuffer->SenseKey == SCSI_SK_RECOVERED_ERROR) {
							osm_printk("%s: SenseKey = SCSI_SK_RECOVERED_ERROR\n", __func__);
							osm_printk("%s: ASC = 0x%x, ASCQ = 0x%x, SKS[0] = 0x%x, SKS[1] = 0x%x, SKS[2] = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier, senseBuffer->SenseKeySpecific[0], senseBuffer->SenseKeySpecific[1], senseBuffer->SenseKeySpecific[2]);
						} else {
							osm_printk("%s: SenseKey = SCSI_SK_MEDIUM_ERROR\n", __func__);
							osm_printk("%s: ASC = 0x%x, ASCQ = 0x%x, SKS[0] = 0x%x, SKS[1] = 0x%x, SKS[2] = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier, senseBuffer->SenseKeySpecific[0], senseBuffer->SenseKeySpecific[1], senseBuffer->SenseKeySpecific[2]);
						}
					#endif
					if (pDevice->Connection & DC_ATAPI)
						break;
					if (pDevice->EntryIndex != ID_NOT_MAPPED) {
						pDevice->CmdErrorCount++;
						if (pDevice->CmdErrorCount > ARCSAS_MAX_DEVICE_RECOVERED_MEDIUM_ERROR_COUNT) {
							arcsas_remove_device(pDevice);
							pscsi_cmnd->result = DID_NO_CONNECT << 16;
						}
					} else {
						pscsi_cmnd->result = DID_NO_CONNECT << 16;
					}
					break;
				case SCSI_SK_HARDWARE_ERROR:
					#if (ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_HAS_SENSE_ERROR)
						osm_printk("%s: SCSI_SK_HARDWARE_ERROR\n", __func__);
						osm_printk("%s: ASC = 0x%x, ASCQ = 0x%x, SKS[0] = 0x%x, SKS[1] = 0x%x, SKS[2] = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier, senseBuffer->SenseKeySpecific[0], senseBuffer->SenseKeySpecific[1], senseBuffer->SenseKeySpecific[2]);
					#endif
					if (pDevice->Connection & DC_ATAPI)
						break;
					if (pDevice->EntryIndex != ID_NOT_MAPPED) {
						arcsas_remove_device(pDevice);
					}
					pscsi_cmnd->result = DID_NO_CONNECT << 16;
					break;
				case SCSI_SK_ILLEGAL_REQUEST:
					#if (ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_HAS_SENSE_ERROR)
						osm_printk("%s: SenseKey = SCSI_SK_ILLEGAL_REQUEST\n", __func__);
						osm_printk("%s: ASC = 0x%x, ASCQ = 0x%x, SKS[0] = 0x%x, SKS[1] = 0x%x, SKS[2] = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier, senseBuffer->SenseKeySpecific[0], senseBuffer->SenseKeySpecific[1], senseBuffer->SenseKeySpecific[2]);
					#endif
					if (pDevice->Connection & DC_ATAPI)
						break;
					if ((pDevice->EntryIndex != ID_NOT_MAPPED) && (senseBuffer->AdditionalSenseCode == SCSI_ASC_LOGICAL_UNIT_NOT_SUPPORTED))/* scsi id lun */{
						if (DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex))/* scsi id lun */ {
							pDevice->Status = DEVICE_STATUS_NO_DEVICE;
							if (IS_ENCLOSURE(pDevice)) {
								arcsas_free_enclosure_entry_index(pACB, pDevice->EntryIndex);/* scsi id lun */
							}
							arcsas_port_abort_ccbs(pDevice->pPort, CCB_STATUS_NO_DEVICE, pDevice);
							if (pDevice->pExpander != NULL) {
								osm_list_del_init(&pDevice->Exp_Queue_Pointer);
								osm_atomic_dec(&pDevice->pExpander->Device_Number);
							}
							osm_list_del_init(&pDevice->Port_Queue_Pointer);
							osm_atomic_dec(&pDevice->pPort->Device_Number);
							arcsas_free_lun_device(pCore, pDevice);
						} else {
							if (arcsas_device_babies_lun_check(pDevice)) {  
								arcsas_port_abort_ccbs(pDevice->pPort, CCB_STATUS_NO_DEVICE, pDevice);
								pDevice->Dev_Type = DT_LOGICAL_UNIT_NOT_PRESENT_DEVICE;
								pDevice->State = DEVICE_STATE_INIT_DONE;
							} else {
								arcsas_remove_device(pDevice); /*need handle if this device had LUN device*/
							}
						}
					}
					pscsi_cmnd->result = DID_ABORT << 16;
					break;
				case SCSI_SK_ABORTED_COMMAND:
					#if (ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_HAS_SENSE_ERROR)
						osm_printk("%s:SenseKey = SCSI_SK_ABORTED_COMMAND, ASC = 0x%x, ASCQ = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier);
					#endif
					pscsi_cmnd->result = DID_ABORT << 16;
					break;
				case SCSI_SK_NOT_READY:
					#if (ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_HAS_SENSE_ERROR)
						osm_printk("%s:SenseKey = SCSI_SK_NOT_READY, ASC = 0x%x, ASCQ = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier);
					#endif
					pscsi_cmnd->result = DID_NO_CONNECT << 16;
					break;
				default:
					#if (ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_HAS_SENSE_ERROR)
						osm_printk("%s: SenseKey = 0x%x, ASC = 0x%x, ASCQ = 0x%x\n", __func__, senseBuffer->SenseKey, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier);
					#endif
					pscsi_cmnd->result = DID_ABORT << 16;
					break;
			}
#if LINUX_VERSION_CODE <= KERNEL_VERSION(5, 13, 0)
			pscsi_cmnd->result = DRIVER_SENSE << 24 | SAM_STAT_CHECK_CONDITION;
#else
			pscsi_cmnd->result = (DID_OK << 16) | SAM_STAT_CHECK_CONDITION;
#endif
			break;
		case CCB_STATUS_ERROR_WITH_SENSE://This ccb is from API only
			senseBuffer->ResponseCode = SCSI_SENSE_CURRENT_ERRORS;
			pscsi_cmnd->result = DID_OK << 16;
			/* Sense buffer data is valid already. */
			osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_ERROR_WITH_SENSE\n", __func__, pscsi_cmnd, pCCB);
			break;
		case CCB_STATUS_SUCCESS_WITH_SENSE: //This ccb is from passthrough cmd only
#if LINUX_VERSION_CODE <= KERNEL_VERSION(5, 13, 0)
			pscsi_cmnd->result = DRIVER_SENSE << 24 | SAM_STAT_CHECK_CONDITION;
#else
			pscsi_cmnd->result = (DID_OK << 16) | SAM_STAT_CHECK_CONDITION;
#endif
			break;
		case CCB_STATUS_CHECK_CONDITION:
			pscsi_cmnd->result = DID_OK << 16 | SAM_STAT_CHECK_CONDITION;
			break;
		case CCB_STATUS_REQUEST_SENSE: {
			struct CommandControlBlock	*pCCBNew;

			pCCBNew = arcsas_core_get_coreccb_from_list(pCore);
			if (pCCBNew)
				arcsas_core_send_request_sense_ccb(pCore, pDevice, pCCBNew, pCCB);
			return;
			}
		case CCB_STATUS_ERROR:
			pscsi_cmnd->result = DID_ERROR << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_TIMEOUT)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_ERROR, DID_ERROR\n", __func__, pscsi_cmnd, pCCB);
				osm_printk("%s: pCore->Running_CCB[0x%x] = 0x%p\n", __func__, pCCB->SlotNo, pCore->Running_CCB[pCCB->SlotNo]);
			#endif
			break;
		case CCB_STATUS_INVALID_REQUEST:
			pscsi_cmnd->result = DID_ABORT << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: pCCB = 0x%p scsi_dev = 0x%p CCB_STATUS_INVALID_REQUEST\n", __func__, pCCB, pscsi_cmnd);
			osm_printk("%s: scsi_cmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x], id:lun = %d:%d\n", __func__,
				pscsi_cmnd,
				pscsi_cmnd->cmnd[0],
				pscsi_cmnd->cmnd[1],
				pscsi_cmnd->cmnd[2],
				pscsi_cmnd->cmnd[3],
				pscsi_cmnd->cmnd[4],
				pscsi_cmnd->cmnd[5],
				pscsi_cmnd->cmnd[6],
				pscsi_cmnd->cmnd[7],
				pscsi_cmnd->cmnd[8],
				pscsi_cmnd->cmnd[9],
				pscsi_cmnd->device->id,
				(u16)pscsi_cmnd->device->lun
			);
			#endif
			break;
		case CCB_STATUS_DATA_OVER_RUN:
			pscsi_cmnd->result = DID_OK << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_PORT_ABORT_CCB)
				osm_printk("%s: pscsi_cmnd = 0x%p, ID:LUN = %d: %d pCCB = 0x%p CCB_STATUS_DATA_OVER_RUN\n",
				    __func__, pscsi_cmnd, pscsi_cmnd->device->id, (u16)pscsi_cmnd->device->lun, pCCB);
			#endif
			break;
		case CCB_STATUS_DID_ABORT:
			pscsi_cmnd->result = DID_ABORT << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_PORT_ABORT_CCB)
				osm_printk("%s: pscsi_cmnd = 0x%p, ID:LUN = %d: %d pCCB = 0x%p CCB_STATUS_DID_ABORT\n",
				    __func__, pscsi_cmnd, pscsi_cmnd->device->id, (u16)pscsi_cmnd->device->lun, pCCB);
			#endif
			break;	
		case CCB_STATUS_DID_RESET:
			pscsi_cmnd->result = DID_SOFT_ERROR << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_PORT_ABORT_CCB)
				osm_printk("%s: pscsi_cmnd = 0x%p, ID:LUN = %d: %d pCCB = 0x%p CCB_STATUS_DID_RESET\n",
				    __func__, pscsi_cmnd, pscsi_cmnd->device->id, (u16)pscsi_cmnd->device->lun, pCCB);
			#endif
			break;
		case CCB_STATUS_DID_TIMEOUT:
			pscsi_cmnd->result = DID_TIME_OUT << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_PORT_ABORT_CCB)
				osm_printk("%s: pscsi_cmnd = 0x%p, ID:LUN = %d: %d pCCB = 0x%p CCB_STATUS_DID_TIMEOUT, serial_number = , start_time = , request = 0x%p\n", __func__, pscsi_cmnd, pscsi_cmnd->device->id, (u16)pscsi_cmnd->device->lun, pCCB, pscsi_cmnd->request);
			#endif
			break;
		default:
			pscsi_cmnd->result = DID_ABORT << 16;
	}
	if (pCCB->Scratch_Buffer)
		arcsas_core_free_sas_scratch_buffer_to_list(pCore, pCCB->Scratch_Buffer);
	if (arcsas_use_sg(pscsi_cmnd)) {
		#if ARCSAS_DBG_OS_CCB_CONTEXT
			osm_printk("%s: No.1 scsi cmd buffer = 0x%p\n", __func__, arcsas_rq_bf(pscsi_cmnd));
		#endif
		arcsas_unmap_sg(pscsi_cmnd);
	} else if (arcsas_rq_bflen(pscsi_cmnd)) {
		#if ARCSAS_DBG_OS_CCB_CONTEXT
			osm_printk("%s: No.2\n", __func__);
		#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
		dma_unmap_single(&(pACB->pcidev->dev), pCCB->dma_handle, arcsas_rq_bflen(pscsi_cmnd),
		    pscsi_cmnd->sc_data_direction);
#else
		pci_unmap_single(pACB->pcidev, pCCB->dma_handle, arcsas_rq_bflen(pscsi_cmnd),
		    pscsi_cmnd->sc_data_direction);
#endif
	}
	arcsas_free_coreccb_to_pool(pCore, pCCB);
	arcsas_post_waiting_list_ccb(pCore, pDevice);
	pDevice->queued_cmd_request--;
#if	LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 21)
	pscsi_cmnd->scsi_done(pscsi_cmnd);
#else
	scsi_done(pscsi_cmnd);
#endif
	return;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_os_sas_send_report_lun_taskccb(PDomain_Device pDevice, PCCB pOrigCCB)
{
	struct _Core_Extension *pCore;
	PCCB pNewCCB;
	PSAS_Scratch_Buffer pSASSB;
	PARCSAS_SG_Table pSGTable;
	unsigned int length;

	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB)
		osm_printk("%s:\n", __func__);
	#endif
	pCore = pDevice->pCore;
	pNewCCB = arcsas_core_get_coreccb_from_list(pCore);
	if (pNewCCB == NULL) {
		/* no internal requests available - wait for now */
		pDevice->Is_Waiting = true; 
		return;
	}
	pSASSB = arcsas_core_get_sas_scratch_buffer_from_list(pCore);
	if (pSASSB == NULL) {
		unsigned long flags;
		pDevice->Is_Waiting = true;
		arcsas_spin_lock_irqsave(&pCore->ccb_list_lock, &flags);
		osm_list_add_tail(&pNewCCB->Queue_Pointer, &pCore->Core_CCB_List);
		if (pNewCCB->CoreContext) {
			arcsas_free_core_context_to_pool(pCore, &pNewCCB->CoreContext);
		}
		arcsas_spin_unlock_irqrestore(&pCore->ccb_list_lock, &flags);
		#if ARCSAS_DBG_OS_CCB_CONTEXT
			osm_printk("ERROR: No more free SAS scratch buffer. Request aborted.\n");
		#endif
		return;
	}
	pSGTable = &pNewCCB->SG_Table;
	pNewCCB->Cdb[0] = SCSI_CMD_REPORT_LUNS;
	osm_memset(&pNewCCB->Cdb[1], 0, 5); /* bytes 1 - 5: reserved,set to zero. Cdb[2]:00b ....SELECT REPORT  Peripheral device addressing method */
	length = (ARCSAS_LUN_SUPPORTED + 1) * 8;
	pNewCCB->Cdb[6] = (unsigned char) (length >> 24) & 0xff; /*MSB:ALLOCATION LENGTH*/
	pNewCCB->Cdb[7] = (unsigned char) (length >> 16) & 0xff;
	pNewCCB->Cdb[8] = (unsigned char) (length >> 8) & 0xff;
	pNewCCB->Cdb[9] = (unsigned char) length & 0xff; /*LSB:ALLOCATION LENGTH*/
	pNewCCB->Cdb[10] = 0; /* reserved */
	pNewCCB->Cdb[11] = 0; /* control */
	pNewCCB->Tag = 0xaa;
	pNewCCB->Org_Req = pOrigCCB;
	pNewCCB->EntryIndex = pDevice->EntryIndex;
	pNewCCB->pDevice = pDevice;
	pNewCCB->Cmd_Initiator = (void *)pCore;
	pNewCCB->Data_Transfer_Length = length;/* length need to enlarge and support 128 lun device if needed */
	pNewCCB->Cmd_Flag |= CMD_FLAG_DATA_IN;
	pNewCCB->Data_Buffer = pSASSB->Scratch_Vir;
	pNewCCB->Scratch_Buffer = pSASSB;
	pNewCCB->Completion = (void(*)(void *, struct CommandControlBlock *))arcsas_os_sas_core_ccb_callback;
	arcsas_sg_table_init(pSGTable, ARCSAS_CORE_SG_ENTRIES);/* Make the SG table. */
	arcsas_sg_table_append(pSGTable, pSASSB->Scratch_DMA.parts.low, pSASSB->Scratch_DMA.parts.high, length);
	arcsas_post_ccb(pCore, pNewCCB);/* Send this internal request */
	return;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_done_os_inquiry_ccb_callback(PACB pACB, PCCB pCCB)
{
	struct scsi_cmnd *pscsi_cmd;
	PSCSI_INQ_DATA pInquiryData = NULL;
	struct scatterlist *sg;
	unsigned char *buffer;
	#if ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB
		osm_printk("%s:\n", __func__);
	#endif
	#if ARCSAS_DBG_OS_CCB_CONTEXT
		osm_printk("%s: pCCB = 0x%p\n", __func__, pCCB);
	#endif
	if ((DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pCCB->EntryIndex) == 0) && (pCCB->CCB_Status != CCB_STATUS_SUCCESS)) {
		pscsi_cmd = (struct scsi_cmnd *)pCCB->Org_Req;
		sg = arcsas_rq_bf(pscsi_cmd);
		buffer = arcsas_map_sg_page(sg) + sg->offset;
		pInquiryData = (PSCSI_INQ_DATA)buffer;
		pInquiryData->DeviceType = 0x0D;
		pInquiryData->Version = 1;
		pInquiryData->AdditionalLength = 31;
		osm_memcpy(pInquiryData->VendorId, "  Areca  ", 8);
		osm_memcpy(pInquiryData->ProductId, " DummyDevice    ", 16);
		arcsas_unmap_buffer(buffer);
		pCCB->CCB_Status = CCB_STATUS_SUCCESS;
		#if ARCSAS_DBG_OS_CCB_CONTEXT
			osm_printk("%s: Areca DummyDevice\n", __func__);
		#endif
	}
	arcsas_os_ccb_done(pACB, pCCB);
	return;
}
/*
************************************************************************
************************************************************************
*/
PSMP_Scratch_Buffer arcsas_core_get_smp_scratch_from_list(PCore_Extension pCore)
{
	PSMP_Scratch_Buffer pSMPSB;
	unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock_irqsave(&pCore->smpsb_list_lock, &flags);
	if (!osm_list_empty(&pCore->SMP_Scratch_List)) {
		pSMPSB = (PSMP_Scratch_Buffer)osm_list_first_entry(&pCore->SMP_Scratch_List, struct _SMP_Scratch_Buffer, Queue_Pointer);
		osm_list_del_init(&pSMPSB->Queue_Pointer);
		arcsas_spin_unlock_irqrestore(&pCore->smpsb_list_lock, &flags);
		return pSMPSB;
	}
	arcsas_spin_unlock_irqrestore(&pCore->smpsb_list_lock, &flags);
	return NULL;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_core_free_smp_scratch_to_list(PCore_Extension pCore, PSMP_Scratch_Buffer *pSMPSB)
{
	unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	// TBD: reset SMP Request - what values need resetting?
	if (*pSMPSB) {
		arcsas_spin_lock_irqsave(&pCore->smpsb_list_lock, &flags);
		#if ARCSAS_DBG_LINKEDLIST
			osm_printk("line %d \n", __LINE__);
		#endif
		osm_list_add_tail(&(*pSMPSB)->Queue_Pointer, &pCore->SMP_Scratch_List);
		*pSMPSB = NULL;
		arcsas_spin_unlock_irqrestore(&pCore->smpsb_list_lock, &flags);
	}
	return;
}
/*
************************************************************************
************************************************************************
*/
PPRD_Buffer arcsas_core_get_prd_buffer_from_list(PCore_Extension pCore)
{
	PPRD_Buffer pPRD_Buffer;
	unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock_irqsave(&pCore->prd_list_lock, &flags);
	if (osm_list_empty(&pCore->PRD_Buffer_List)) {
		arcsas_spin_unlock_irqrestore(&pCore->prd_list_lock, &flags);
		return NULL;
	}
	pPRD_Buffer = (PPRD_Buffer)osm_list_first_entry(&pCore->PRD_Buffer_List, struct _PRD_Buffer, Queue_Pointer);
	osm_list_del_init(&pPRD_Buffer->Queue_Pointer);
	arcsas_spin_unlock_irqrestore(&pCore->prd_list_lock, &flags);
	return pPRD_Buffer;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_core_free_prd_buffer_to_list(PCore_Extension pCore, PPRD_Buffer *pPRD_Buffer)
{
	unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif

	if (*pPRD_Buffer) {
		arcsas_spin_lock_irqsave(&pCore->prd_list_lock, &flags);
		osm_list_add_tail(&(*pPRD_Buffer)->Queue_Pointer, &pCore->PRD_Buffer_List);
		*pPRD_Buffer = NULL;
		arcsas_spin_unlock_irqrestore(&pCore->prd_list_lock, &flags);
	}
	return;
}
/*
************************************************************************
************************************************************************
*/
unsigned char arcsas_get_next_active_dep_entry_index(PACB pACB, unsigned short dep_entry_index)
{
	unsigned short next_dep_entry_index, temp_dep_entry_index;
	unsigned char i;

	if (dep_entry_index == ID_NOT_MAPPED) {
		for (i = 0; i < ARCSAS_DISK_SUPPORTED; i++) {
			if ((temp_dep_entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, i)) == ID_NOT_MAPPED) {
				continue;
			}
			return temp_dep_entry_index;
		}
	} else {
		for (i = 0; i < ARCSAS_DISK_SUPPORTED; i++) {
			if ((temp_dep_entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, i)) != dep_entry_index) {
				continue;
			}
			for (i = i + 1; i < ARCSAS_DISK_SUPPORTED; i++) {
				if ((next_dep_entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, i)) != ID_NOT_MAPPED) {
					return next_dep_entry_index;
				}
			}
		}
	}
	return ID_NOT_MAPPED;
}
/*
************************************************************************
************************************************************************
*/
PCore_Context arcsas_core_get_context_from_pool(PCore_Extension pCore)
{
	PCore_Context pContext;
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	/*remove the lock in case of segmentation fault in customer kernel 2.6.19-18gtc */
	arcsas_spin_lock(&pCore->context_list_lock);
	if (osm_list_empty(&pCore->Context_List)) {
		arcsas_spin_unlock(&pCore->context_list_lock);
		return NULL;
	}
	pContext = ((PCore_Context)osm_list_first_entry(&pCore->Context_List, struct _CORE_CONTEXT, Queue_Pointer));
	osm_list_del_init(&pContext->Queue_Pointer);
	/*Reset some variables.*/
	pContext->Context_Type = CORE_CONTEXT_TYPE_NONE;
	arcsas_spin_unlock(&pCore->context_list_lock);
	return pContext;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_free_core_context_to_pool(PCore_Extension pCore, PCore_Context *pContext)
{
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock(&pCore->context_list_lock);
	osm_list_add_tail(&(*pContext)->Queue_Pointer, &pCore->Context_List);
	*pContext = NULL;
	arcsas_spin_unlock(&pCore->context_list_lock);
	return;
}

/*
************************************************************************
************************************************************************
*/
void arcsas_zero_ccb(struct CommandControlBlock *pCCB)
{
	PARCSAS_SGENTRY pSGEntry;
	void *pSenseBuffer;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pSGEntry = pCCB->SG_Table.pSGEntry;
	pSenseBuffer = pCCB->Sense_Buffer;
	osm_memset(pCCB, 0, sizeof(struct CommandControlBlock));
	OSM_INIT_LIST_HEAD(&pCCB->Complete_Queue_Pointer);
	pCCB->SG_Table.pSGEntry = pSGEntry;
	pCCB->Sense_Buffer = pSenseBuffer;
	return;
}
/*
************************************************************************
************************************************************************
*/
struct CommandControlBlock *arcsas_core_get_coreccb_from_list(PCore_Extension pCore)
{
	unsigned long flags;
	struct CommandControlBlock *pCCB;
	struct AdapterControlBlock *pACB;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pACB = pCore->pACB;
	arcsas_spin_lock_irqsave(&pCore->ccb_list_lock, &flags);
	if (osm_list_empty(&pCore->Core_CCB_List)) {
		if (printk_ratelimit())
			osm_printk("ccb pool empty\n ");
		arcsas_spin_unlock_irqrestore(&pCore->ccb_list_lock, &flags);
		return NULL;
	}
	pCCB = (struct CommandControlBlock *)osm_list_first_entry(&pCore->Core_CCB_List, struct CommandControlBlock, Queue_Pointer);
	osm_list_del_init(&pCCB->Queue_Pointer);
	pCore->free_ccb_count--;
	arcsas_zero_ccb(pCCB);
	pCCB->CoreContext = arcsas_core_get_context_from_pool(pCore);
	arcsas_spin_unlock_irqrestore(&pCore->ccb_list_lock, &flags);
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s: pCCB = 0x%p\n", __func__, pCCB);
	#endif
	return pCCB;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_free_coreccb_to_pool(PCore_Extension pCore, struct CommandControlBlock *pCCB)
{
	unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pCCB->Scratch_Buffer = NULL;
	if (!IS_A_SMP_REQ(pCCB) && !IS_A_PM_REQ(pCCB)) {
		arcsas_spin_lock_irqsave(&(pCCB->pDevice->sent_ccb_list_lock), &(pCCB->pDevice->lock_flags));
		#if ARCSAS_DBG_CONTEXT
			osm_printk("%s: No.0\n", __func__);
		#endif
		osm_list_del_init(&pCCB->Queue_Pointer);
		arcsas_spin_unlock_irqrestore(&(pCCB->pDevice->sent_ccb_list_lock), &(pCCB->pDevice->lock_flags));
	} else if (IS_A_PM_REQ(pCCB)) {
		#if ARCSAS_DBG_CONTEXT
			osm_printk("%s: No.1\n", __func__);
		#endif
		arcsas_spin_lock_irqsave(&(pCCB->pPM->sent_ccb_list_lock), &flags);
		osm_list_del_init(&pCCB->Queue_Pointer);
		arcsas_spin_unlock_irqrestore(&(pCCB->pPM->sent_ccb_list_lock), &flags);
	} else {
		#if ARCSAS_DBG_CONTEXT
			osm_printk("%s: No.2\n", __func__);
		#endif
		arcsas_spin_lock_irqsave(&pCore->ccb_waiting_lst_lock, &flags);
		osm_list_del_init(&pCCB->Queue_Pointer);
		arcsas_spin_unlock_irqrestore(&pCore->ccb_waiting_lst_lock, &flags);
	}
	arcsas_spin_lock_irqsave(&pCore->ccb_list_lock, &flags);
	#if ARCSAS_DBG_CONTEXT
		osm_printk("%s: No.3\n", __func__);
	#endif
	osm_list_add_tail(&pCCB->Queue_Pointer, &pCore->Core_CCB_List);
	pCore->free_ccb_count++;
	if (pCCB->CoreContext) {
		arcsas_free_core_context_to_pool(pCore, &pCCB->CoreContext);
	}
	arcsas_spin_unlock_irqrestore(&pCore->ccb_list_lock, &flags);
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s: pCCB = 0x%p\n", __func__, pCCB);
	#endif
}
/*
************************************************************************
************************************************************************
*/
static void arcsas_teardown_dma_pool(struct AdapterControlBlock *pACB)
{
	unsigned char core_index;
	struct _Core_Extension *pCore;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	for (core_index = 0; core_index < pACB->core_num; core_index++) {
		pCore = pACB->pCore[core_index];
		if (pCore == NULL)
			continue;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
		if (pCore->sg_tbl_pool != NULL) {
			dma_pool_free(pCore->sg_tbl_pool, pCore->sg_tbl_pool_vaddr, pCore->sg_tbl_pool_dma);
			dma_pool_destroy(pCore->sg_tbl_pool);
		}
		if (pCore->sas_scrtch_pool != NULL) {
			dma_pool_free(pCore->sas_scrtch_pool, pCore->sas_scrtch_pool_vaddr, pCore->sas_scrtch_pool_dma);
			dma_pool_destroy(pCore->sas_scrtch_pool);
		}
		if (pCore->cmpl_queue_pool != NULL) {
			dma_pool_free(pCore->cmpl_queue_pool, pCore->cmpl_queue_pool_vaddr, pCore->cmpl_queue_pool_dma);
			dma_pool_destroy(pCore->cmpl_queue_pool);
		}
		if (pCore->post_queue_pool != NULL) {
			dma_pool_free(pCore->post_queue_pool, pCore->post_queue_pool_vaddr, pCore->post_queue_pool_dma);
			dma_pool_destroy(pCore->post_queue_pool);
		}
		if (pCore->rcv_fis_pool != NULL) {
			dma_pool_free(pCore->rcv_fis_pool, pCore->rcv_fis_pool_vaddr, pCore->rcv_fis_pool_dma);
			dma_pool_destroy(pCore->rcv_fis_pool);
		}
		if (pCore->cmnd_tbl_pool != NULL) {
			dma_pool_free(pCore->cmnd_tbl_pool, pCore->cmnd_tbl_pool_vaddr, pCore->cmnd_tbl_pool_dma);
			dma_pool_destroy(pCore->cmnd_tbl_pool);
		}
		if (pCore->cmnd_list_pool != NULL) {
			dma_pool_free(pCore->cmnd_list_pool, pCore->cmnd_list_pool_vaddr, pCore->cmnd_list_pool_dma);
			dma_pool_destroy(pCore->cmnd_list_pool);
		}
#else
		if (pCore->sg_tbl_pool != NULL) {
			pci_pool_free(pCore->sg_tbl_pool, pCore->sg_tbl_pool_vaddr, pCore->sg_tbl_pool_dma);
			pci_pool_destroy(pCore->sg_tbl_pool);
		}
		if (pCore->sas_scrtch_pool != NULL) {
			pci_pool_free(pCore->sas_scrtch_pool, pCore->sas_scrtch_pool_vaddr, pCore->sas_scrtch_pool_dma);
			pci_pool_destroy(pCore->sas_scrtch_pool);
		}
		if (pCore->cmpl_queue_pool != NULL) {
			pci_pool_free(pCore->cmpl_queue_pool, pCore->cmpl_queue_pool_vaddr, pCore->cmpl_queue_pool_dma);
			pci_pool_destroy(pCore->cmpl_queue_pool);
		}
		if (pCore->post_queue_pool != NULL) {
			pci_pool_free(pCore->post_queue_pool, pCore->post_queue_pool_vaddr, pCore->post_queue_pool_dma);
			pci_pool_destroy(pCore->post_queue_pool);
		}
		if (pCore->rcv_fis_pool != NULL) {
			pci_pool_free(pCore->rcv_fis_pool, pCore->rcv_fis_pool_vaddr, pCore->rcv_fis_pool_dma);
			pci_pool_destroy(pCore->rcv_fis_pool);
		}
		if (pCore->cmnd_tbl_pool != NULL) {
			pci_pool_free(pCore->cmnd_tbl_pool, pCore->cmnd_tbl_pool_vaddr, pCore->cmnd_tbl_pool_dma);
			pci_pool_destroy(pCore->cmnd_tbl_pool);
		}
		if (pCore->cmnd_list_pool != NULL) {
			pci_pool_free(pCore->cmnd_list_pool, pCore->cmnd_list_pool_vaddr, pCore->cmnd_list_pool_dma);
			pci_pool_destroy(pCore->cmnd_list_pool);
		}
#endif
	}	
}
/*
************************************************************************
************************************************************************
*/
Bool arcsas_tag_is_empty(PTag_Stack pTagStack)
{
	if (pTagStack->Top == 0)
		return true;
	else
		return false;
}
/*
************************************************************************
** Tag: index of DevicesExpandersPMs_EntryIndex[]
** PTag_Stack: Port_Tag_Pool, Expander_Tag_Pool, PM_Tag_Pool, CmdSlot_Tag_Pool, Device_Tag_Pool
**
** if TagStackType!=FIFO_TAG, use FILO,
** if TagStackType==FIFO_TAG, use FIFO, PtrOut is the next tag to get
** and Top is the number of available tags in the stack
** when use FIFO, get tag from PtrOut and free tag to (PtrOut+Top)%Size
************************************************************************
*/
unsigned short arcsas_get_pool_tag(PTag_Stack pTagStack)
{
	unsigned short nTag;
	unsigned long flags;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock_irqsave(&pTagStack->lock, &flags);
	if (pTagStack->TagStackType == FIFO_TAG) {
		nTag = pTagStack->Stack[pTagStack->PtrOut++];
		if (pTagStack->PtrOut >= pTagStack->Size) {
			pTagStack->PtrOut = 0;
		}
		pTagStack->Top--;
	} else {
		nTag = pTagStack->Stack[--pTagStack->Top];
	}
	arcsas_spin_unlock_irqrestore(&pTagStack->lock, &flags);
	return nTag;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_init_tag(PTag_Stack pTagStack, unsigned short size)
{
	unsigned short i;

	#if ARCSAS_DBG_INIT
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock_init(&pTagStack->lock);
	pTagStack->Top = size;
	pTagStack->TagStackType = FILO_TAG;
	pTagStack->PtrOut = 0;
	for (i = 0; i < size; i++) {
		pTagStack->Stack[i] = size - 1 - i;
	}
}
/*
************************************************************************
************************************************************************
*/
void arcsas_init_cmdslot(PTag_Stack pTagStack, unsigned short size)
{
	unsigned short i;

	#if ARCSAS_DBG_INIT
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock_init(&pTagStack->lock);
	pTagStack->Top = size;
	pTagStack->TagStackType = FIFO_TAG;
	pTagStack->PtrOut = 0;
	for (i = 0; i < size; i++) {
		pTagStack->Stack[i] = i;
	}
}
/*
************************************************************************
************************************************************************
*/
void arcsas_release_pool_tag(PTag_Stack pTagStack, unsigned short tag)
{
	unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock_irqsave(&pTagStack->lock, &flags);
//	arcsas_spin_lock(&pTagStack->lock);
	if (pTagStack->TagStackType == FIFO_TAG) {
		pTagStack->Stack[(pTagStack->PtrOut + pTagStack->Top) % pTagStack->Size] = tag;
		pTagStack->Top++;
	} else {
		pTagStack->Stack[pTagStack->Top++] = tag;
	}
	arcsas_spin_unlock_irqrestore(&pTagStack->lock, &flags);
//	arcsas_spin_unlock(&pTagStack->lock);
}
/*
************************************************************************
************************************************************************
*/
static Bool arcsas_core_get_DMA_memory(struct AdapterControlBlock *pACB, unsigned short PRDBufferCount,
    unsigned short slotCount, unsigned short deliverQueueSize, unsigned short completeQueueSize, unsigned short SASScratchCount)
{
	Assigned_Uncached_Memory dmaResource;
	PCore_Extension pCore = NULL;
	PSAS_Scratch_Buffer pSASSB;
	PPRD_Buffer pPRD_Buffer;
	ARCSAS_U64 memDMA, tmpPhyAddress, cmdTblPhyAddr, cmdStructPhyAddr;
	void *memVir;
	u32 offset, size;
	unsigned short i;
	PARCSAS_Command_Structure pCmdStruct;
	PARCSAS_Command_Header pCmdHeader;

	#if ARCSAS_DBG_INIT
		osm_printk("%s:\n", __func__);
	#endif
	switch (pACB->adapter_type) {
		case ARC_1330: {
			u32 size = 0; 
			unsigned short core_index;
			osm_dma_addr_t dma_addr;
			void *vaddr;
			struct dma_pool *dma_pool_temp;

			for (core_index = 0; core_index < pACB->core_num; core_index++) {
				pCore = pACB->pCore[core_index];
				/*
				************************************************************************
				**		Deploy Command Header + Table DMA
				************************************************************************ 
				*/
				size = sizeof(struct _ARCSAS_Command_Structure) * slotCount;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas command list pool", &(pACB->pcidev->dev), size, 64, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas command list pool", pACB->pcidev, size, 64, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create command list DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got command list pool\n", __func__);
				#endif
				pCore->cmnd_list_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->cmnd_list_pool, GFP_ATOMIC, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->cmnd_list_pool, GFP_ATOMIC, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc command list DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr)); 
				dmaResource.Byte_Size = size;
				pCore->cmnd_list_pool_dma = dma_addr;
				pCore->cmnd_list_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 64) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, size);
				pCore->CmdSlot_DMA = memDMA;
				pCore->CmdSlot = memVir;
				/*
				************************************************************************
				**		Init Command Header
				************************************************************************
				*/
				pCmdStruct = (PARCSAS_Command_Structure)pCore->CmdSlot;
				cmdStructPhyAddr.value = pCore->CmdSlot_DMA.value;
				cmdTblPhyAddr.value = pCore->CmdSlot_DMA.value + offsetof(struct _ARCSAS_Command_Structure, Command_Table);
				for (i=0; i < slotCount; i++) {
					pCmdHeader = &pCmdStruct->Command_Header;
					pCmdStruct->Cmd_Structure_Addr = ARCSAS_CPU_TO_LE32(cmdStructPhyAddr.parts.low);
					pCmdStruct->Cmd_Structure_Addr_High = ARCSAS_CPU_TO_LE32(cmdStructPhyAddr.parts.high);
					tmpPhyAddress = cmdTblPhyAddr;
					tmpPhyAddress.value += offsetof(struct _ARCSAS_Command_Table, Open_Addr_Frame);
					pCmdHeader->OpenAddrFrame_Address = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.low);
					pCmdHeader->OpenAddrFrame_Address_High = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.high);
					tmpPhyAddress = cmdTblPhyAddr;
					tmpPhyAddress.value += offsetof(struct _ARCSAS_Command_Table, Status_Buff);
					pCmdHeader->StatusBuff_Address = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.low);
					pCmdHeader->StatusBuff_Address_High = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.high);
					pCmdHeader->Interface_select = 0xC0C0C000;
					pCmdStruct++;
					cmdStructPhyAddr.value += sizeof(struct _ARCSAS_Command_Structure);
					cmdTblPhyAddr.value += sizeof(struct _ARCSAS_Command_Structure);
				}
				/*
				************************************************************************
				**		Deploy Received FIS DMA
				************************************************************************ 
				*/
				size = MAX_RX_FIS_POOL_SIZE_1330;/* Received FIS */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas received FIS pool",  &(pACB->pcidev->dev), size, 256, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas received FIS pool", pACB->pcidev, size, 256, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create received FIS DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got received FIS pool\n", __func__);
				#endif
				pCore->rcv_fis_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->rcv_fis_pool, GFP_ATOMIC, &dma_addr); 
#else
				vaddr = pci_pool_alloc(pCore->rcv_fis_pool, GFP_ATOMIC, &dma_addr); 
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc command list DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->rcv_fis_pool_dma = dma_addr;
				pCore->rcv_fis_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 256) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, size);
				pCore->RX_FIS_DMA = memDMA;
				pCore->RX_FIS = memVir;
				/*
				************************************************************************
				**		Deploy Delivery Queue DMA
				************************************************************************
				*/
				size = (sizeof(struct _Delivery_Q_ENTRY) * (deliverQueueSize + 1)) * ARCSAS_1330_DELIVERY_Q_NUM;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas post queue pool",  &(pACB->pcidev->dev), size, 64, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas post queue pool", pACB->pcidev, size, 64, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create post queue DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got post queue pool\n", __func__);
				#endif
				pCore->post_queue_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->post_queue_pool, GFP_ATOMIC, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->post_queue_pool, GFP_ATOMIC, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc post queue DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->post_queue_pool_dma = dma_addr;
				pCore->post_queue_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, size);
				size = sizeof(struct _Delivery_Q_ENTRY) * deliverQueueSize;
				for (i=0; i < ARCSAS_MAX_QUEUE_NUM; i++) {
					pCore->queues[i].delv_q = (delv_q_context *)memVir;
					pCore->queues[i].delv_q_dma = memDMA;
					memVir = (unsigned char *)memVir + size;
					memDMA = arcsas_u64_add_u32(memDMA, size);
				}
				size = sizeof(delv_q_context);
				for (i=0; i < ARCSAS_MAX_QUEUE_NUM; i++) {
					pCore->queues[i].delv_q_shadow = (delv_q_context *)memVir;
					pCore->queues[i].delv_q_shadow_dma = memDMA;
					memVir = (unsigned char *)memVir + size;
					memDMA = arcsas_u64_add_u32(memDMA, size);
				}
				/*
				************************************************************************
				**		Deploy Completion Queue DMA
				************************************************************************
				*/
				size = (sizeof(struct _Completion_Q_ENTRY) * completeQueueSize +
				    SHADOW_PTR_SIZE) * ARCSAS_1330_COMPLETION_Q_NUM;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas completion queue pool", &(pACB->pcidev->dev), size, 64, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas completion queue pool", pACB->pcidev, size, 64, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create completion queue DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got completion queue pool\n", __func__);
				#endif
				pCore->cmpl_queue_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->cmpl_queue_pool, GFP_ATOMIC, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->cmpl_queue_pool, GFP_ATOMIC, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc completion queue DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->cmpl_queue_pool_dma = dma_addr;
				pCore->cmpl_queue_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, size);
				size = sizeof(struct _Completion_Q_ENTRY) * completeQueueSize;
				for (i=0; i < ARCSAS_MAX_QUEUE_NUM; i++) {
					pCore->queues[i].cmpl_q = memVir;
					pCore->queues[i].cmpl_q_dma = memDMA;
					memVir = (unsigned char *)memVir + size;
					memDMA = arcsas_u64_add_u32(memDMA, size);
				}
				size = SHADOW_PTR_SIZE;
				for (i=0; i < ARCSAS_MAX_QUEUE_NUM; i++) {
					pCore->queues[i].cmpl_q_shadow = memVir;
					pCore->queues[i].cmpl_q_shadow_dma = memDMA;
					memVir = (unsigned char *)memVir + size;
					memDMA = arcsas_u64_add_u32(memDMA, size);
				}
				/*
				************************************************************************
				**		Deploy SAS Scratch Buffer DMA
				************************************************************************
				*/
				size = SAS_SCRATCH_BUFFER_SIZE * SASScratchCount;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas sas scratch buffer pool", &(pACB->pcidev->dev), size, 64, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas sas scratch buffer pool", pACB->pcidev, size, 64, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create sas scratch buffer DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got sas scratch buffer pool\n", __func__);
				#endif
				pCore->sas_scrtch_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->sas_scrtch_pool, GFP_ATOMIC, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->sas_scrtch_pool, GFP_ATOMIC, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc sas scratch buffer DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->sas_scrtch_pool_dma = dma_addr;
				pCore->sas_scrtch_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, size);
				for (i = 0; i < SASScratchCount; i++) {
					pSASSB = (PSAS_Scratch_Buffer)osm_list_first_entry(&pCore->SAS_Scratch_List,
					    struct _SAS_Scratch_Buffer, Queue_Pointer);
					pSASSB->Scratch_Vir = memVir;
					pSASSB->Scratch_DMA = memDMA;
					osm_list_move_tail(&pSASSB->Queue_Pointer, &pCore->SAS_Scratch_List);
					memVir = (unsigned char *)memVir + SAS_SCRATCH_BUFFER_SIZE;
					memDMA = arcsas_u64_add_u32(memDMA, SAS_SCRATCH_BUFFER_SIZE);
				}
				/*
				************************************************************************
				**		Deploy SG Table DMA and SG Table DMA Trash Bucket
				************************************************************************
				*/
				size = MAX_HBA_CMD_12G_PRD_BUFFER_SIZE * PRDBufferCount;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas sg table pool", &(pACB->pcidev->dev), size, 64, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas sg table pool", pACB->pcidev, size, 64, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create sg table DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got sg table pool\n", __func__);
				#endif
				pCore->sg_tbl_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->sg_tbl_pool, GFP_KERNEL, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->sg_tbl_pool, GFP_KERNEL, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc sg table DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->sg_tbl_pool_dma = dma_addr;
				pCore->sg_tbl_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, size);
				for (i = 0; i < PRDBufferCount; i++) {
					u32 j, chain_cnt;
					psgd_12g_prd pPRDEntry, pChainPRDEntry;
					ARCSAS_U64 address64;

					pPRD_Buffer = (PPRD_Buffer)osm_list_first_entry(&pCore->PRD_Buffer_List,
					    struct _PRD_Buffer, Queue_Pointer);
					pPRD_Buffer->PRD_VIRTUAL = memVir;
					pPRD_Buffer->PRD_PHYSICAL = memDMA;
					pPRDEntry = (psgd_12g_prd)memVir;
					for (j = 0; j < ARCSAS_MAX_SG_PER_CMD; j++) {
						pPRDEntry->sci.interface_select = 0xc0;
						pPRDEntry++;
					}
					if (ARCSAS_OS_SG_ENTRIES >= 255) {
						pPRDEntry = (psgd_12g_prd)memVir;
						/* point to first chain sg*/
						pChainPRDEntry = (psgd_12g_prd)NumToPtr(PtrToNum(pPRDEntry) + (254 * sizeof(sgd_12g_prd)));
						address64 = arcsas_u64_add_u32(memDMA, 255 * sizeof(sgd_12g_prd)); /*NEXT PRD Base Address*/
						/*
						**********************************
						**   address this prd chain sg
						**  255,382,509,636,763,890,1017
						**********************************
						*/
						for (chain_cnt = 255; chain_cnt <= ARCSAS_OS_SG_ENTRIES; chain_cnt += 127) {
							pChainPRDEntry->baseAddr.parts.low = ARCSAS_CPU_TO_LE32(address64.parts.low); /* PRD table DWord0: (31:0)NEXT PRD Base Address Lower 32 bits*/
							pChainPRDEntry->baseAddr.parts.high = ARCSAS_CPU_TO_LE32(address64.parts.high); /* PRD table DWord1: (31:0)NEXT PRD Base Address Upper 32 bits*/
							pChainPRDEntry->sci.sgl_chained = 1;
							pChainPRDEntry = (psgd_12g_prd)NumToPtr(PtrToNum(pChainPRDEntry) +
							    (127 * sizeof(sgd_12g_prd)));
							address64 = arcsas_u64_add_u32(address64, 127 * sizeof(sgd_12g_prd));
						}
					}
					osm_list_move_tail(&pPRD_Buffer->Queue_Pointer, &pCore->PRD_Buffer_List);
					memVir = (unsigned char *)memVir + MAX_HBA_CMD_12G_PRD_BUFFER_SIZE;
					memDMA = arcsas_u64_add_u32(memDMA, MAX_HBA_CMD_12G_PRD_BUFFER_SIZE);
				}
			}
			break;
		}
		case ARC_1320: {
			u32 size = 0; 
			unsigned short core_index;
			osm_dma_addr_t dma_addr;
			void *vaddr;
			struct dma_pool *dma_pool_temp;

			for (core_index = 0; core_index < pACB->core_num; core_index++) {
				pCore = pACB->pCore[core_index];
				/*
				************************************************************************
				**		Deploy Command Header DMA
				************************************************************************ 
				*/
				size = sizeof(struct _ARCSAS_Command_Header) * slotCount;	/* Command Header*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas command list pool", &(pACB->pcidev->dev), size, 64, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas command list pool", pACB->pcidev, size, 64, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create command list DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got command list pool\n", __func__);
				#endif
				pCore->cmnd_list_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->cmnd_list_pool, GFP_ATOMIC, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->cmnd_list_pool, GFP_ATOMIC, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc command list DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr)); 
				dmaResource.Byte_Size = size;
				pCore->cmnd_list_pool_dma = dma_addr;
				pCore->cmnd_list_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 64) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->CmdSlot_DMA = memDMA;
				pCore->CmdSlot = memVir;
				/*
				************************************************************************
				**		Deploy Received FIS DMA
				************************************************************************ 
				*/
				size = MAX_RX_FIS_POOL_SIZE_1320;/* Received FIS */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas received FIS pool", &(pACB->pcidev->dev), size, 256, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas received FIS pool", pACB->pcidev, size, 256, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create received FIS DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got received FIS pool\n", __func__);
				#endif
				pCore->rcv_fis_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->rcv_fis_pool, GFP_ATOMIC, &dma_addr); 
#else
				vaddr = pci_pool_alloc(pCore->rcv_fis_pool, GFP_ATOMIC, &dma_addr); 
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc command list DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->rcv_fis_pool_dma = dma_addr;
				pCore->rcv_fis_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 256) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->RX_FIS_DMA = memDMA;
				pCore->RX_FIS = memVir;
				/*
				************************************************************************
				**		Deploy Command Table DMA
				************************************************************************
				*/
				size = sizeof(struct _ARCSAS_Command_Table) * slotCount;	/* Command Table */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas command table pool", &(pACB->pcidev->dev), size, 128, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas command table pool", pACB->pcidev, size, 128, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create command table DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got command table pool\n", __func__);
				#endif
				pCore->cmnd_tbl_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->cmnd_tbl_pool, GFP_ATOMIC, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->cmnd_tbl_pool, GFP_ATOMIC, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc command table DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				pCore->cmnd_tbl_pool_dma = dma_addr;
				pCore->cmnd_tbl_pool_vaddr = vaddr;
				dmaResource.Byte_Size = size;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 128) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->Cmd_Table_DMA = memDMA;
				pCore->Cmd_Table = memVir;
				/*
				************************************************************************
				**		Init Command Header
				************************************************************************
				*/
				pCmdHeader = (PARCSAS_Command_Header)pCore->CmdSlot;
				cmdTblPhyAddr = pCore->Cmd_Table_DMA;
				for (i=0; i < slotCount; i++) {
					tmpPhyAddress = cmdTblPhyAddr;
					tmpPhyAddress.value += offsetof(struct _ARCSAS_Command_Table, table);
					pCmdHeader->Cmd_Table_Address = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.low);
					pCmdHeader->Cmd_Table_Address_High = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.high);
					tmpPhyAddress = cmdTblPhyAddr;
					tmpPhyAddress.value += offsetof(struct _ARCSAS_Command_Table, Open_Addr_Frame);
					pCmdHeader->OpenAddrFrame_Address = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.low);
					pCmdHeader->OpenAddrFrame_Address_High = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.high);
					tmpPhyAddress = cmdTblPhyAddr;
					tmpPhyAddress.value += offsetof(struct _ARCSAS_Command_Table, Status_Buff);
					pCmdHeader->StatusBuff_Address = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.low);
					pCmdHeader->StatusBuff_Address_High = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.high);
					pCmdHeader++;
					cmdTblPhyAddr.value += sizeof(struct _ARCSAS_Command_Table);
				}
				/*
				************************************************************************
				**		Deploy Post Queue DMA
				************************************************************************
				*/
				size = sizeof(struct _POST_Q_ENTRY) * deliverQueueSize;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas post queue pool", &(pACB->pcidev->dev), size, 8, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas post queue pool", pACB->pcidev, size, 8, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create post queue DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got post queue pool\n", __func__);
				#endif
				pCore->post_queue_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->post_queue_pool, GFP_ATOMIC, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->post_queue_pool, GFP_ATOMIC, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc post queue DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->post_queue_pool_dma = dma_addr;
				pCore->post_queue_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->PostQ_DMA = memDMA;
				pCore->PostQ = memVir;
				/*
				************************************************************************
				**		Deploy Completion Queue DMA
				************************************************************************
				*/
				size = sizeof(struct _DONE_Q_ENTRY) * completeQueueSize;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas completion queue pool", &(pACB->pcidev->dev), size, 8, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas completion queue pool", pACB->pcidev, size, 8, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create completion queue DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got completion queue pool\n", __func__);
				#endif
				pCore->cmpl_queue_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->cmpl_queue_pool, GFP_ATOMIC, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->cmpl_queue_pool, GFP_ATOMIC, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc completion queue DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->cmpl_queue_pool_dma = dma_addr;
				pCore->cmpl_queue_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->DoneQ_DMA = memDMA;
				pCore->DoneQ = memVir;
				/*
				************************************************************************
				**		Deploy SAS Scratch Buffer DMA
				************************************************************************
				*/
				size = SAS_SCRATCH_BUFFER_SIZE * SASScratchCount;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas sas scratch buffer pool", &(pACB->pcidev->dev), size, 8, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas sas scratch buffer pool", pACB->pcidev, size, 8, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create sas scratch buffer DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got sas scratch buffer pool\n", __func__);
				#endif
				pCore->sas_scrtch_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->sas_scrtch_pool, GFP_ATOMIC, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->sas_scrtch_pool, GFP_ATOMIC, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc sas scratch buffer DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->sas_scrtch_pool_dma = dma_addr;
				pCore->sas_scrtch_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				for (i = 0; i < SASScratchCount; i++) {
					pSASSB = (PSAS_Scratch_Buffer)osm_list_first_entry(&pCore->SAS_Scratch_List,
					    struct _SAS_Scratch_Buffer, Queue_Pointer);
					pSASSB->Scratch_Vir = memVir;
					pSASSB->Scratch_DMA = memDMA;
					osm_list_move_tail(&pSASSB->Queue_Pointer, &pCore->SAS_Scratch_List);
					memVir = (unsigned char *)memVir + SAS_SCRATCH_BUFFER_SIZE;
					memDMA = arcsas_u64_add_u32(memDMA, SAS_SCRATCH_BUFFER_SIZE);
				}
				/*
				************************************************************************
				**		Deploy SG Table DMA and SG Table DMA Trash Bucket
				************************************************************************
				*/
				size = MAX_HBA_CMD_6G_PRD_BUFFER_SIZE * PRDBufferCount;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				dma_pool_temp = dma_pool_create("arcsas sg table pool", &(pACB->pcidev->dev), size, 8, 0);
#else
				dma_pool_temp = pci_pool_create("arcsas sg table pool", pACB->pcidev, size, 8, 0);
#endif
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create sg table DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("%s:got sg table pool\n", __func__);
				#endif
				pCore->sg_tbl_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
				vaddr = dma_pool_alloc(pCore->sg_tbl_pool, GFP_KERNEL, &dma_addr);
#else
				vaddr = pci_pool_alloc(pCore->sg_tbl_pool, GFP_KERNEL, &dma_addr);
#endif
				if (vaddr == NULL) {
					osm_printk("Fail to alloc sg table DMA pool.\n");
					return false;
				}
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->sg_tbl_pool_dma = dma_addr;
				pCore->sg_tbl_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				for (i = 0; i < PRDBufferCount; i++) {
					pPRD_Buffer = (PPRD_Buffer)osm_list_first_entry(&pCore->PRD_Buffer_List,
					    struct _PRD_Buffer, Queue_Pointer);
					pPRD_Buffer->PRD_VIRTUAL = memVir;
					pPRD_Buffer->PRD_PHYSICAL = memDMA;
					if (ARCSAS_OS_SG_ENTRIES >= 255) {
						u32 chain_cnt;
						ARCSAS_U64 address64;
						psgd_6g_prd pPRDEntry, pChainPRDEntry;

						pPRDEntry = (psgd_6g_prd)memVir;
						/* point to first chain sg*/
						pChainPRDEntry = (psgd_6g_prd)NumToPtr(PtrToNum(pPRDEntry) + (254 * sizeof(sgd_6g_prd)));
						address64 = arcsas_u64_add_u32(memDMA, 255 * sizeof(sgd_6g_prd)); /*NEXT PRD Base Address*/
						/*
						**********************************
						**   address this prd chain sg
						**  255,382,509,636,763,890,1017
						**********************************
						*/
						for (chain_cnt = 255; chain_cnt <= ARCSAS_OS_SG_ENTRIES; chain_cnt += 127) {
							pChainPRDEntry->baseAddr.parts.low = ARCSAS_CPU_TO_LE32(address64.parts.low); /* PRD table DWord0: (31:0)NEXT PRD Base Address Lower 32 bits*/
							pChainPRDEntry->baseAddr.parts.high = ARCSAS_CPU_TO_LE32(address64.parts.high); /* PRD table DWord1: (31:0)NEXT PRD Base Address Upper 32 bits*/
							pChainPRDEntry->sci.sgl_chained = 1;
							pChainPRDEntry = (psgd_6g_prd)NumToPtr(PtrToNum(pChainPRDEntry) +
							    (127 * sizeof(sgd_6g_prd)));
							address64 = arcsas_u64_add_u32(address64, 127 * sizeof(sgd_6g_prd));
						}
					}
					memVir = (unsigned char *)memVir + MAX_HBA_CMD_6G_PRD_BUFFER_SIZE;
					memDMA = arcsas_u64_add_u32(memDMA, MAX_HBA_CMD_6G_PRD_BUFFER_SIZE);
					osm_list_move_tail(&pPRD_Buffer->Queue_Pointer, &pCore->PRD_Buffer_List);
				}
			}
			break;
		}
		case ARC_1300: {
			void *vaddr;
			osm_dma_addr_t dma_addr;
			struct dma_pool *dma_pool_temp;

			pCore = pACB->pCore[0];
			/*
			************************************************************************
			** Assign uncached memory for command list(64 byte align)
			************************************************************************
			*/
			size = sizeof(struct _ARCSAS_Command_Header) * slotCount;/* Command List*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			dma_pool_temp = dma_pool_create("arcsas command list pool", &(pACB->pcidev->dev), size, 64, 0);
#else
			dma_pool_temp = pci_pool_create("arcsas command list pool", pACB->pcidev, size, 64, 0);
#endif
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create command list DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:got command list pool\n", __func__);
			#endif
			pCore->cmnd_list_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			vaddr = dma_pool_alloc(pCore->cmnd_list_pool, GFP_ATOMIC, &dma_addr);
#else
			vaddr = pci_pool_alloc(pCore->cmnd_list_pool, GFP_ATOMIC, &dma_addr);
#endif
			if (vaddr == NULL) {
				osm_printk("Fail to alloc command list DMA pool.\n");
				return false;
			}
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr)); 
			dmaResource.Byte_Size = size;
			pCore->cmnd_list_pool_dma = dma_addr;
			pCore->cmnd_list_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 64) - memDMA.value);
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:command list pool offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			osm_memset(memVir, 0, dmaResource.Byte_Size);
			pCore->CmdSlot_DMA = memDMA;
			pCore->CmdSlot = memVir;
			/*
			************************************************************************
			** Assign uncached memory for received FIS(256 byte align)
			************************************************************************
			*/
			size = MAX_RX_FIS_POOL_SIZE_1300;/* Received FIS */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			dma_pool_temp = dma_pool_create("arcsas received FIS pool", &(pACB->pcidev->dev), size, 256, 0);
#else
			dma_pool_temp = pci_pool_create("arcsas received FIS pool", pACB->pcidev, size, 256, 0);
#endif
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create received FIS DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:got received FIS pool\n", __func__);
			#endif
			pCore->rcv_fis_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			vaddr = dma_pool_alloc(pCore->rcv_fis_pool, GFP_ATOMIC, &dma_addr); 
#else
			vaddr = pci_pool_alloc(pCore->rcv_fis_pool, GFP_ATOMIC, &dma_addr); 
#endif
			if (vaddr == NULL) {
				osm_printk("Fail to alloc received FIS DMA pool.\n");
				return false;
			}
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->rcv_fis_pool_dma = dma_addr;
			pCore->rcv_fis_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 256) - memDMA.value);
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:received FIS pool offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			osm_memset(memVir, 0, dmaResource.Byte_Size);
			pCore->RX_FIS_DMA = memDMA;
			pCore->RX_FIS = memVir;
			/*
			************************************************************************
			** Assign the 32 command tables.(128 byte align)
			************************************************************************
			*/
			size = sizeof(struct _ARCSAS_Command_Table) * slotCount;	/* Command Table */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			dma_pool_temp = dma_pool_create("arcsas command table pool", &(pACB->pcidev->dev), size, 128, 0);
#else
			dma_pool_temp = pci_pool_create("arcsas command table pool", pACB->pcidev, size, 128, 0);
#endif
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create command table DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:got command table pool\n", __func__);
			#endif
			pCore->cmnd_tbl_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			vaddr = dma_pool_alloc(pCore->cmnd_tbl_pool, GFP_ATOMIC, &dma_addr);
#else
			vaddr = pci_pool_alloc(pCore->cmnd_tbl_pool, GFP_ATOMIC, &dma_addr);
#endif
			if (vaddr == NULL) {
				osm_printk("Fail to alloc command table DMA pool.\n");
				return false;
			}
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			pCore->cmnd_tbl_pool_dma = dma_addr;
			pCore->cmnd_tbl_pool_vaddr = vaddr;
			dmaResource.Byte_Size = size;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 128) - memDMA.value);
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:command table offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			pCore->Cmd_Table = memVir;
			pCore->Cmd_Table_DMA = memDMA;
			/*
			************************************************************************
			**		Init Command Header
			************************************************************************
			*/
			pCmdHeader = (PARCSAS_Command_Header)pCore->CmdSlot;
			cmdTblPhyAddr = pCore->Cmd_Table_DMA;
			for (i=0; i < slotCount; i++) {
				tmpPhyAddress = cmdTblPhyAddr;
				tmpPhyAddress.value += offsetof(struct _ARCSAS_Command_Table, table);
				pCmdHeader->Cmd_Table_Address = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.low);
				pCmdHeader->Cmd_Table_Address_High = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.high);
				tmpPhyAddress = cmdTblPhyAddr;
				tmpPhyAddress.value += offsetof(struct _ARCSAS_Command_Table, Open_Addr_Frame);
				pCmdHeader->OpenAddrFrame_Address = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.low);
				pCmdHeader->OpenAddrFrame_Address_High = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.high);
				tmpPhyAddress = cmdTblPhyAddr;
				tmpPhyAddress.value += offsetof(struct _ARCSAS_Command_Table, Status_Buff);
				pCmdHeader->StatusBuff_Address = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.low);
				pCmdHeader->StatusBuff_Address_High = ARCSAS_CPU_TO_LE32(tmpPhyAddress.parts.high);
				pCmdHeader++;
				cmdTblPhyAddr.value += sizeof(struct _ARCSAS_Command_Table);
			}
			/*
			************************************************************************
			** Assign delivery queue.
			************************************************************************
			*/
			size = sizeof(struct _POST_Q_ENTRY) * deliverQueueSize;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			dma_pool_temp = dma_pool_create("arcsas post queue pool", &(pACB->pcidev->dev), size, 8, 0);
#else
			dma_pool_temp = pci_pool_create("arcsas post queue pool", pACB->pcidev, size, 8, 0);
#endif
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create post queue DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:got post queue pool\n", __func__);
			#endif
			pCore->post_queue_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			vaddr = dma_pool_alloc(pCore->post_queue_pool, GFP_ATOMIC, &dma_addr);
#else
			vaddr = pci_pool_alloc(pCore->post_queue_pool, GFP_ATOMIC, &dma_addr);
#endif
			if (vaddr == NULL) {
				osm_printk("Fail to alloc post queue DMA pool.\n");
				return false;
			}
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->post_queue_pool_dma = dma_addr;
			pCore->post_queue_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s: post queue offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			pCore->PostQ = memVir;/* odin delivery window of scsi,sata,sas... command */
			pCore->PostQ_DMA = memDMA;
			/*
			************************************************************************
			** Assign completion queue.
			************************************************************************
			*/
			size = sizeof(struct _DONE_Q_ENTRY) * completeQueueSize;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			dma_pool_temp = dma_pool_create("arcsas completion queue pool", &(pACB->pcidev->dev), size, 8, 0);
#else
			dma_pool_temp = pci_pool_create("arcsas completion queue pool", pACB->pcidev, size, 8, 0);
#endif
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create completion queue DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:got completion queue pool\n", __func__);
			#endif
			pCore->cmpl_queue_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			vaddr = dma_pool_alloc(pCore->cmpl_queue_pool, GFP_ATOMIC, &dma_addr);
#else
			vaddr = pci_pool_alloc(pCore->cmpl_queue_pool, GFP_ATOMIC, &dma_addr);
#endif
			if (vaddr == NULL) {
				osm_printk("Fail to alloc completion queue DMA pool.\n");
				return false;
			}
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->cmpl_queue_pool_dma = dma_addr;
			pCore->cmpl_queue_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:completion queue offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			pCore->DoneQ_DMA = memDMA;
			pCore->DoneQ = memVir; 
			/*
			************************************************************************
			** Assign the scratch buffer(8 byte align)
			** Number of Taret/Device can not be determined until 
			************************************************************************
			*/
			size = SAS_SCRATCH_BUFFER_SIZE * SASScratchCount;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			dma_pool_temp = dma_pool_create("arcsas sas scratch buffer pool", &(pACB->pcidev->dev), size, 8, 0);
#else
			dma_pool_temp = pci_pool_create("arcsas sas scratch buffer pool", pACB->pcidev, size, 8, 0);
#endif
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create sas scratch buffer DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:got sas scratch buffer pool\n", __func__);
			#endif
			pCore->sas_scrtch_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			vaddr = dma_pool_alloc(pCore->sas_scrtch_pool, GFP_ATOMIC, &dma_addr);
#else
			vaddr = pci_pool_alloc(pCore->sas_scrtch_pool, GFP_ATOMIC, &dma_addr);
#endif
			if (vaddr == NULL) {
				osm_printk("Fail to alloc sas scratch buffer DMA pool.\n");
				return false;
			}
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->sas_scrtch_pool_dma = dma_addr;
			pCore->sas_scrtch_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:sas scratch buffer offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			osm_memset(memVir, 0, dmaResource.Byte_Size);
			for (i = 0; i < SASScratchCount; i++) {
				pSASSB = (PSAS_Scratch_Buffer)osm_list_first_entry(&pCore->SAS_Scratch_List,
				    struct _SAS_Scratch_Buffer, Queue_Pointer);
				pSASSB->Scratch_Vir = memVir;
				pSASSB->Scratch_DMA = memDMA;
				osm_list_move_tail(&pSASSB->Queue_Pointer, &pCore->SAS_Scratch_List);
				memVir = (unsigned char *)memVir + SAS_SCRATCH_BUFFER_SIZE;
				memDMA = arcsas_u64_add_u32(memDMA, SAS_SCRATCH_BUFFER_SIZE);
			}
			/*
			************************************************************************
			** Assign the PRD buffer 
			************************************************************************
			*/
			size = MAX_HBA_CMD_3G_PRD_BUFFER_SIZE * PRDBufferCount;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			dma_pool_temp = dma_pool_create("arcsas sg table pool", &(pACB->pcidev->dev), size, 8, 0);
#else
			dma_pool_temp = pci_pool_create("arcsas sg table pool", pACB->pcidev, size, 8, 0);
#endif
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create sg table DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:got sg table pool\n", __func__);
			#endif
			pCore->sg_tbl_pool = dma_pool_temp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
			vaddr = dma_pool_alloc(pCore->sg_tbl_pool, GFP_KERNEL, &dma_addr);
#else
			vaddr = pci_pool_alloc(pCore->sg_tbl_pool, GFP_KERNEL, &dma_addr);
#endif
			if (vaddr == NULL) {
				osm_printk("Fail to alloc sg table DMA pool.\n");
				return false;
			}
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->sg_tbl_pool_dma = dma_addr;
			pCore->sg_tbl_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
			#if ARCSAS_DBG_INIT_CONTEXT
				osm_printk("%s:sg table offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			osm_memset(memVir, 0, dmaResource.Byte_Size);
			for (i = 0; i < PRDBufferCount; i++) {
				pPRD_Buffer = (PPRD_Buffer)osm_list_first_entry(&pCore->PRD_Buffer_List,
				    struct _PRD_Buffer, Queue_Pointer);
				pPRD_Buffer->PRD_VIRTUAL = memVir;
				pPRD_Buffer->PRD_PHYSICAL = memDMA;
				osm_list_move_tail(&pPRD_Buffer->Queue_Pointer, &pCore->PRD_Buffer_List);
				memDMA = arcsas_u64_add_u32(memDMA, MAX_HBA_CMD_3G_PRD_BUFFER_SIZE);
				memVir = (unsigned char *)memVir + MAX_HBA_CMD_3G_PRD_BUFFER_SIZE; 
			}
			break;
		}
		default: {
			_ASSERT();
		}
	}
	return true;
}
/*
************************************************************************
************************************************************************
*/
Bool arcsas_core_initialize(struct AdapterControlBlock *pACB)
{
	unsigned char devCount, expanderCount, pmCount, EventElementCount;
	unsigned short slotCount, i, deliverQueueSize, completeQueueSize, core_index;
	unsigned short CoreCCBCount;
	unsigned short SASScratchCount, SMPScratchCount;
	unsigned short contextCount, PRDBufferCount;
	u32 discovery_buffer_size;
	u32 sgSize = 0;
	u32 size, usedSize;
	unsigned int sgEntryCount;
	PCore_Extension pCore = NULL;
	PCore_Context pContext;
	PSAS_Scratch_Buffer pSASSB;
	PSMP_Scratch_Buffer pSMPSB;
	PDomain_Port port;
	PDomain_Phy phy;
	PPRD_Buffer pPRD_Buffer;
	void *pResource, *temp, *tmpSG;
	struct CommandControlBlock *pCCB;

	#if ARCSAS_DBG_INIT
		osm_printk("%s:\n", __func__);
	#endif
	size = arcsas_core_get_cached_memory_quota(pACB);
	pResource = osm_vmalloc(size + 4096);
	if (pResource == NULL) {
		osm_printk("Areca sas controller core initialize, allocate core resource error.\n");
		return false;
	}
	pACB->pCacheMem = pResource;
	osm_memset(pResource, 0, size);
	devCount = expanderCount = pmCount = CoreCCBCount = sgEntryCount = SASScratchCount = SMPScratchCount = 0;
	contextCount = PRDBufferCount = slotCount = deliverQueueSize = completeQueueSize = EventElementCount = 0;
	#if ARCSAS_DBG_INFO
		osm_printk("%s: ARCSAS_MIN_DISK_ID = %d\n", __func__, ARCSAS_MIN_DISK_ID);
		osm_printk("%s: ARCSAS_MAX_DISK_ID = %d\n", __func__, ARCSAS_MAX_DISK_ID);
		osm_printk("%s: ARCSAS_MIN_DEVICE_ID = %d\n", __func__, ARCSAS_MIN_DEVICE_ID);
		osm_printk("%s: ARCSAS_MAX_DEVICE_ID = %d\n", __func__, ARCSAS_MAX_DEVICE_ID);
		osm_printk("%s: ARCSAS_MIN_ENCLOSURE_ID = %d\n", __func__, ARCSAS_MIN_ENCLOSURE_ID);
		osm_printk("%s: ARCSAS_MAX_ENCLOSURE_ID = %d\n", __func__, ARCSAS_MAX_ENCLOSURE_ID);
		osm_printk("%s: ARCSAS_MIN_EXPANDER_ID = %d\n", __func__, ARCSAS_MIN_EXPANDER_ID);
		osm_printk("%s: ARCSAS_MAX_EXPANDER_ID = %d\n", __func__, ARCSAS_MAX_EXPANDER_ID);
		osm_printk("%s: ARCSAS_MIN_PM_ID = %d\n", __func__, ARCSAS_MIN_PM_ID);
		osm_printk("%s: ARCSAS_MAX_PM_ID = %d\n", __func__, ARCSAS_MAX_PM_ID);
		osm_printk("%s: ARCSAS_MAX_DEVICE_NUMBER = %d\n", __func__, ARCSAS_MAX_DEVICE_NUMBER);
	#endif
	arcsas_get_core_supported_counts(&EventElementCount,
					&devCount, 
					&expanderCount,
					&pmCount,
					&CoreCCBCount,
					&sgEntryCount,
					&PRDBufferCount,
					&SASScratchCount,
					&SMPScratchCount,
					&slotCount,
					&contextCount,
					&deliverQueueSize,
					&completeQueueSize);
	pACB->Expander_Count_Supported = expanderCount;
	pACB->Device_Count_Supported = devCount;
	pACB->PortMutiplier_Count_Supported = pmCount;
	pACB->DevicesLun = (PDomain_Device)pResource;	
	temp = pResource + osm_roundup(sizeof(Domain_Device) * (ARCSAS_LUN_SUPPORTED + 1), 8);
	pACB->Devices = (PDomain_Device)temp;
	#if ARCSAS_DBG_INIT_CONTEXT
		osm_printk("%s:pACB->Devices = 0x%p\n", __func__, pACB->Devices);
	#endif
	temp += osm_roundup(sizeof(Domain_Device) * devCount, 8);
	pACB->Expanders = (PDomain_Expander)temp;
	temp += osm_roundup(sizeof(Domain_Expander) * expanderCount, 8);
	pACB->PMs = (PDomain_PM)temp;
	temp += osm_roundup(sizeof(Domain_PM) * pmCount, 8);
	pACB->EventElement = (PEVENT_ELEMENT)temp;
   	 temp += osm_roundup(sizeof(struct _EVENT_ELEMENT) * EventElementCount, 8);
	pACB->Device_Tag_Pool.Stack = (unsigned short *)temp;
	pACB->Device_Tag_Pool.Size = devCount;
	temp += sizeof(struct _Tag_Stack) * devCount;
	arcsas_init_tag(&pACB->Device_Tag_Pool, pACB->Device_Tag_Pool.Size);
	pACB->Expander_Tag_Pool.Stack = (unsigned short *)temp;
	pACB->Expander_Tag_Pool.Size = expanderCount;
	temp += sizeof(struct _Tag_Stack) * expanderCount;
	arcsas_init_tag(&pACB->Expander_Tag_Pool, pACB->Expander_Tag_Pool.Size);
	pACB->PM_Tag_Pool.Stack = (unsigned short *)temp;
	pACB->PM_Tag_Pool.Size = pmCount;
	temp += sizeof(struct _Tag_Stack) * pmCount;
	arcsas_init_tag(&pACB->PM_Tag_Pool, pACB->PM_Tag_Pool.Size);
	for (i = 0; i <= ARCSAS_LUN_SUPPORTED; i++) {
		pACB->DevicesLun[i].Status = DEVICE_STATUS_NO_DEVICE;
		//osm_atomic_set(&pACB->DevicesLun[i].Timer_Status, STOP_RUNNING);
		pACB->DevicesLun[i].register_set = ID_NOT_MAPPED;
		pACB->DevicesLun[i].EntryIndex = ID_NOT_MAPPED;/*this value is (target_lun,target_id) */
		pACB->DevicesLun[i].Timeout_Count = 0;
		pACB->DevicesLun[i].Retry_Count = 0;
		pACB->DevicesLun[i].saslun[1] = 0x00;
	}
	for (i = 0; i <= pACB->Device_Count_Supported; i++) {/*0...136=136*/
		pACB->Devices[i].Status = DEVICE_STATUS_NO_DEVICE;
		//osm_atomic_set(&pACB->Devices[i].Timer_Status, STOP_RUNNING);
		pACB->Devices[i].EntryIndex = ID_NOT_MAPPED;
		pACB->Devices[i].Timeout_Count = 0;
		pACB->Devices[i].register_set = ID_NOT_MAPPED;
		pACB->Devices[i].Retry_Count = 0;
		pACB->Devices[i].saslun[1] = 0x00;
	}
	pACB->last_device_mapped_id = 0;/*(0..135)*/
	for (i = 0; i <= ARCSAS_MAX_PM_ID; i++) {/*(0..135,136..143,144..151)*/
		pACB->DevicesExpandersPMs_EntryIndex[i] = ID_NOT_MAPPED;
	}
	for (i = 0; i < expanderCount; i++) {
		osm_atomic_set(&pACB->Expanders[i].Device_Number, 0);
		pACB->Expanders[i].Phy_Count = 0;
		pACB->Expanders[i].pParent = NULL;
		pACB->Expanders[i].pPort = (void *)NULL;
		OSM_INIT_LIST_HEAD(&pACB->Expanders[i].Exp_Queue_Pointer);
		OSM_INIT_LIST_HEAD(&pACB->Expanders[i].Device_List);
	}
	pACB->last_expander_mapped_id = ARCSAS_MIN_EXPANDER_ID;/*136....143*/
	pACB->last_portmutiplier_mapped_id = ARCSAS_MIN_PM_ID;/*144....151*/
//	max_core_num = (pACB->acb_flag & ACB_FLAG_DUAL_CORE) ? 2 : 1;
	for (core_index = 0; core_index < pACB->core_num; core_index++) {
		pCore = pACB->pCore[core_index] = (PCore_Extension)temp;
		temp = NumToPtr(PtrToNum(pCore) + osm_roundup(sizeof(Core_Extension), 8));
		pCore->pACB = pACB;
		pCore->core_index = core_index;
		//pCore->Timer_Status = STOP_RUNNING;
		//OSM_INIT_LIST_HEAD(&pCore->Wait_RegisterSet_CCB_List);
		OSM_INIT_LIST_HEAD(&pCore->CCB_Waiting_List);
		OSM_INIT_LIST_HEAD(&pCore->CCB_Complete_List);
		OSM_INIT_LIST_HEAD(&pCore->Context_List);
		OSM_INIT_LIST_HEAD(&pCore->SAS_Scratch_List);
		OSM_INIT_LIST_HEAD(&pCore->SMP_Scratch_List);
		OSM_INIT_LIST_HEAD(&pCore->PRD_Buffer_List);
		OSM_INIT_LIST_HEAD(&pCore->Core_CCB_List);
		arcsas_spin_lock_init(&pCore->ccb_cmpl_list_lock);
		arcsas_spin_lock_init(&pCore->ccb_waiting_lst_lock);
		arcsas_spin_lock_init(&pCore->ccb_list_lock);
		arcsas_spin_lock_init(&pCore->context_list_lock);
		arcsas_spin_lock_init(&pCore->prd_list_lock);
		arcsas_spin_lock_init(&pCore->sassb_list_lock);
		arcsas_spin_lock_init(&pCore->smpsb_list_lock);
		arcsas_spin_lock_init(&pCore->doneq_lock);
		arcsas_spin_lock_init(&pCore->CmdSlot_Tag_Pool.lock);
		arcsas_spin_lock_init(&pCore->Port_Tag_Pool.lock);
		arcsas_spin_lock_init(&pCore->send_cmd_lock);
		if (pACB->adapter_type != ARC_1330)
			tasklet_init(&pCore->isr_tasklet, arcsas_interrupt_service_bh, (unsigned long)pCore);
		pCore->Slot_Count_Supported = slotCount;	/*512*/
		pCore->Max_PostQ_Index = deliverQueueSize;
		/* 
		************************************************************************
		** For completion queue,
		** Minus 1 because the first entry in completion queue is the pointer not real entry
		************************************************************************
		*/
		pCore->Max_DoneQ_Index = completeQueueSize - 1;	//522
		switch (pACB->adapter_type) {
			case ARC_1330: {
				pCore->core_flag |= (core_index ? ARCSAS_CORE_FLAG_HBA_CORE_ID1 : ARCSAS_CORE_FLAG_HBA_CORE_ID0);
				pCore->Phy_Num = ARCSAS_1330_PHY_NUM;
				pCore->MaxRegisterSet = ARCSAS_1330_REGSET_NUM;
				pCore->last_port_index = 0;
				pCore->MaxCmdSlotWidth = 13;
				pCore->queue_num = ARCSAS_MAX_QUEUE_NUM;
				for (i = 0; i < pCore->queue_num; i++) {
					OSM_INIT_LIST_HEAD(&pCore->queues[i].waiting_queue);
					OSM_INIT_LIST_HEAD(&pCore->queues[i].high_priority_queue);
					OSM_INIT_LIST_HEAD(&pCore->queues[i].complete_queue);
					arcsas_spin_lock_init(&pCore->queues[i].waiting_queue_lock);
					arcsas_spin_lock_init(&pCore->queues[i].high_priority_queue_lock);
					arcsas_spin_lock_init(&pCore->queues[i].complete_queue_lock);
					pCore->queues[i].pCore = pCore;
					pCore->queues[i].msix_idx = core_index * pCore->queue_num + i;
					pCore->queues[i].id = i;
					pCore->queues[i].delv_q_size = deliverQueueSize;
					pCore->queues[i].cmpl_q_size = completeQueueSize;
					pCore->queues[i].last_delv_q = 0x3fff;
					pCore->queues[i].last_cmpl_q = 0x3fff;
				}
				pCore->handlers[HANDLER_SATA].send_command = arcsas_sata_send_command;
			//	pCore->handlers[HANDLER_SATA].process_command = arcsas_sata_process_command;
			//	pCore->handlers[HANDLER_SATA].error_handler = arcsas_sata_error_handler;
				pCore->handlers[HANDLER_PM].send_command = arcsas_pm_send_command;
			//	pCore->handlers[HANDLER_PM].process_command = arcsas_pm_process_command;
			//	pCore->handlers[HANDLER_PM].error_handler = NULL;
				pCore->handlers[HANDLER_SSP].send_command = arcsas_ssp_send_command;
			//	pCore->handlers[HANDLER_SSP].process_command = arcsas_ssp_process_command;
			//	pCore->handlers[HANDLER_SSP].error_handler = arcsas_ssp_error_handler;
				pCore->handlers[HANDLER_SMP].send_command = arcsas_smp_send_command;
			//	pCore->handlers[HANDLER_SMP].process_command = arcsas_smp_process_command;
			//	pCore->handlers[HANDLER_SMP].error_handler = NULL;
				pCore->handlers[HANDLER_STP].send_command = arcsas_sata_send_command;
			//	pCore->handlers[HANDLER_STP].process_command = arcsas_stp_process_command;
			//	pCore->handlers[HANDLER_STP].error_handler = arcsas_sata_error_handler;
				pCore->handlers[HANDLER_ENC].send_command = arcsas_ssp_send_command;
			//	pCore->handlers[HANDLER_ENC].process_command = arcsas_ssp_process_command;
			//	pCore->handlers[HANDLER_ENC].error_handler = NULL;
#if	SUPPORT_PHY_POWER_MODE
				pCore->PHY_power_mode_DIPM = 0;
				pCore->PHY_power_mode_HIPM = 0;
				pCore->PHY_power_mode_HIPM_timer = 10;
				pCore->PHY_power_mode_port_map = 0xffff;
				pCore->PHY_power_mode_timer_index = 0xffff;
#endif
				break;
			}
			case ARC_1320: {
				pCore->core_flag |= (core_index ? ARCSAS_CORE_FLAG_HBA_CORE_ID1 : ARCSAS_CORE_FLAG_HBA_CORE_ID0);
				pCore->Phy_Num = ARCSAS_1320_PHY_NUM;
				pCore->MaxRegisterSet = ARCSAS_1320_REGSET_NUM;
				pCore->last_port_index = 0;
				pCore->MaxCmdSlotWidth = 9;
				break;
			}
			case ARC_1300: {
				pCore->core_flag |= ARCSAS_CORE_FLAG_HBA_CORE_ID0;
				pCore->Phy_Num = ARCSAS_1300_PHY_NUM;
				pCore->MaxRegisterSet = ARCSAS_1300_REGSET_NUM;
				pCore->last_port_index = 0;
				pCore->MaxCmdSlotWidth = 9;
				break;
			}
			default: {
				_ASSERT();
			}
		}
		pCore->free_ccb_count = CoreCCBCount;
	#if	ARCSAS_DBG_DISCOVERY_CONTEXT
		osm_printk("arcmsr: original has free CCB count = 0x%x\n", pCore->free_ccb_count);
	#endif
		pCore->Running_CCB = (PCCB *)temp;
		temp += osm_roundup(sizeof(struct CommandControlBlock *) * slotCount, 8);
		sgSize = sizeof(struct _ARCSAS_SGENTRY) * sgEntryCount;
		tmpSG = temp;
		temp = temp + sgSize * CoreCCBCount;
		for (i = 0; i < CoreCCBCount; i++) {/*CoreCCBCount=128*/
			pCCB = (struct CommandControlBlock *)temp;
	#if	ARCSAS_DBG_DISCOVERY_CONTEXT
		osm_printk("arcmsr: pCCB = 0x%p\n", temp);
	#endif
			pCCB->SG_Table.pSGEntry = (PARCSAS_SGENTRY)tmpSG;
			osm_list_add_tail(&pCCB->Queue_Pointer, &pCore->Core_CCB_List);	/*link all ccb pointer*/
			tmpSG += sgSize;
			temp += sizeof(struct CommandControlBlock);			/*struct _CCB is 64byte aligned.*/
		}
		temp = NumToPtr(osm_roundup(PtrToNum(temp), 8));
		for (i = 0; i < contextCount; i++) {
			pContext = (PCore_Context)temp;
			pContext->Context_Type = CORE_CONTEXT_TYPE_NONE;
			OSM_INIT_LIST_HEAD(&pContext->Queue_Pointer);
			osm_list_add_tail(&pContext->Queue_Pointer, &pCore->Context_List);
			temp += sizeof(struct _CORE_CONTEXT);
		}
		temp = NumToPtr(osm_roundup(PtrToNum(temp), 8));
		for (i = 0; i < SASScratchCount; i++) {
			pSASSB = (PSAS_Scratch_Buffer)temp;
			OSM_INIT_LIST_HEAD(&pSASSB->Queue_Pointer);
			osm_list_add_tail(&pSASSB->Queue_Pointer, &pCore->SAS_Scratch_List);
			temp += sizeof(struct _SAS_Scratch_Buffer);
		}
		temp = NumToPtr(osm_roundup(PtrToNum(temp), 8));
		for (i = 0; i < SMPScratchCount; i++) {
			pSMPSB = (PSMP_Scratch_Buffer)temp;
			OSM_INIT_LIST_HEAD(&pSMPSB->Queue_Pointer);
			osm_list_add_tail(&pSMPSB->Queue_Pointer, &pCore->SMP_Scratch_List);
			temp += sizeof(struct _SMP_Scratch_Buffer);
		}
		temp = NumToPtr(osm_roundup(PtrToNum(temp), 8));
		for (i = 0; i < PRDBufferCount; i++) {
			pPRD_Buffer = (PPRD_Buffer)temp;
			OSM_INIT_LIST_HEAD(&pPRD_Buffer->Queue_Pointer);
			osm_list_add_tail(&pPRD_Buffer->Queue_Pointer, &pCore->PRD_Buffer_List);
			temp += sizeof(struct _PRD_Buffer);
		}
		pCore->CmdSlot_Tag_Pool.Stack = (unsigned short *)temp;
		pCore->CmdSlot_Tag_Pool.Size = slotCount;
		temp += sizeof(struct _Tag_Stack) * slotCount;
		arcsas_init_cmdslot(&pCore->CmdSlot_Tag_Pool, pCore->CmdSlot_Tag_Pool.Size);
		pCore->Port_Tag_Pool.Stack = (unsigned short *)temp;
		pCore->Port_Tag_Pool.Size = pCore->Phy_Num;
		temp += sizeof(struct _Tag_Stack) * pCore->Phy_Num;
		arcsas_init_tag(&pCore->Port_Tag_Pool, pCore->Phy_Num);
		for (i = 0; i < pCore->Phy_Num; i++) {
			port = &pCore->Ports[i];
			port->Port_Index = (unsigned char) i;
			port->Port_State = PORT_STATE_IDLE;
			osm_atomic_set(&port->DiscoveryInstance, 0);
			port->pCore = pCore;
			osm_atomic_set(&port->Device_Number, 0);
			osm_atomic_set(&port->Expander_Number, 0);
			OSM_INIT_LIST_HEAD(&port->Device_List);
			OSM_INIT_LIST_HEAD(&port->Expander_List);
			/* PORT TYPE will be determined after arcsas_init_chip. */
			port->Type = PORT_TYPE_SAS; /*default*/
			phy = &pCore->Phy[i];
			phy->id = (u8)i;
			if (pACB->adapter_type == ARC_1330)
				phy->asic_id = log_2_asic[i];
			else
				phy->asic_id = (u8)i;
			phy->sata_ssu_timer = NO_CURRENT_TIMER;
		}
		pCore->last_port_index = 0;
		for (i = 0; i < pCore->Phy_Num; i++) {
			pCore->Ports_EntryIndex[i] = ID_NOT_MAPPED;
		}
		for (i = 0; i < 16; i++) {
			pCore->Resetting_Slot[i] = 0;
		}
		pCore->last_PostQ_index = 0xffff;
		pCore->last_DoneQ_index = 0xffff;
	}
	usedSize = (u32)(PtrToNum(temp) - PtrToNum(pResource));
//	discovery_buffer_size = (pACB->acb_flag & ACB_FLAG_DUAL_CORE) ? ((size - usedSize) >> 1) : (size - usedSize);
	discovery_buffer_size = sizeof(Discover_Extension) + sizeof(Discover_Target) * expanderCount;
	pACB->pCore[0]->pDiscoverBuffer[0] = temp;
	arcsas_discovery_set_resource(pACB->pCore[0]->pDiscoverBuffer[0], discovery_buffer_size, expanderCount);
	temp += discovery_buffer_size;
	pACB->pCore[0]->pDiscoverBuffer[1] = temp;
	arcsas_discovery_set_resource(pACB->pCore[0]->pDiscoverBuffer[1], discovery_buffer_size, expanderCount);
	temp += discovery_buffer_size;
	if (pACB->acb_flag & ACB_FLAG_DUAL_CORE) {
		pACB->pCore[1]->pDiscoverBuffer[0] = temp;
		arcsas_discovery_set_resource(pACB->pCore[1]->pDiscoverBuffer[0], discovery_buffer_size, expanderCount);
		temp += discovery_buffer_size;
		pACB->pCore[1]->pDiscoverBuffer[1] = temp;
		arcsas_discovery_set_resource(pACB->pCore[1]->pDiscoverBuffer[1], discovery_buffer_size, expanderCount);
	}
	if (!arcsas_core_get_DMA_memory(pACB, PRDBufferCount, slotCount, deliverQueueSize, completeQueueSize, SASScratchCount)) {
		arcsas_teardown_dma_pool(pACB);
		return false;
	}
	return true;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_pre_init_pm(PDomain_PM pPM)
{
	#if ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PM_SM
		osm_printk("%s:\n", __func__);
	#endif
	osm_memset((unsigned char *)pPM, 0, sizeof(Domain_PM));
	//pPM->Timer_Status = STOP_RUNNING;
	arcsas_setup_timer(&pPM->timer, arcsas_pm_ccb_timeout, (unsigned long)pPM);
	arcsas_setup_timer(&pPM->timer1, arcsas_timer_wait_pm_phy_ready, (unsigned long)pPM);
	pPM->register_set = ID_NOT_MAPPED;
	OSM_INIT_LIST_HEAD(&pPM->Sent_CCB_List);
	arcsas_spin_lock_init(&pPM->sent_ccb_list_lock);
}
/*
************************************************************************
************************************************************************
*/
static void arcsas_pre_init_device(PDomain_Device pDevice, PACB pACB)
{
	struct dv_event_work	*dv_event;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pDevice->Status = DEVICE_STATUS_NO_DEVICE;
	//arcsas_setup_timer(&pDevice->timer, arcsas_os_ccb_timeout, (unsigned long)pDevice);
	pDevice->Queue_Depth = 32;//Originally this is 255
	pDevice->register_set = ID_NOT_MAPPED;
	pDevice->Timeout_Count = 0;
	pDevice->Retry_Count = 0;
	OSM_INIT_LIST_HEAD(&pDevice->Port_Queue_Pointer);
	OSM_INIT_LIST_HEAD(&pDevice->Exp_Queue_Pointer);
	OSM_INIT_LIST_HEAD(&pDevice->Sent_CCB_List);
	OSM_INIT_LIST_HEAD(&pDevice->CCB_Waiting_List);
	arcsas_spin_lock_init(&pDevice->sent_ccb_list_lock);
	arcsas_spin_lock_init(&pDevice->ncq_lock);
//#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 19, 0)
	dv_event = arcsas_get_dv_event_work(pACB);
	dv_event->pACB = pACB;
	dv_event->pDevice = pDevice;
	pDevice->dv_event = dv_event;
	dv_event = arcsas_get_dv_event_work(pACB);
	dv_event->pACB = pACB;
	dv_event->pDevice = pDevice;
	pDevice->dv_event1 = dv_event;
//#endif
#if 0
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
		OSM_INIT_WORK(&pDevice->arcsas_add_scsi_device, arcsas_add_scsi_device_wqfn);
		OSM_INIT_WORK(&pDevice->arcsas_remove_scsi_device, arcsas_remove_scsi_device_wqfn);
	#else
		OSM_INIT_WORK(&pDevice->arcsas_add_scsi_device, arcsas_add_scsi_device_wqfn, pDevice);
		OSM_INIT_WORK(&pDevice->arcsas_remove_scsi_device, arcsas_remove_scsi_device_wqfn, pDevice);
	#endif
#endif
	pDevice->psdev = NULL;
	osm_atomic_set(&pDevice->Outstanding_CCB, 0);
	osm_atomic_set(&pDevice->TaskFile_Error_Count, 0);
	osm_atomic_set(&pDevice->Timer_Status, 0);
	#if ARCSAS_DBG_CONTEXT
		osm_printk("%s: pDevice = 0x%p, &pDevice->arcsas_add_scsi_device = 0x%p, &pDevice->arcsas_remove_scsi_device = 0x%p\n",
		    __func__, pDevice, &pDevice->arcsas_add_scsi_device, &pDevice->arcsas_remove_scsi_device);
	#endif
}
/*
************************************************************************
************************************************************************
*/
PDomain_Device arcsas_get_device_lun_from_pool(PCore_Extension pCore, unsigned short EntryIndex)
{
	unsigned char i = 0;
	struct AdapterControlBlock *pACB;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pACB = pCore->pACB;
	for (i = 0; i <= ARCSAS_LUN_SUPPORTED; i++) {
		if (pACB->DevicesLun[i].saslun[1] == 0x00 && (pACB->DevicesLun[i].EntryIndex == ID_NOT_MAPPED)) {
			/* saslun=0 imply that this Device_Lun is free */
			break;
		}
	}
	if (i > ARCSAS_LUN_SUPPORTED) {
		#if ARCSAS_DBG_CONTEXT
			osm_printk("Ran out of Lun device IDs!\n");
		#endif
		return NULL;
	}
	pACB->DevicesLun[i].pCore = pCore;
	arcsas_pre_init_device(&pACB->DevicesLun[i], pACB);
	pACB->DevicesLun[i].EntryIndex = EntryIndex;/*this value is (target_lun, target_id) */
	/* Set some default value. */
	pACB->DevicesLun[i].MDMA_Mode = 0xFF;
	pACB->DevicesLun[i].PIO_Mode = 0xFF;
	pACB->DevicesLun[i].UDMA_Mode = 0xFF;
	pACB->DevicesLun[i].Current_MDMA = 0xFF;
	pACB->DevicesLun[i].Current_PIO = 0xFF;
	pACB->DevicesLun[i].Current_UDMA = 0xFF;
	pACB->DevicesLun[i].NegotiatedLinkRate = 0;
	pACB->DevicesLun[i].queue_id = 0;
	pACB->DevicesLun[i].queue = &pCore->queues[0];
	osm_atomic_inc(&pACB->all_device_count);
	return &pACB->DevicesLun[i];
}
/*
************************************************************************
************************************************************************
*/
PDomain_Device arcsas_get_device_from_pool(PCore_Extension pCore, u8 phyid)
{
	unsigned char i = 0;
	unsigned short index;
	struct AdapterControlBlock *pACB;
	PDomain_Device pDevice;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pACB = pCore->pACB;
	if (arcsas_tag_is_empty(&pACB->Device_Tag_Pool)) {
		osm_printk("Device pool empty!\n");
		return NULL;
	}
	if (phyid != ID_NOT_MAPPED) {
		i = phyid;
		if (pCore->core_flag & ARCSAS_CORE_FLAG_HBA_CORE_ID1)
			i += pCore->Phy_Num;
		if (pACB->DevicesExpandersPMs_EntryIndex[i] != ID_NOT_MAPPED)
			goto gd1;
		if (pACB->last_device_mapped_id < i)
			pACB->last_device_mapped_id = i;
		goto gd0;
	}
gd1:
	for (i = ARCSAS_MIN_DEVICE_ID; i <= ARCSAS_MAX_DEVICE_ID; i++) {/*0...135*/
		if (pACB->DevicesExpandersPMs_EntryIndex[i] == ID_NOT_MAPPED) {
			/* update the last ID into last_device_mapped_id */
			if (pACB->last_device_mapped_id < i) {
				pACB->last_device_mapped_id = i;/*0...135*/
			}
			break;
		}
	}
	if (i > ARCSAS_MAX_DEVICE_ID) {
		osm_printk("Ran out of device IDs!\n");
		return NULL;
	}
gd0:
	index = arcsas_get_pool_tag(&pACB->Device_Tag_Pool); /* Tag: dep_entry_index of DevicesExpandersPMs_EntryIndex[] */
	pACB->DevicesExpandersPMs_EntryIndex[i] = (unsigned char)index; /* 0....135 */
	pDevice = &pACB->Devices[index];
	pDevice->pCore = pCore;
	arcsas_pre_init_device(pDevice, pACB);
	pDevice->EntryIndex = i;	/* 0...135 */
	pDevice->State = DEVICE_STATE_IDLE;
	/* Set some default value. */
	pDevice->MDMA_Mode = 0xFF;
	pDevice->PIO_Mode = 0xFF;
	pDevice->UDMA_Mode = 0xFF;
	pDevice->Current_MDMA = 0xFF;
	pDevice->Current_PIO = 0xFF;
	pDevice->Current_UDMA = 0xFF;
	pDevice->NegotiatedLinkRate = 0;
	if (pACB->adapter_type == ARC_1330) {
	//	pDevice->queue_id = 0;
	//	pDevice->queue = &pCore->queues[0];
		pDevice->queue_id = i % pCore->queue_num;
		pDevice->queue = &pCore->queues[pDevice->queue_id];
	}
	osm_atomic_inc(&pACB->all_device_count);
	#if ARCSAS_DBG_CONTEXT || ARCSAS_DBG_PLUG_UNPLUG || ARCSAS_DBG_SAS_SM_CONTEXT || ARCSAS_DBG_SATA_SM_CONTEXT
		osm_printk("%s: pDevice = 0x%p, pDevice->EntryIndex = 0x%x\n", __func__, &pACB->Devices[index],
		    pACB->Devices[index].EntryIndex);
	#endif
	return pDevice;
}
/*
************************************************************************
************************************************************************
*/
static Bool arcsas_generate_os_ccb_sg_table(struct AdapterControlBlock *pACB, struct scsi_cmnd *pscsi_cmnd, struct CommandControlBlock *pCCB)
{
	unsigned int i;
	int sg_count;
	struct scatterlist *sg;
	PARCSAS_SG_Table pSGTable = &pCCB->SG_Table;

	#if ARCSAS_DBG_OS_CCB
		osm_printk("%s:\n", __func__);
	#endif
	if (arcsas_rq_bflen(pscsi_cmnd) > ARCSAS_MAX_CMDXFER_LENGTH) {
		osm_printk("arcsas: request buffer length is beyong the range.\n");
		return false;
	}
	if (arcsas_use_sg(pscsi_cmnd)) {
		#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: use sg\n", __func__);
		#endif
		sg_count = arcsas_map_sg(pscsi_cmnd);
		#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: sg_count = %d\n", __func__, sg_count);
		#endif
		if ((unlikely(sg_count > ARCSAS_OS_SG_ENTRIES)) || (unlikely(sg_count < 0))) {
			osm_printk("arcsas: sg dma mapping failed.\n");
			return false;
		}
		arcsas_sg_table_init(pSGTable, sg_count);
		scsi_for_each_sg(pscsi_cmnd, sg, sg_count, i) {
			arcsas_sg_table_append(pSGTable, cpu_to_le32(dma_addr_lo32(sg_dma_address(sg))),
			    cpu_to_le32(dma_addr_hi32(sg_dma_address(sg))), cpu_to_le32(sg_dma_len(sg)));
		}
	} else if (arcsas_rq_bflen(pscsi_cmnd)) {
		dma_addr_t dma_addr;
		#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: Not use sg\n", __func__);
		#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
		dma_addr = dma_map_single(&(pACB->pcidev->dev), arcsas_rq_bf(pscsi_cmnd), arcsas_rq_bflen(pscsi_cmnd),
		    pscsi_cmnd->sc_data_direction);
#else
		dma_addr = pci_map_single(pACB->pcidev, arcsas_rq_bf(pscsi_cmnd), arcsas_rq_bflen(pscsi_cmnd),
		    pscsi_cmnd->sc_data_direction);
#endif
		/* We hide it here for later unmap.*/
		arcsas_sg_table_init(pSGTable, 1);
	//	pscsi_cmnd->SCp.dma_handle = dma_addr;
		pCCB->dma_handle = dma_addr;
		arcsas_sg_table_append(pSGTable, cpu_to_le32(dma_addr_lo32(dma_addr)),
		    cpu_to_le32(dma_addr_hi32(dma_addr)), cpu_to_le32(arcsas_rq_bflen(pscsi_cmnd)));
	}
	return true;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_done_os_ccb_callback(void *this, struct CommandControlBlock *pCCB)
{
	/* Return this request to OS. */
	PDomain_Device pDevice;
	PACB pACB = (PACB)this;
	#if ARCSAS_DBG_OS_CCB
		osm_printk("%s: pCCB = 0x%p, cdb[0] = 0x%x\n", __func__, pCCB, pCCB->Cdb[0]);
	#endif
	pDevice = pCCB->pDevice;
	arcsas_free_ncq_tag(pDevice, pCCB);
	arcsas_os_ccb_done(pACB, pCCB);
}
/*
************************************************************************
************************************************************************
*/
unsigned long arcsas_build_os_ccb(struct AdapterControlBlock *pACB, struct scsi_cmnd *pscsi_cmnd, PDomain_Device *ppDevice)
{
	u64 timeout = 20;
	PCore_Extension pCore;
	struct CommandControlBlock *pOSCCB = NULL;
	PDomain_Port pPort;
	unsigned short entry_index;
	PDomain_Device pDevice;
	/*
	*************************************************
	** Set three flags:	CMD_FLAG_NON_DATA
	**		CMD_FLAG_DATA_IN
	**		CMD_FLAG_DMA
	** Currently data in/out all go thru DMA
	*************************************************
	*/
	#if ARCSAS_DBG_OS_CCB
		osm_printk("%s:\n", __func__);
	#endif
	if (pscsi_cmnd->device->lun) {
		unsigned char i;
		*ppDevice = NULL;
		entry_index = MAKE_DEVICE_ENTRY_INDEX(pscsi_cmnd->device->id, pscsi_cmnd->device->lun);
		for (i = 0; i <= ARCSAS_LUN_SUPPORTED; i++) {
			if (pACB->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
				*ppDevice = &pACB->DevicesLun[i];
				break;
			}
		}
	} else {
		if (pscsi_cmnd->device->id <= ARCSAS_MAX_DEVICE_ID) {
			*ppDevice = ((entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, pscsi_cmnd->device->id)) ==
			    ID_NOT_MAPPED) ? NULL : &pACB->Devices[entry_index];
			#if ARCSAS_DBG_OS_CCB_CONTEXT
				osm_printk("%s: pDevice = 0x%p, entry_index = 0x%x\n", __func__, *ppDevice, entry_index);
			#endif
		} else {
			*ppDevice = NULL;/*hot unplug, device is disconnected*/
			#if ARCSAS_DBG_OS_CCB_CONTEXT
				osm_printk("%s: pDevice = NULL\n", __func__);
			#endif
		}
	}
	pDevice = *ppDevice;
	if ((pDevice == NULL) || ((pDevice)->Status == DEVICE_STATUS_NO_DEVICE)) {
		return 0;
	}
	pCore = pDevice->pCore;
	pPort = pDevice->pPort;
	if (pCore->free_ccb_count < 4)
		return 0;
	if (pDevice->State != DEVICE_STATE_INIT_DONE) {
		#if ARCSAS_DBG_OS_CCB
			osm_printk("%s: Device is 0x%x 0x%x\n", __func__, pDevice->State, pDevice->Status);
			osm_printk("%s: scmnd(0x%p) id:lun = %x:%x\n", __func__, pscsi_cmnd, pscsi_cmnd->device->id,
			    (u32)pscsi_cmnd->device->lun);
		#endif
		return SCSI_MLQUEUE_DEVICE_BUSY;
	}
	if ((pDevice->Connection & DC_ATAPI) && (pDevice->queued_cmd_request > ARCSAS_MAX_CMD_ATAPI))
		return SCSI_MLQUEUE_DEVICE_BUSY;
	if ((pPort->Setting & PORT_SETTING_DURING_RECOVERY) || osm_atomic_read(&pPort->DiscoveryInstance)) {
		unsigned long rtn;
#if	ARCSAS_DBG_PLUG_UNPLUG
		if (printk_ratelimit()) {
			osm_printk("%s: Port is under recovery.(0x%x 0x%x)\n", __func__, pPort->Setting,
			    osm_atomic_read(&pPort->DiscoveryInstance));
			osm_printk("%s: scsicmd(0x%p, 0x%p, )[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x]\n",
					__func__,
					pscsi_cmnd,
				//	pscsi_cmnd->serial_number,
					pscsi_cmnd->request,
				//	pscsi_cmnd->request->start_time,
					pscsi_cmnd->cmnd[0],
					pscsi_cmnd->cmnd[1],
					pscsi_cmnd->cmnd[2],
					pscsi_cmnd->cmnd[3],
					pscsi_cmnd->cmnd[4],
					pscsi_cmnd->cmnd[5],
					pscsi_cmnd->cmnd[6],
					pscsi_cmnd->cmnd[7],
					pscsi_cmnd->cmnd[8],
					pscsi_cmnd->cmnd[9]
			);
		}
#endif
		//mod_timer(&pPort->DiscoveryPolling_timer, jiffies + 10 * HZ);
		rtn = SCSI_MLQUEUE_HOST_BUSY;
		return rtn;
	}
	pOSCCB = arcsas_core_get_coreccb_from_list(pCore);
	if (pOSCCB == NULL) {
		return 0;
	}
	if (arcsas_generate_os_ccb_sg_table(pACB, pscsi_cmnd, pOSCCB)) { /*fill pCCB->SG_Table*/
		osm_memcpy(pOSCCB->Cdb, pscsi_cmnd->cmnd, pscsi_cmnd->cmd_len);
		pOSCCB->Cdb_len = pscsi_cmnd->cmd_len;
		pOSCCB->EntryIndex = MAKE_DEVICE_ENTRY_INDEX(pscsi_cmnd->device->id, pscsi_cmnd->device->lun);
		pOSCCB->pDevice = pDevice;
		pOSCCB->PRD_Buffer = NULL;
		pOSCCB->Data_Buffer = NULL;
		pOSCCB->Data_Transfer_Length = cpu_to_le32((arcsas_rq_bflen(pscsi_cmnd)));
		pOSCCB->Sense_Buffer = pscsi_cmnd->sense_buffer;
		pOSCCB->Sense_Info_Buffer_Length = SCSI_SENSE_BUFFERSIZE;
		pOSCCB->Time_Out = timeout; // timeout in second
		pOSCCB->Org_Req = pscsi_cmnd;
	//	pOSCCB->pscsi_cmnd = pscsi_cmnd;
	//	pOSCCB->Tag = pscsi_cmnd->tag;
		pOSCCB->CCB_Status = CCB_STATUS_PENDING;
		pOSCCB->Completion = (void(*)(void *, PCCB))arcsas_done_os_ccb_callback;
		pOSCCB->Cmd_Initiator = (void *)pACB;
		pOSCCB->TargetId = (u16)pscsi_cmnd->device->id;
		pOSCCB->Lun = (u32)pscsi_cmnd->device->lun;
		switch (pscsi_cmnd->sc_data_direction) {
			case DMA_NONE:
				break;
			case DMA_FROM_DEVICE:
				pOSCCB->Cmd_Flag |= CMD_FLAG_DATA_IN;
				break;
			case DMA_TO_DEVICE:
				pOSCCB->Cmd_Flag |= CMD_FLAG_DMA;
				break;
			case DMA_BIDIRECTIONAL:
				osm_printk("arcsas : unexpected DMA_BIDIRECTIONAL.\n");
				break;
			default:
				break;
		}
	} else {
		unsigned long flags;
		arcsas_spin_lock_irqsave(&pCore->ccb_list_lock, &flags);
		osm_list_add_tail(&pOSCCB->Queue_Pointer, &pCore->Core_CCB_List);
		if (pOSCCB->CoreContext) {
			arcsas_free_core_context_to_pool(pCore, &pOSCCB->CoreContext);
		}
		arcsas_spin_unlock_irqrestore(&pCore->ccb_list_lock, &flags);
		//Although it fails to generate os ccb sg table here, it will return the ccb address to tell
		//arcsas_queuecommand that the ccb pool is not empty.
	}
	return (unsigned long)pOSCCB;
}
/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 21)
static int arcsas_queuecommand_lck(struct scsi_cmnd *pscsi_cmnd)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
static int arcsas_queuecommand_lck(struct scsi_cmnd *pscsi_cmnd, void (*done)(struct scsi_cmnd *))
#endif
{
	unsigned long pccb;
	struct Scsi_Host *host;
	struct AdapterControlBlock *pACB;
	struct pci_dev *pcidev;
	PDomain_Device pDevice = NULL;

	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB)
		if (!SCSI_IS_READ(pscsi_cmnd->cmnd[0]) && !SCSI_IS_WRITE(pscsi_cmnd->cmnd[0]))
			osm_printk("%s: id:lun = %d:%d, scmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x], buffer_len(0x%x), start_time = , request = 0x%p\n", __func__,
			pscsi_cmnd->device->id,
			(u16)pscsi_cmnd->device->lun,
			pscsi_cmnd,
			pscsi_cmnd->cmnd[0],
			pscsi_cmnd->cmnd[1],
			pscsi_cmnd->cmnd[2],
			pscsi_cmnd->cmnd[3],
			pscsi_cmnd->cmnd[4],
			pscsi_cmnd->cmnd[5],
			pscsi_cmnd->cmnd[6],
			pscsi_cmnd->cmnd[7],
			pscsi_cmnd->cmnd[8],
			pscsi_cmnd->cmnd[9],
			arcsas_rq_bflen(pscsi_cmnd),
		//	pscsi_cmnd->request->start_time,
			pscsi_cmnd->request
		);
	#endif
	host = pscsi_cmnd->device->host;
	pACB = (struct AdapterControlBlock *)host->hostdata;
	if (pACB->acb_flag & ACB_FLAG_ADAPTER_REMOVED) {
		pscsi_cmnd->result = (DID_NO_CONNECT << 16);
#if	LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 21)
		pscsi_cmnd->scsi_done(pscsi_cmnd);
#else
		scsi_done(pscsi_cmnd);
#endif
		return 0;
	}
#if	LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 21)
	pscsi_cmnd->scsi_done = done;
#endif
	pscsi_cmnd->result = 0;
	pcidev = pACB->pcidev;
	if (unlikely(osm_atomic_read(&pACB->outstandingccbcount) > host->can_queue)) {
		if (printk_ratelimit())
			osm_printk("queuecommand: SCSI_MLQUEUE_HOST_BUSY\n ");
		return SCSI_MLQUEUE_HOST_BUSY;
	}
	if (unlikely(pscsi_cmnd->device->channel)) {
		pscsi_cmnd->result = DID_BAD_TARGET << 16;
#if	LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 21)
		pscsi_cmnd->scsi_done(pscsi_cmnd);
#else
		scsi_done(pscsi_cmnd);
#endif
		return 0;
	}
	if (pACB->acb_flag & ACB_FLAG_SYNC_DEVICE_ID) {
		if (printk_ratelimit())
			osm_printk("queuecommand: Clearing flash device map!!\n ");
		return SCSI_MLQUEUE_HOST_BUSY;
	}
	if (pscsi_cmnd->device->id == ARCSAS_VIRTUAL_DEVICE_ID) {
		arcsas_handle_virtual_ses_device(pACB, pscsi_cmnd);
		return 0;
	}
	#if ARCSAS_DBG_OS_CCB_CONTEXT
		osm_printk("%s: calling arcsas_build_os_ccb\n", __func__);
	#endif
	pccb = arcsas_build_os_ccb(pACB, pscsi_cmnd, &pDevice);
	if (pccb) {
		if ((pccb == SCSI_MLQUEUE_HOST_BUSY) || (pccb == SCSI_MLQUEUE_DEVICE_BUSY)) {
			#if ARCSAS_DBG_OS_CCB_CONTEXT
				osm_printk("%s: No.1\n", __func__);
			#endif
			return pccb;
		} else {
		#if ARCSAS_DBG_CDB
			arcsas_print_cdb((struct CommandControlBlock *)pccb);
		#endif
			arcsas_post_ccb(((struct CommandControlBlock *)pccb)->pDevice->pCore, (struct CommandControlBlock *)pccb);
			pDevice->queued_cmd_request++;
		}
		return 0;
	} else {
		if ((pDevice == NULL) || (pDevice->Status == DEVICE_STATUS_NO_DEVICE)) {
			pscsi_cmnd->result = DID_NO_CONNECT << 16;//This is important. Driver should tell scsi-layer actively
#if	LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 21)
			pscsi_cmnd->scsi_done(pscsi_cmnd);
#else
			scsi_done(pscsi_cmnd);
#endif
			#if ARCSAS_DBG_OS_CCB_CONTEXT
				osm_printk("%s: No.3\n", __func__);
			#endif
			return 0;
		} else {
			#if ARCSAS_DBG_OS_CCB
				osm_printk("%s: No.4\n", __func__);
			#endif
			return SCSI_MLQUEUE_HOST_BUSY;
		}
	}
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
DEF_SCSI_QCMD(arcsas_queuecommand)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 37)
static int arcsas_queuecommand(struct scsi_cmnd *pscsi_cmnd, void (*done)(struct scsi_cmnd *))
{
	unsigned long pccb;
	struct Scsi_Host *host;
	struct AdapterControlBlock *pACB;
	struct pci_dev *pcidev;
	PDomain_Device pDevice = NULL;

	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB)
		if (!SCSI_IS_READ(pscsi_cmnd->cmnd[0]) && !SCSI_IS_WRITE(pscsi_cmnd->cmnd[0]))
			osm_printk("%s: scmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x], buffer_len(0x%x), id:lun = %d:%d, start_time = , request = 0x%p\n", __func__,
			pscsi_cmnd,
			pscsi_cmnd->cmnd[0],
			pscsi_cmnd->cmnd[1],
			pscsi_cmnd->cmnd[2],
			pscsi_cmnd->cmnd[3],
			pscsi_cmnd->cmnd[4],
			pscsi_cmnd->cmnd[5],
			pscsi_cmnd->cmnd[6],
			pscsi_cmnd->cmnd[7],
			pscsi_cmnd->cmnd[8],
			pscsi_cmnd->cmnd[9],
			arcsas_rq_bflen(pscsi_cmnd),
			pscsi_cmnd->device->id,
			(u16)pscsi_cmnd->device->lun,
		//	pscsi_cmnd->request->start_time,
			pscsi_cmnd->request
		);
	#endif
	host = pscsi_cmnd->device->host;
	pACB = (struct AdapterControlBlock *)host->hostdata;
	if (pACB->acb_flag & ACB_FLAG_ADAPTER_REMOVED) {
		pscsi_cmnd->result = (DID_NO_CONNECT << 16);
#if	LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 21)
		pscsi_cmnd->scsi_done(pscsi_cmnd);
#else
		scsi_done(pscsi_cmnd);
#endif
		return 0;
	}
#if	LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 21)
	pscsi_cmnd->scsi_done = done;
#endif
	pscsi_cmnd->result = 0;
	pcidev = pACB->pcidev;
	if (unlikely(osm_atomic_read(&pACB->outstandingccbcount) > host->can_queue)) {
		if (printk_ratelimit())
			osm_printk("queuecommand: SCSI_MLQUEUE_HOST_BUSY\n ");
		return SCSI_MLQUEUE_HOST_BUSY;
	}
	if (unlikely(pscsi_cmnd->device->channel)) {
		pscsi_cmnd->result = DID_BAD_TARGET << 16;
#if	LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 21)
		pscsi_cmnd->scsi_done(pscsi_cmnd);
#else
		scsi_done(pscsi_cmnd);
#endif
		return 0;
	}
	if (pACB->acb_flag & ACB_FLAG_SYNC_DEVICE_ID) {
		return SCSI_MLQUEUE_HOST_BUSY;
	}
	if (pscsi_cmnd->device->id == ARCSAS_VIRTUAL_DEVICE_ID) {
		arcsas_handle_virtual_ses_device(pACB, pscsi_cmnd);
		return 0;
	}
	pccb = arcsas_build_os_ccb(pACB, pscsi_cmnd, &pDevice);
	if (pccb) {
		if ((pccb == SCSI_MLQUEUE_HOST_BUSY) || (pccb == SCSI_MLQUEUE_DEVICE_BUSY)) {
			#if ARCSAS_DBG_OS_CCB
				osm_printk("%s: No.1 \n", __func__);
			#endif
			return pccb;
		} else {
			#if ARCSAS_DBG_OS_CCB_CONTEXT
				osm_printk("%s: No.2\n", __func__);
			#endif
			#if ARCSAS_DBG_CDB
			arcsas_print_cdb((struct CommandControlBlock *)pccb);
			#endif
			arcsas_post_ccb(((struct CommandControlBlock *)pccb)->pDevice->pCore, (struct CommandControlBlock *)pccb);
			pDevice->queued_cmd_request++;
		}
		return 0;
	} else {
		if ((pDevice == NULL) || (pDevice->Status == DEVICE_STATUS_NO_DEVICE)) {
			#if ARCSAS_DBG_OS_CCB
				osm_printk("%s:No.3 scmnd(0x%p) id:lun = %d:%d DID_NO_CONNECT(0x%p)\n", __func__,
				    pscsi_cmnd, pscsi_cmnd->device->id, (u16)pscsi_cmnd->device->lun, pDevice);
			#endif
			pscsi_cmnd->result = DID_NO_CONNECT << 16;//This is important. Driver should tell scsi-layer actively
#if	LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 21)
			pscsi_cmnd->scsi_done(pscsi_cmnd);
#else
			scsi_done(pscsi_cmnd);
#endif
			return 0;
		} else {
			#if ARCSAS_DBG_OS_CCB
				osm_printk("%s:No.4 scmnd(0x%p) id:lun = %d:%d SCSI_MLQUEUE_HOST_BUSY\n", __func__,
				    pscsi_cmnd, pscsi_cmnd->device->id, (u16)pscsi_cmnd->device->lun);
			#endif
			return SCSI_MLQUEUE_HOST_BUSY;
		}
	}
}
#endif
/*
************************************************************************
************************************************************************
*/
static void arcsas_shutdown(struct pci_dev *pcidev)
{
	struct Scsi_Host *host;
	struct AdapterControlBlock *pACB;
#if ARCSAS_DBG_UPDATE_FLASH_DEVICE_MAP
	int i;
#endif

	#if ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PCI_REMOVE
		osm_printk("%s:\n", __func__);
	#endif
	host = pci_get_drvdata(pcidev);
	pACB = (struct AdapterControlBlock *) host->hostdata;
	if (pACB->acb_flag & ACB_FLAG_ADAPTER_REMOVED)
		return;
	scsi_remove_host(host);
	arcsas_free_irq(pcidev, pACB);
	arcsas_check_update_flash_device_map(pACB, false);/* rom region 0x30000....0x30400, 8 bytes pattern for each device 8*128=1024 */
#if ARCSAS_DBG_UPDATE_FLASH_DEVICE_MAP
	for (i = 0; i < ARCSAS_DISK_SUPPORTED; i++) {
		osm_printk("pACB->Flash_Devices_Map[%d].Pattern.value = 0x%llx \n", i, pACB->Flash_Devices_Map[i].Pattern.value);
	}
#endif
	arcsas_disable_interrupt(pACB);
	arcsas_teardown_dma_pool(pACB);
	osm_vfree(pACB->pCacheMem);
	destroy_workqueue(pACB->scsi_lookup);
	flush_scheduled_work();
	kfree(pACB->dv_event);
	arcsas_iounmap(pACB);
	pci_release_regions(pcidev);
	pci_set_drvdata(pcidev, NULL);
	scsi_host_put(host);
	pci_disable_device(pcidev);
}

static void arcsas_iounmap(struct AdapterControlBlock *pACB)
{
	switch (pACB->adapter_type) {
		case ARC_1330:
		case ARC_1320: {
			iounmap(pACB->mem_map_base_ext);
			break;
		}
		case ARC_1300: {
			iounmap(pACB->pCore[0]->mem_map_base);
			break;
		}
		default: {
			_ASSERT();
			break;
		}
	}
}
/*
************************************************************************
************************************************************************
*/
static void arcsas_remove_devices(struct AdapterControlBlock *pACB)
{
	int  i;
	uint id, lun;
	struct scsi_device *psdev;

	for (i = ARCSAS_MIN_DEVICE_ID; i < ARCSAS_MAX_DEVICE_ID; i++) {/*(0..135,136..143,144..151)*/
		if (pACB->DevicesExpandersPMs_EntryIndex[i] == ID_NOT_MAPPED)
			continue;
		psdev = scsi_device_lookup(pACB->host, 0, i, 0);
		if (psdev != NULL) {
			scsi_remove_device(psdev);
			scsi_device_put(psdev);
		}
	}
	for (i = 0; i <= ARCSAS_LUN_SUPPORTED; i++) {
		if (pACB->DevicesLun[i].EntryIndex == ID_NOT_MAPPED)
			continue;
		id = DEVICE_ENTRY_INDEX_TO_TARGET_ID(pACB->DevicesLun[i].EntryIndex);
		lun = DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pACB->DevicesLun[i].EntryIndex);
		psdev = scsi_device_lookup(pACB->host, 0, id, lun);
		if (psdev != NULL) {
			scsi_remove_device(psdev);
			scsi_device_put(psdev);
		}
	}
}

static void arcsas_remove(struct pci_dev *pcidev)
{
	struct Scsi_Host *host;
	PDomain_Device pDevice;
	PDomain_PM pPM;
	struct AdapterControlBlock *pACB;
	unsigned short index;
	unsigned char i;
	u16 dev_id;


#if ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PCI_REMOVE
	osm_printk("%s:\n", __func__);
#endif
	host = pci_get_drvdata(pcidev);
	pACB = (struct AdapterControlBlock *)host->hostdata;
	pci_read_config_word(pcidev, PCI_DEVICE_ID, &dev_id);
	if (dev_id == 0xffff) {
		pACB->acb_flag |= ACB_FLAG_ADAPTER_REMOVED;
		arcsas_remove_devices(pACB);
	}
	for (i = ARCSAS_MIN_DEVICE_ID; i < ARCSAS_MAX_DEVICE_ID; i++) {/*(0..135,136..143,144..151)*/
		if (pACB->DevicesExpandersPMs_EntryIndex[i] == ID_NOT_MAPPED)
			continue;
		pDevice = &pACB->Devices[Get_DevicesExpandersPMs_EntryIndex(pACB, i)];/*(0..135)*/
		arcsas_delete_device_timer(pDevice);
	}
	for (i = ARCSAS_MIN_PM_ID; i <= ARCSAS_MAX_PM_ID; i++) {
		if (pACB->DevicesExpandersPMs_EntryIndex[i] == ID_NOT_MAPPED)
			continue;
		index = pACB->DevicesExpandersPMs_EntryIndex[i];
		pPM = &pACB->PMs[index];
		arcsas_delete_pm_timers(pPM);
	}
	scsi_remove_host(host);
	arcsas_free_irq(pcidev, pACB);
	if (pACB->eh_handler)
		kthread_stop(pACB->eh_handler);
	arcsas_disable_interrupt(pACB);
	destroy_workqueue(pACB->scsi_lookup);
	flush_scheduled_work();
	kfree(pACB->dv_event);
	arcsas_teardown_dma_pool(pACB);
	osm_vfree(pACB->DevicesLun);
	arcsas_iounmap(pACB);
	pci_release_regions(pcidev);
	pci_set_drvdata(pcidev, NULL);
	scsi_host_put(host);
	pci_disable_device(pcidev);
	arcsas_show_time();
	return;
}
/*
************************************************************************
************************************************************************
*/
static int arcsas_bios_param(struct scsi_device *sdev,
		struct block_device *bdev, sector_t capacity, int *geom)
{
	int heads, sectors, cylinders;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 7, 0)
	if (scsi_partsize(bdev, capacity, geom))
		return 0;
#endif
	heads = 255;
	sectors = 56;
	cylinders = sector_div(capacity, heads * sectors);
	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;
	return 0;
}
/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth)
{
	PDomain_Device pDevice = NULL;
	int res = 0;
	unsigned short entry_index;
	struct Scsi_Host *host;
	struct AdapterControlBlock *pACB;

	host = sdev->host;
	pACB = (struct AdapterControlBlock *)host->hostdata;
	if (sdev->lun) {
		unsigned char i;
		entry_index = MAKE_DEVICE_ENTRY_INDEX(sdev->id, sdev->lun);
		for (i = 0; i <= ARCSAS_LUN_SUPPORTED; i++) {
			if (pACB->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
				pDevice = &pACB->DevicesLun[i];
				break;
			}
		}
	} else {
		entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, sdev->id);
		if (entry_index == ID_NOT_MAPPED) {
			pDevice = NULL;/*hot unplug, device is disconnected*/
		} else {
			pDevice = &pACB->Devices[entry_index];
		}
	}
	if (pDevice == NULL) {
		osm_printk("Fail to change queue_depth of scsi id:lun(%d:%d)\n", sdev->id, (u32)sdev->lun);
		return -EOPNOTSUPP;
	}
	if (IS_SSP(pDevice)) {
		res = min(queue_depth, SAS_MAX_QD);
		if (sdev->tagged_supported)
			scsi_change_queue_depth(sdev, res);
		else {
			scsi_change_queue_depth(sdev, 1);
			res = 1;
		}
	} else if (IS_STP_OR_SATA_ATAPI(pDevice)) {
		res = min(queue_depth, ARCSAS_MAX_CMD_PERDEV);
		if (sdev->tagged_supported)
			scsi_change_queue_depth(sdev, res);
		else {
			scsi_change_queue_depth(sdev, 1);
			res = 1;
		}			
	}
	osm_printk("queue_depth of scsi id:lun(%d:%d) has changed to %d\n", sdev->id, (u32)sdev->lun, res);
	return res;
}
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth, int reason)
{
	unsigned short entry_index;
	int res = 0;
	PDomain_Device pDevice = NULL;
	struct Scsi_Host *host;
	struct AdapterControlBlock *pACB;

	if (reason != SCSI_QDEPTH_DEFAULT)
		return -EOPNOTSUPP;
	host = sdev->host;
	pACB = (struct AdapterControlBlock *)host->hostdata;
	if (sdev->lun) {
		unsigned char i;
		entry_index = MAKE_DEVICE_ENTRY_INDEX(sdev->id, sdev->lun);
		for (i = 0; i <= ARCSAS_LUN_SUPPORTED; i++) {
			if (pACB->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
				pDevice = &pACB->DevicesLun[i];
				break;
			}
		}
	} else {
		entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, sdev->id);
		if (entry_index == ID_NOT_MAPPED) {
			pDevice = NULL;/*hot unplug, device is disconnected*/
		} else {
			pDevice = &pACB->Devices[entry_index];
		}
	}
	if (pDevice == NULL) {
		osm_printk("Fail to change queue_depth of scsi id:lun(%d:%d)\n", sdev->id, (u32)sdev->lun);
		return -EOPNOTSUPP;
	}
	if (IS_SSP(pDevice)) {
		res = min(queue_depth, SAS_MAX_QD);
		if (sdev->tagged_supported)
			scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
		else {
			scsi_adjust_queue_depth(sdev, 0, 1);
			res = 1;
		}
	} else if (IS_STP_OR_SATA_ATAPI(pDevice)) {
		res = min(queue_depth, ARCSAS_MAX_CMD_PERDEV);
		if (sdev->tagged_supported)
			scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
		else {
			scsi_adjust_queue_depth(sdev, 0, 1);
			res = 1;
		}			
	}
	osm_printk("queue_depth of scsi id:lun(%d:%d) has changed to %d\n", sdev->id, (u32)sdev->lun, res);
	return res;
}
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11)
static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth)
{
	unsigned short entry_index;
	int res = 0;
	PDomain_Device pDevice = NULL;
	struct Scsi_Host *host;
	struct AdapterControlBlock *pACB;

	host = sdev->host;
	pACB = (struct AdapterControlBlock *)host->hostdata;
	if (sdev->lun) {
		unsigned char i;
		entry_index = MAKE_DEVICE_ENTRY_INDEX(sdev->id, sdev->lun);
		for (i = 0; i <= ARCSAS_LUN_SUPPORTED; i++) {
			if (pACB->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
				pDevice = &pACB->DevicesLun[i];
				break;
			}
		}
	} else {
		entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, sdev->id);
		if (entry_index == ID_NOT_MAPPED) {
			pDevice = NULL;/*hot unplug, device is disconnected*/
		} else {
			pDevice = &pACB->Devices[entry_index];
		}
	}
	if (pDevice == NULL) {
		osm_printk("Fail to change queue_depth of scsi id:lun(%d:%d)\n", sdev->id, sdev->lun);
		return -EOPNOTSUPP;
	}
	if (IS_SSP(pDevice)) {
		res = min(queue_depth, SAS_MAX_QD);
		if (sdev->tagged_supported)
			scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
		else {
			scsi_adjust_queue_depth(sdev, 0, 1);
			res = 1;
		}
	} else if (IS_STP_OR_SATA_ATAPI(pDevice)) {
		res = min(queue_depth, ARCSAS_MAX_CMD_PERDEV);
		if (sdev->tagged_supported)
			scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
		else {
			scsi_adjust_queue_depth(sdev, 0, 1);
			res = 1;
		}			
	}
	osm_printk("queue_depth of scsi id:lun(%d:%d) has changed to %d\n", sdev->id, sdev->lun, res);
	return res;
}
#else
static ssize_t arcsas_change_queue_depth(struct device *dev, const char *buf, size_t count)
{
	struct scsi_device *sdev = to_scsi_device(dev);
	PDomain_Device pDevice = NULL;
	unsigned short entry_index;
	int queue_depth;
	int res = 0;
	struct Scsi_Host *host;
	struct AdapterControlBlock *pACB;

	host = sdev->host;
	pACB = (struct AdapterControlBlock *)host->hostdata;
	if (sdev->lun) {
		unsigned char i;
		entry_index = MAKE_DEVICE_ENTRY_INDEX(sdev->id, sdev->lun);
		for (i = 0; i <= ARCSAS_LUN_SUPPORTED; i++) {
			if (pACB->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
				pDevice = &pACB->DevicesLun[i];
				break;
			}
		}
	} else {
		entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, sdev->id);
		if (entry_index == ID_NOT_MAPPED) {
			pDevice = NULL;/*hot unplug, device is disconnected*/
		} else {
			pDevice = &pACB->Devices[entry_index];
		}
	}
	if (pDevice == NULL) {
		osm_printk("Fail to change queue_depth of scsi id:lun(%d:%d)\n", sdev->id, sdev->lun);
		return -EINVAL;
	}
	queue_depth = simple_strtoul(buf, NULL, 0);
	if (IS_SSP(pDevice)) {
		res = min(queue_depth, SAS_MAX_QD);
		if (sdev->tagged_supported)
			scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
		else {
			scsi_adjust_queue_depth(sdev, 0, 1);
			res = 1;
		}
	} else if (IS_STP_OR_SATA_ATAPI(pDevice)) {
		res = min(queue_depth, ARCSAS_MAX_CMD_PERDEV);
		if (sdev->tagged_supported)
			scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
		else {
			scsi_adjust_queue_depth(sdev, 0, 1);
			res = 1;
		}			
	}
	osm_printk("queue_depth of scsi id:lun(%d:%d) has changed to %d\n", sdev->id, sdev->lun, res);
	return res;
}
static struct device_attribute arcsas_queue_depth_attr = 
{
	.attr = {
		.name =	"queue_depth",
		.mode =	S_IRUSR | S_IWUSR,
	},
	.store = arcsas_change_queue_depth
};
static struct device_attribute *arcsas_scsi_device_attr[] = 
{
	&arcsas_queue_depth_attr,
	NULL,
};
#endif
/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
static irqreturn_t arcsas_msix_check_int(int irq, void *dev_id)
{
	return IRQ_WAKE_THREAD;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
static irqreturn_t arcsas_msix_thread_1330(int irq, void *dev_id)
#else
static irqreturn_t arcsas_msix_thread_1330(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
	irqreturn_t rtn;
	struct msix_data *msix_data = dev_id;

	rtn = arcsas_interrupt_msix_1330(msix_data);
	return rtn;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
static irqreturn_t arcsas_interrupt(int irq, void *dev_id)
#else
static irqreturn_t arcsas_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
	irqreturn_t rtn;
	u32  int_cause;
	struct AdapterControlBlock *pACB = dev_id;

	if (pACB->acb_flag & ACB_FLAG_ADAPTER_REMOVED)
		return IRQ_HANDLED;
	int_cause = arcsas_disable_hba_isr(dev_id);
	rtn = arcsas_interrupt_check(dev_id);
	arcsas_enable_hba_isr(dev_id, int_cause);
	return rtn;
}

const char *arcsas_info(struct Scsi_Host *host)
{
	char *type;
	struct AdapterControlBlock *pACB = (struct AdapterControlBlock *) host->hostdata;
	static char buf[256];

	switch (pACB->adapter_type) {
		case ARC_1330:
			type = "ARC1330";
			break;
		case ARC_1320:
			type = "ARC1320";
			break;
		case ARC_1300:
			type = "ARC1300";
			break;
		default:
			type = "unknown";
			break;
	}
	sprintf(buf, "Areca %s SAS Host Bus Adapter, arcsas version %s\n",
			type, ARCSAS_DRIVER_VERSION);
	return buf;
}
/*
************************************************************************
************************************************************************
*/
static Bool arcsas_initialize_hba(struct AdapterControlBlock *pACB, struct pci_dev *pcidev)
{
	unsigned long addr;
	unsigned long range;
	void __iomem *mem_map_base = NULL;
	void __iomem *mem_map_base_ext = NULL;

#if ARCSAS_DBG_INIT
		osm_printk("%s:\n", __func__);
#endif
	osm_atomic_set(&pACB->outstandingccbcount, 0);
	osm_atomic_set(&pACB->all_device_count, 0);
	pACB->enclosure_mask = 0;
	/*refer to pCore->Mmio_Base = controller.Base_Address[MV_PCI_BAR] in Core_ModuleInitialize function of core_exp.c*/
	switch (pACB->adapter_type) {
		case ARC_1330: {
			u32	tmp;
			pACB->main_irq_mask = IRQ_SAS_A_1330;
			addr  = pci_resource_start(pcidev, 0);
			range = pci_resource_len(pcidev, 0);
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
			mem_map_base_ext = ioremap(addr, range);
	#else
			if (pci_resource_flags(pcidev, 0) & IORESOURCE_CACHEABLE) {
				mem_map_base_ext = ioremap(addr, range);
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("line %d Cachable I/O memory\n", __LINE__);
				#endif
			} else {
				mem_map_base_ext = ioremap_nocache(addr, range);	//This is ARC1330 really belongs to.
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("line %d Uncachable I/O memory\n", __LINE__);
				#endif
			}
	#endif
			if (!mem_map_base_ext) {
				osm_printk("mem_map_base_ext failed......\n");
				return false;
			}
			tmp = osm_readl(NumToPtr(PtrToNum(mem_map_base_ext) + ARCSAS_SAS_B_BASE + PORTS_IMPLEMENT_1330));
			if (tmp == 0x000000ff) {
				pACB->acb_flag |= ACB_FLAG_DUAL_CORE;
				pACB->core_num = 2;
				pACB->main_irq_mask |= IRQ_SAS_B_1330;
			} else {
				pACB->acb_flag &= ~ACB_FLAG_DUAL_CORE;
				pACB->core_num = 1;
			}
			if (!arcsas_core_initialize(pACB)) {
				osm_printk("core initialization is not finished..........\n");
				return false;
			}
			/*
			************************************************************************
			** pCore->mem_map_base: (SAS/SATA A Register) port0...port3 20000h-23FFFh
			** pCore->mem_map_base: (SAS/SATA B Register) port4...port7 24000h-24FFFh
			************************************************************************
			*/
			pACB->mem_map_base_ext = mem_map_base_ext;
			pACB->pCore[0]->mem_map_base_ext = mem_map_base_ext;
			pACB->pCore[0]->mem_map_base = NumToPtr(PtrToNum(mem_map_base_ext) + ARCSAS_SAS_A_BASE);
			if (pACB->core_num == 2) {
				pACB->pCore[1]->mem_map_base_ext = mem_map_base_ext;
				pACB->pCore[1]->mem_map_base = NumToPtr(PtrToNum(mem_map_base_ext) + ARCSAS_SAS_B_BASE);
			}
			if (!arcsas_update_acb_bios_flag_from_bios_1330(pACB)) {
				osm_printk("arcsas_update_acb_bios_flag_from_bios_1330 failed\n");
				return false;
			}
			break;
		}
		case ARC_1320: {
			addr  = pci_resource_start(pcidev, 2);
			range = pci_resource_len(pcidev, 2);
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
			mem_map_base_ext = ioremap(addr, range);
	#else
			if (pci_resource_flags(pcidev, 2) & IORESOURCE_CACHEABLE) {
				mem_map_base_ext = ioremap(addr, range);
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("line %d Cachable I/O memory\n", __LINE__);
				#endif
			} else {
				mem_map_base_ext = ioremap_nocache(addr, range);//This is ARC1320 really belongs to.
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("line %d Uncachable I/O memory\n", __LINE__);
				#endif
			}
	#endif
			if (!mem_map_base_ext) {
				osm_printk("mem_map_base_ext failed......\n");
				return false;
			}
			pACB->acb_flag |= ACB_FLAG_DUAL_CORE;
			pACB->core_num = 2;
			if (!arcsas_core_initialize(pACB)) {
				osm_printk("core initialization is not finished..........\n");
				return false;
			}
			/*
			************************************************************************
			** pCore->mem_map_base: (SAS/SATA A Register) port0...port3 20000h-23FFFh
			** pCore->mem_map_base: (SAS/SATA B Register) port4...port7 24000h-24FFFh
			************************************************************************
			*/
			pACB->mem_map_base_ext = mem_map_base_ext;
			pACB->pCore[0]->mem_map_base_ext = mem_map_base_ext;
			pACB->pCore[1]->mem_map_base_ext = mem_map_base_ext;
			pACB->pCore[0]->mem_map_base = NumToPtr(PtrToNum(mem_map_base_ext) + ARCSAS_SAS_A_BASE);
			pACB->pCore[1]->mem_map_base = NumToPtr(PtrToNum(mem_map_base_ext) + ARCSAS_SAS_B_BASE);
			if (!arcsas_update_acb_bios_flag_from_bios(pACB)) {
				osm_printk("arcsas_update_acb_bios_flag_from_bios failed\n");
				return false;
			}
			break;
		}
		case ARC_1300: {
			addr  = (unsigned long)pci_resource_start(pcidev, 2);
			range = pci_resource_len(pcidev, 2);
			if (!addr || !range) {
				return false;
			}
			if (pci_resource_flags(pcidev, 2) & IORESOURCE_MEM) {
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
				mem_map_base_ext = (void *)ioremap(addr, range);
	#else
				if (pci_resource_flags(pcidev, 2) & IORESOURCE_CACHEABLE) {
					#if ARCSAS_DBG_INIT_CONTEXT
						osm_printk("line %d Cachable I/O memory\n", __LINE__);
					#endif
					mem_map_base_ext = (void *)ioremap(addr, range);
				} else {
					#if ARCSAS_DBG_INIT_CONTEXT
						osm_printk("line %d Uncachable I/O memory\n", __LINE__);
					#endif
					mem_map_base_ext = (void *)ioremap_nocache(addr, range);
				}
	#endif
			} else {
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("I/O port\n");
				#endif
				mem_map_base_ext = (void *)addr;
			}
			addr  = pci_resource_start(pcidev, 4);
			range = pci_resource_len(pcidev, 4);
			if (!addr || !range) {
				osm_printk("addr or range failed\n");
				return false;
			}
			if (pci_resource_flags(pcidev, 4) & IORESOURCE_MEM) {
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
				mem_map_base = ioremap(addr, range);
	#else
				if (pci_resource_flags(pcidev, 4) & IORESOURCE_CACHEABLE) {
					mem_map_base = ioremap(addr, range);
					#if ARCSAS_DBG_INIT_CONTEXT
						osm_printk("line %d Cachable I/O memory mem_map_base = 0x%p\n", __LINE__, mem_map_base);
					#endif
				} else {
					mem_map_base = ioremap_nocache(addr, range);//This is ARC1300 really belongs to.
					#if ARCSAS_DBG_INIT_CONTEXT
						osm_printk("line %d Uncachable I/O memory mem_map_base = 0x%p\n", __LINE__, mem_map_base);
					#endif
				}
	#endif
			} else {
				#if ARCSAS_DBG_INIT_CONTEXT
					osm_printk("I/O port\n");
				#endif
				mem_map_base_ext = (void *)addr;
			}
			if (!mem_map_base) {
				if (pci_resource_flags(pcidev, 2) & IORESOURCE_MEM) {
					osm_printk("line %d ARC1300 iounmap\n", __LINE__);
					iounmap(mem_map_base_ext);
				}
				osm_printk("mem_map_base failed\n");
				return false;
			}
			pACB->acb_flag &= ~ACB_FLAG_DUAL_CORE;
			pACB->core_num = 1;
			if (!arcsas_core_initialize(pACB)) {
				osm_printk("line %d ARC1300 iounmap\n", __LINE__);
				iounmap(mem_map_base);
				iounmap(mem_map_base_ext);
				return false;
			}
			pACB->pCore[0]->io_map_base = mem_map_base_ext;
			pACB->pCore[0]->mem_map_base = mem_map_base;
			if (!arcsas_update_acb_bios_flag_from_bios(pACB)) {
				osm_printk("arcsas_update_acb_bios_flag_from_bios failed\n");
				return false;
			}
			break;
		}
		default:{
			_ASSERT();
		}
	}
	if (arcsas_reset_controller(pACB) == false) {
		arcsas_teardown_dma_pool(pACB);
		arcsas_iounmap(pACB);
		osm_printk("arcsas_reset_controller failed, iounmap!\n");
		return false;
	}
	return arcsas_init_chip(pACB);
}
/*
************************************************************************
************************************************************************
*/
int arcsas_error_handler(struct AdapterControlBlock *pACB)
{
	unsigned short i;
	u32 DoneQ_index;
	int rtn = SUCCESS;
	unsigned long flags;
	struct CommandControlBlock *pCCB, *pCCBtimeout;
	PCore_Extension pCore;
	PDomain_Device pDevice;
	PDomain_PM pPM;
	struct _Domain_Port *pPort;
	COMPLETION_QUEUE_ENTRY entry;
	PCOMPLETION_QUEUE_ENTRY pDoneQ;

	//set_current_state(TASK_INTERRUPTIBLE);
	do {
start:
		//if (osm_list_empty(&pACB->timeout_list)) {
		//	osm_printk("Error handler sleeping\n");
		//	set_current_state(TASK_INTERRUPTIBLE);
		//	schedule();
		//	continue;
		//}
		//__set_current_state(TASK_RUNNING);
		//osm_printk("Error handler waking up\n");
		arcsas_spin_lock_irqsave(&pACB->ehandler_lock, &flags);
		if (osm_list_empty(&pACB->timeout_list)) {
			arcsas_spin_unlock_irqrestore(&pACB->ehandler_lock, &flags);
			break;
		}
		pCCBtimeout = (struct CommandControlBlock *)osm_list_first_entry(&pACB->timeout_list,
		    struct CommandControlBlock, Queue_Pointer);
		#if ARCSAS_DBG_TIMEOUT
			osm_printk("%s: No.1 pCCBtimeout = 0x%p\n", __func__, pCCBtimeout);
		#endif
		osm_list_del_init(&pCCBtimeout->Queue_Pointer);
		arcsas_spin_unlock_irqrestore(&pACB->ehandler_lock, &flags);
		if (IS_A_PM_REQ(pCCBtimeout)) {
			pPM = pCCBtimeout->pPM;
			pPort = pPM->pPort;
			pCore = pPort->pCore;
			pDoneQ = (PCOMPLETION_QUEUE_ENTRY)pCore->DoneQ;
			for (i = 0; i < pCore->Slot_Count_Supported; i++) {/*Slot_Count_Supported = 512*/
				pCCB = pCore->Running_CCB[i];
				if (pCCB != NULL) {
					if (pPM->EntryIndex != pCCB->EntryIndex) {
						/*
						*************************************************
						** this ccb is not belong to this timeout Device
						*************************************************
						*/
						continue;
					}
					arcsas_spin_lock_irqsave(&pCore->doneq_lock, &flags);
					/* check if this is being completed */
					//DoneQ_index = pCore->last_DoneQ_index;
					/* This while-loop section below is to prevent new done entry pops out while you check the completion queue
					** If it really comes up, bad luck, just skip the checking process in the completion queue.
					** Then it will be aborted and retried in the following sas/sata_error_handling
					*/
					#if ARCSAS_DBG_TIMEOUT
						osm_printk("%s: No.2 pCCB(addr=0x%p), pCore->current_DoneQ_index = 0x%x, i = 0x%x\n",
						    __func__, pCCB, pCore->current_DoneQ_index, i);
					#endif
					for (DoneQ_index = pCore->last_DoneQ_index; DoneQ_index == ARCSAS_LE32_TO_CPU(((u32 *)pDoneQ)[0]); DoneQ_index++) {
						if (DoneQ_index >= pCore->Max_DoneQ_Index) {/* 522 */
							DoneQ_index = 0;
						}
						entry = pDoneQ[DoneQ_index + 1];
						ARCSAS_LE32_TO_CPU_PTR(&entry);
						if (!(entry.ATTENTION) && ((unsigned short)entry.SLOT_NM) == i) {
							/* the timed out request is actually in the completion queue we handle that and move on */
							if (pCCB == pCCBtimeout) {
								/*Areca SAS controller core device ccb timeout*/
								osm_printk("Receive the timeout ccb from Done_queue, device 0x%p scsi id %d, scsi lun %d.\n", pPM, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pPM->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pPM->EntryIndex));
								osm_printk("Time_Out = %lld, CCB_Status = 0x%x.\n", pCCB->Time_Out, pCCB->CCB_Status);
								osm_printk("Cdb[0] = 0x%x  Cdb[1] = 0x%x  Cdb[2] = 0x%x  Cdb[3] = 0x%x  Cdb[4] = 0x%x\n", pCCB->Cdb[0], pCCB->Cdb[1], pCCB->Cdb[2], pCCB->Cdb[3], pCCB->Cdb[4]);
								//osm_writel(INT_CMD_CMPL_MASK, pCore->mem_map_base + COMMON_IRQ_STATE);// clear CMD_CMPLT ASAP
								arcsas_complete_command_in_doneQ(pCore, pCCB, pPort);
								arcsas_spin_unlock_irqrestore(&pCore->doneq_lock, &flags);
								//arcsas_interrupt_service_bh((unsigned long)pCore);
								goto start;
							}
							break;
						}
					}
					#if ARCSAS_DBG_TIMEOUT
						osm_printk("%s: No.3\n", __func__);
					#endif
					arcsas_spin_unlock_irqrestore(&pCore->doneq_lock, &flags);
				}
			}
			if (pPM->Timeout_Count >= DEVICE_DOWN_LEVEL) {
				pPM->Error_Handling_State = EH_SET_DISK_DOWN;
			} else if (pPM->Timeout_Count >= DEVICE_RESET_LEVEL) {
				pPM->Error_Handling_State = EH_SET_DISK_RESET;
			} else if (pPM->Timeout_Count >= PORT_RESET_LEVEL) {
				pPM->Error_Handling_State = EH_PORT_RESET;
			}
			rtn = arcsas_pm_error_handling(pPM, pCCBtimeout);
		} else {
			pDevice = pCCBtimeout->pDevice;
			pPort = pDevice->pPort;
			pCore = pPort->pCore;
			pDoneQ = (PCOMPLETION_QUEUE_ENTRY)pCore->DoneQ;
			for (i = 0; i < pCore->Slot_Count_Supported; i++) {
				pCCB = pCore->Running_CCB[i];
				if (pCCB != NULL) {
					if (pDevice->EntryIndex != pCCB->EntryIndex) {
						/*
						*************************************************
						** this ccb is not belong to this timeout Device
						*************************************************
						*/
						continue;
					}
					arcsas_spin_lock_irqsave(&pCore->doneq_lock, &flags);
					/* check if this is being completed */
					//DoneQ_index = pCore->last_DoneQ_index;
					/* This while-loop section below is to prevent new done entry pops out while you check the completion queue
					** If it really comes up, bad luck, just skip the checking process in the completion queue.
					** Then it will be aborted and retried in the following sas/sata_error_handling
					*/
					#if ARCSAS_DBG_TIMEOUT
						osm_printk("%s: No.4 pCCB(addr=0x%p), pCore->current_DoneQ_index = 0x%x, i = 0x%x\n", __func__, pCCB, pCore->current_DoneQ_index, i);
					#endif
					for (DoneQ_index = pCore->last_DoneQ_index; DoneQ_index == ARCSAS_LE32_TO_CPU(((u32 *)pDoneQ)[0]); DoneQ_index++) {
						if (DoneQ_index >= pCore->Max_DoneQ_Index) {
							DoneQ_index = 0;
						}
						entry = pDoneQ[DoneQ_index + 1];
						ARCSAS_LE32_TO_CPU_PTR(&entry);
						if (!(entry.ATTENTION) && ((unsigned short)entry.SLOT_NM) == i) {
							/* the timed out request is actually in the completion queue we handle that and move on */
							if (pCCB == pCCBtimeout) {
								/*Areca SAS controller core device ccb timeout*/
								osm_printk("Receive the timeout ccb from Done_queue, device 0x%p scsi id %d, scsi lun %d.\n", pDevice, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
								osm_printk("Time_Out = %lld, CCB_Status = 0x%x.\n", pCCB->Time_Out, pCCB->CCB_Status);
								osm_printk("Cdb[0] = 0x%x  Cdb[1] = 0x%x  Cdb[2] = 0x%x  Cdb[3] = 0x%x  Cdb[4] = 0x%x\n", pCCB->Cdb[0], pCCB->Cdb[1], pCCB->Cdb[2], pCCB->Cdb[3], pCCB->Cdb[4]);
								//osm_writel(INT_CMD_CMPL_MASK, pCore->mem_map_base + COMMON_IRQ_STATE);// clear CMD_CMPLT ASAP
								arcsas_complete_command_in_doneQ(pCore, pCCB, pPort);
								arcsas_spin_unlock_irqrestore(&pCore->doneq_lock, &flags);
								//arcsas_interrupt_service_bh((unsigned long)pCore);
								//arcsas_enable_hba_isr(pACB);
								goto start;
							}
							break;
						}
					}
					#if ARCSAS_DBG_TIMEOUT
						osm_printk("%s: No.5\n", __func__);
					#endif
					arcsas_spin_unlock_irqrestore(&pCore->doneq_lock, &flags);
				}
			}
			if (pDevice->Timeout_Count >= DEVICE_DOWN_LEVEL) {
				pDevice->Error_Handling_State = EH_SET_DISK_DOWN;
			} else if (pDevice->Timeout_Count >= DEVICE_RESET_LEVEL) {
				pDevice->Error_Handling_State = EH_SET_DISK_RESET;
			} else if (pDevice->Timeout_Count >= PORT_RESET_LEVEL) {
				pDevice->Error_Handling_State = EH_PORT_RESET;
			}
			if ((pPort->Type & PORT_TYPE_SAS) && IS_SSP(pDevice)) {
				/* this port is SAS device */
				rtn = arcsas_sas_error_handling(pCCBtimeout);
			} else if ((pDevice != NULL) && IS_STP_OR_SATA_ATAPI(pDevice)) {
				rtn = arcsas_sata_error_handling(pCCBtimeout);
				#if ARCSAS_DBG_TIMEOUT
					osm_printk("%s: No.6\n", __func__);
				#endif
			}
		}
		#if ARCSAS_DBG_TIMEOUT
			osm_printk("%s: No.7\n", __func__);
		#endif
		//set_current_state(TASK_INTERRUPTIBLE);
	} while (!osm_list_empty(&pACB->timeout_list));
	#if ARCSAS_DBG_TIMEOUT
		osm_printk("%s: leaving now\n", __func__);
	#endif
	//__set_current_state(TASK_RUNNING);
	//pACB->eh_handler = NULL;
	return rtn;
}
/*
************************************************************************
************************************************************************
*/
void arcsas_eh_wakeup(struct AdapterControlBlock *pACB)
{
	wake_up_process(pACB->eh_handler);
}
/*
************************************************************************
************************************************************************
*/
static int arcsas_set_dma_mask(struct pci_dev *pcidev)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0))
	if (!dma_set_mask(&pcidev->dev, DMA_BIT_MASK(64))) {
		if (dma_set_coherent_mask(&pcidev->dev, DMA_BIT_MASK(64))) {
			if (dma_set_coherent_mask(&pcidev->dev, DMA_BIT_MASK(32))) {
				osm_printk("arcsas: set DMA 64/32-bit mask failed\n");
				return -ENXIO;
			}
		}
	} else {
		if (dma_set_mask(&pcidev->dev, DMA_BIT_MASK(32)) ||
		    dma_set_coherent_mask(&pcidev->dev, DMA_BIT_MASK(32))) {
			osm_printk("arcsas: set DMA 32-bit mask failed\n");
			return -ENXIO;
		}
	}
	return 0;
#else
	if (!pci_set_dma_mask(pcidev, DMA_BIT_MASK(64))) {
		if (pci_set_consistent_dma_mask(pcidev, DMA_BIT_MASK(64))) {
			if (pci_set_consistent_dma_mask(pcidev, DMA_BIT_MASK(32))) {
				osm_printk("arcsas: set DMA 64-bit mask failed\n");
				return -ENXIO;
			}
		}
	} else {
		if (pci_set_dma_mask(pcidev, DMA_BIT_MASK(32)) ||
		    pci_set_consistent_dma_mask(pcidev, DMA_BIT_MASK(32))) {
			osm_printk("arcsas: set DMA 32-bit mask failed\n");
			return -ENXIO;
		}
	}
	return 0;
#endif
}

#ifdef CONFIG_PM
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
/*
************************************************************************
************************************************************************
*/
static int arcsas_suspend(struct pci_dev *pdev, pm_message_t state)
{
	struct Scsi_Host *shost = pci_get_drvdata(pdev);
	struct AdapterControlBlock *pACB = (struct AdapterControlBlock *)shost ->hostdata;

	#if ARCSAS_DBG_HIBERNATE
		osm_printk("%s:\n", __func__);
	#endif
	scsi_block_requests(shost);
//	arcsas_disable_interrupt(pACB);
	//for (i = ARCSAS_MIN_DEVICE_ID; i < ARCSAS_MAX_DEVICE_ID; i++) {/*(0..135,136..143,144..151)*/
	//	if (pACB->DevicesExpandersPMs_EntryIndex[i] == ID_NOT_MAPPED) {
	//		continue;
	//	}
	//	pDevice = &pACB->Devices[Get_DevicesExpandersPMs_EntryIndex(pACB, i)];/*(0..135)*/
	//	arcsas_delete_device_timer(pDevice);
	//}
	//for (i = ARCSAS_MIN_PM_ID; i <= ARCSAS_MAX_PM_ID; i++) {
	//	if (pACB->DevicesExpandersPMs_EntryIndex[i] != ID_NOT_MAPPED) {
	//		index = pACB->DevicesExpandersPMs_EntryIndex[i];
	//		pPM = &pACB->PMs[index];
	//		arcsas_delete_pm_timers(pPM);
	//	}
	//}
	arcsas_free_irq(pdev, pACB);
	if (pACB->eh_handler)
		kthread_stop(pACB->eh_handler);
	arcsas_disable_interrupt(pACB);
	arcsas_teardown_dma_pool(pACB);
	osm_vfree(pACB->pCacheMem);
	arcsas_iounmap(pACB);
	pci_save_state(pdev);
	pci_disable_device(pdev);
	pci_set_power_state(pdev, pci_choose_state(pdev, state));
	return 0;
}
/*
************************************************************************
************************************************************************
*/
static int arcsas_resume(struct pci_dev *pcidev)
{
	struct Scsi_Host *shost = pci_get_drvdata(pcidev);
	struct AdapterControlBlock *pACB = (struct AdapterControlBlock *)shost ->hostdata;
	unsigned int retval = -ENXIO; //, core_index;

	#if ARCSAS_DBG_HIBERNATE
		osm_printk("%s:\n", __func__);
	#endif
	pci_set_power_state(pcidev, PCI_D0);
	pci_enable_wake(pcidev, PCI_D0, 0);
	pci_restore_state(pcidev);
	if (pci_enable_device(pcidev)) {
		printk("%s: pci_enable_device error \n", __func__);
		goto phase_0;
	}
	if (arcsas_set_dma_mask(pcidev))
		goto phase_0;
	pci_set_master(pcidev);
	if (!arcsas_initialize_hba(pACB, pcidev)) {
		osm_printk("arcsas: Fail to map pci resources\n");
		goto phase_0;
	}
	retval = arcsas_request_irq(pcidev, pACB);
	if (retval == ARC_FAILURE) {
		osm_printk("Fail to request irq\n");
		goto phase_1;
	}
	arcsas_enable_interrupt(pACB);
	scsi_unblock_requests(shost);
	return 0;
phase_1:
	destroy_workqueue(pACB->scsi_lookup);
	flush_scheduled_work();
	kfree(pACB->dv_event);
phase_0:
	arcsas_teardown_dma_pool(pACB);
	osm_vfree(pACB->DevicesLun);
	arcsas_iounmap(pACB);
	pci_release_regions(pcidev);
	pci_set_drvdata(pcidev, NULL);
	scsi_host_put(shost);
	pci_disable_device(pcidev);
	return -ENODEV;
}
#endif
#endif
/*
************************************************************************
************************************************************************
*/
void arcsas_init_virtual_ses_device(PACB pACB)
{
	unsigned char SasAddr[8] = {0xEF, 0x00, 0xD3, 0x17, 0xdf, 0xb4, 0x01, 0x50};
	unsigned char Model_Number[20] = {"Areca Virtual SES"};
	unsigned char Serial_Number[10] = {"SGPIO01234"};

	osm_memcpy(&pACB->SesSgpioVirtualDevice.SASAddr.value, &SasAddr, 8);
	osm_memcpy(&pACB->SesSgpioVirtualDevice.WWN.value, &SasAddr, 8);
	osm_memcpy(&pACB->SesSgpioVirtualDevice.Model_Number, &Model_Number, 20);
	osm_memcpy(&pACB->SesSgpioVirtualDevice.Serial_Number, &Serial_Number, 10);
	pACB->SesSgpioVirtualDevice.pCore = pACB->pCore[0];
	pACB->SesSgpioVirtualDevice.Dev_Type = DT_SES_DEVICE;
	pACB->SesSgpioVirtualDevice.EntryIndex = ARCSAS_VIRTUAL_DEVICE_ID;
	pACB->SesSgpioVirtualDevice.MaxOfEnclosureSlotNumber = pACB->pCore[0]->Phy_Num * pACB->core_num;
	pACB->SesSgpioVirtualDevice.ses_pages_supported = ARCSAS_BIT(1, SES_PG_ADDITIONAL_ELEMENT_STATUS) |
	    ARCSAS_BIT(1, SES_PG_CONFIGURATION) | ARCSAS_BIT(1, SES_PG_ENCLOSURE_STATUS) | ARCSAS_BIT(1, SES_PG_ELEMENT_DESCRIPTOR);
	pACB->SesSgpioVirtualDevice.State = DEVICE_STATE_INIT_DONE;
	pACB->SesSgpioVirtualDevice.Status = DEVICE_STATUS_EXISTING;
	return;
}
/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 32)
static void arcsas_disable_aspm(struct pci_dev *pdev, u16 state)
{
	int pos;
	u16 reg16;

	/*
	 * Both device and parent should have the same ASPM setting.
	 * Disable ASPM in downstream component first and then upstream.
	 */
	pos = pci_find_capability(pdev, PCI_CAP_ID_EXP);
	pci_read_config_word(pdev, pos + PCI_EXP_LNKCTL, &reg16);
	reg16 &= ~state;
	pci_write_config_word(pdev, pos + PCI_EXP_LNKCTL, reg16);

	if (!pdev->bus->self)
		return;

	pos = pci_find_capability(pdev->bus->self, PCI_CAP_ID_EXP);
	pci_read_config_word(pdev->bus->self, pos + PCI_EXP_LNKCTL, &reg16);
	reg16 &= ~state;
	pci_write_config_word(pdev->bus->self, pos + PCI_EXP_LNKCTL, reg16);
}
#endif
/*
************************************************************************
************************************************************************
*/
static int arcsas_request_msix_irq(PACB pACB, int index, unsigned int vector)
{
	int ret;
	struct msix_data *msix_data;
#if ARCSAS_DBG_INIT
	osm_printk("%s:\n", __func__);
#endif

	msix_data = &pACB->msix_data[index];
	msix_data->pACB = pACB;
	msix_data->index = index;
	msix_data->vector = vector;
	msix_data->root = index / pACB->vectors_per_core;
	msix_data->queue = index % pACB->vectors_per_core;
	snprintf(msix_data->name, 32, "%s-%d", "arcsas", index);
	msix_data->valid = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
	ret = request_threaded_irq(vector, arcsas_msix_check_int, arcsas_msix_thread_1330, 0,
				msix_data->name, msix_data);
#else
	ret = request_irq(vector, arcsas_msix_thread_1330, 0, msix_data->name, msix_data);
#endif
	if (ret) {
		osm_printk("allocate interrupt for msix idx %d vector %d failed\n", index, vector);
		return -1;
	}
	tasklet_init(&pACB->isr_tasklet[index], arcsas_interrupt_service_msix_1330, (unsigned long)msix_data);
	msix_data->valid = 1;
	return 0;
}

static int arcsas_request_irq(struct pci_dev *pdev, PACB pACB)
{
	PCORE pCore;
	struct msix_entry	entries[ARCSAS_NUM_MSIX_VECTORS];
	u32	temp;
#if 0
//#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0)
	unsigned long flags;
	int nvec, i, j;

	if ((msix_enable == 0) || (pACB->adapter_type != ARC_1330))
		goto msi_init0;
	nvec = pci_alloc_irq_vectors(pdev, 1, ARCSAS_NUM_MSIX_VECTORS,
			PCI_IRQ_MSIX);
	if (nvec > 0) {
		pACB->vector_count = nvec;
		pACB->vectors_per_core = pACB->vector_count / (u16)pACB->core_num;
		pACB->msix_data = kzalloc(sizeof(struct msix_data) * nvec, GFP_KERNEL);
		if (!pACB->msix_data) {
			pci_free_irq_vectors(pdev);
			goto msi_init0;
		}
		flags = 0;
		for (i = 0; i < nvec; i++) {
			entries[i].entry = i;
			entries[i].vector = pci_irq_vector(pdev, i);
			if (arcsas_request_msix_irq(pACB, i, entries[i].vector)) {
				osm_printk("arcsas%d: request_irq =%d failed!\n",
					pACB->host->host_no, entries[i].vector);
				for (j = 0 ; j < i ; j++)
					free_irq(entries[j].vector, &pACB->msix_data[j]);
				kfree(pACB->msix_data);
				pci_free_irq_vectors(pdev);
				goto msi_init0;
			}
			pACB->entries[i] = entries[i];
		}
		pACB->acb_flag |= ACB_FLAG_MSIX_ENABLED;
		for (i = 0; i < (int)(pACB->core_num); i++) {
			pCore = pACB->pCore[i];
			pCore->queue_num = pACB->vectors_per_core;
			for (j = 0; j < pCore->queue_num; j++) {
				pCore->queues[j].msix_idx = i * pCore->queue_num + j;
				osm_writel(pCore->queues[j].msix_idx, pCore->mem_map_base + INT_ROUTING_0 + j * 4);
				temp = osm_readl(pCore->mem_map_base + COMPLETITION_Q0_WR_PTR + j * COMPLETITION_Q_REG_SIZE);
				osm_writel(temp, pCore->mem_map_base + COMPLETITION_Q0_RD_PTR + j * COMPLETITION_Q_REG_SIZE);
			}
		}
		osm_printk("arcsas%d: msi-x INT enabled\n", pACB->host->host_no);
	} else {
msi_init0:
		arcsas_disable_msix(pACB);
		arcsas_disable_msix_SM(pACB);
		nvec = pci_alloc_irq_vectors(pdev, 1, ARCSAS_NUM_MSI_VECTORS, PCI_IRQ_MSI);
		if (nvec < 1) {
			nvec = pci_alloc_irq_vectors(pdev, 1, 1, PCI_IRQ_LEGACY);
			if (nvec < 1)
				return FAILED;
		} else
			pACB->acb_flag |= ACB_FLAG_MSI_ENABLED;
		pACB->vector_count = nvec;
		flags = IRQF_SHARED;
		for (i = 0; i < nvec; i++) {
			entries[i].vector = pci_irq_vector(pdev, i);
			if (request_irq(entries[i].vector, arcsas_interrupt,
				flags, "arcsas", pACB)) {
				pr_warn("arcsas%d: request_irq =%d failed!\n",
					pACB->host->host_no, entries[i].vector);
				while (--i >= 0)
					free_irq(pci_irq_vector(pdev, i), pACB);
				pci_free_irq_vectors(pdev);
				return FAILED;
			}
			pACB->entries[i] = entries[i];
		}
		if (pACB->acb_flag & ACB_FLAG_MSI_ENABLED) {
			arcsas_enable_msi(pACB);
			osm_printk("arcsas%d: msi INT enabled\n", pACB->host->host_no);
		}
	}
	return SUCCESS;
//#else
#endif
	int	i, j, cnt, msix_count, base = 0;
	u16	message_ctl;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
	int	nvec;
#endif
#if ARCSAS_DBG_INIT
	osm_printk("%s:\n", __func__);
#endif

	if ((msix_enable == 0) || (pACB->adapter_type != ARC_1330))
		goto msi_int;
	base = pci_find_capability(pdev, PCI_CAP_ID_MSIX);
	if (!base)
		goto msix_off;
	pci_read_config_word(pdev, base + 2, &message_ctl);
	msix_count = (message_ctl & 0x3FF) + 1;
	msix_count = ARCSAS_MIN(ARCSAS_MAX_QUEUE_NUM, msix_count);
	for (i = 0; i < msix_count; i++)
		entries[i].entry = i;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 14, 0)
	cnt = pci_enable_msix_range(pdev, entries, 1, msix_count);
	if (cnt < 0)
		goto msix_off;
#else
	cnt = pci_enable_msix(pdev, entries, msix_count);
	if (cnt < 0)
		goto msix_off;
	if (cnt) {
		if (pci_enable_msix(pdev, entries, cnt))
			goto msix_off;
	} else
		cnt = msix_count;
#endif		
	pACB->vector_count = cnt;
	pACB->vectors_per_core = pACB->vector_count / (u16)pACB->core_num;
	pACB->msix_data = kzalloc(sizeof(struct msix_data) * cnt, GFP_KERNEL);
	if (!pACB->msix_data)
		goto msix_off;
	for (i = 0; i < cnt; i++) {
		if (arcsas_request_msix_irq(pACB, i, entries[i].vector)) {
			osm_printk("arcsas%d: request_irq =%d failed!\n",
				pACB->host->host_no, pdev->irq);
			for (j = 0 ; j < i ; j++)
				free_irq(entries[j].vector, &pACB->msix_data[j]);
			kfree(pACB->msix_data);
			pci_disable_msix(pdev);
			goto msix_off;
		}
		pACB->entries[i] = entries[i];
	}
	pACB->acb_flag |= ACB_FLAG_MSIX_ENABLED;
	for (i = 0; i < (int)(pACB->core_num); i++) {
		pCore = pACB->pCore[i];
		pCore->queue_num = pACB->vectors_per_core;
		for (j = 0; j < pCore->queue_num; j++) {
			pCore->queues[j].msix_idx = i * pCore->queue_num + j;
			osm_writel(pCore->queues[j].msix_idx, pCore->mem_map_base + INT_ROUTING_0 + j * 4);
			temp = osm_readl(pCore->mem_map_base + COMPLETITION_Q0_WR_PTR + j * COMPLETITION_Q_REG_SIZE);
			osm_writel(temp, pCore->mem_map_base + COMPLETITION_Q0_RD_PTR + j * COMPLETITION_Q_REG_SIZE);
		}
	}
//	arcsas_enable_msix(pACB);
//	arcsas_enable_msix_SM(pACB);
	osm_printk("arcsas%d: msi-x INT enabled\n", pACB->host->host_no);
	return ARC_SUCCESS;
msix_off:
	arcsas_disable_msix(pACB);
	arcsas_disable_msix_SM(pACB);
	arcsas_enable_msi(pACB);
msi_int:
	if (msi_enable == 0)
		goto msi_off;
	if (!pci_find_capability(pdev, PCI_CAP_ID_MSI))
		goto msi_off;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
	nvec = pci_alloc_irq_vectors(pdev, 1, 1, PCI_IRQ_MSI);
	if (nvec < 0)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 14, 0)
	if (pci_enable_msi_range(pdev, 1, 1) < 0)
#else
	if (pci_enable_msi(pdev))
#endif		
		goto msi_off;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
	if (request_irq(pci_irq_vector(pdev, 0), arcsas_interrupt, IRQF_SHARED,
				"arcsas", pACB)) {
#elif	LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 19)
	if (request_irq(pdev->irq, arcsas_interrupt, IRQF_SHARED, "arcsas", pACB)) {
#else
	if (request_irq(pdev->irq, arcsas_interrupt, SA_SHIRQ, "arcsas", pACB)) {
#endif
		osm_printk("arcsas%d: request_irq =%d failed!\n",
			pACB->host->host_no, pdev->irq);
		pci_disable_msi(pdev);
		goto msi_off;
	}
	for (i = 0; i < pACB->core_num; i++) {
		pCore = pACB->pCore[i];
		if (pACB->adapter_type == ARC_1330)
			tasklet_init(&pCore->isr_tasklet, arcsas_interrupt_service_1330, (unsigned long)pCore);
	//	else
	//		tasklet_init(&pCore->isr_tasklet, arcsas_interrupt_service_bh, (unsigned long)pCore);
	}
	pACB->acb_flag |= ACB_FLAG_MSI_ENABLED;
//	arcsas_enable_msi(pACB);
	osm_printk("arcsas%d: msi INT enabled\n", pACB->host->host_no);
	return SUCCESS;
msi_off:
	arcsas_disable_msi(pACB);
#if	LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 19)
	if (request_irq(pdev->irq, arcsas_interrupt, IRQF_SHARED, "arcsas", pACB)) {
#else
	if (request_irq(pdev->irq, arcsas_interrupt, SA_SHIRQ, "arcsas", pACB)) {
#endif
		osm_printk("arcsas%d: request_irq = %d failed!\n",
			pACB->host->host_no, pdev->irq);
		return FAILED;
	}
	for (i = 0; i < pACB->core_num; i++) {
		pCore = pACB->pCore[i];
		if (pACB->adapter_type == ARC_1330)
			tasklet_init(&pCore->isr_tasklet, arcsas_interrupt_service_1330, (unsigned long)pCore);
		else
			tasklet_init(&pCore->isr_tasklet, arcsas_interrupt_service_bh, (unsigned long)pCore);
	}
	osm_printk("arcsas%d: legacy INT enabled\n", pACB->host->host_no);
	return SUCCESS;
//#endif
}

static void arcsas_free_irq(struct pci_dev *pdev, PACB pACB)
{
	struct msix_data *msix_data;
	int i;
#if ARCSAS_DBG_INIT
	osm_printk("%s:\n", __func__);
#endif

#if 0
//#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0)
	if (pACB->acb_flag & ACB_FLAG_MSIX_ENABLED) {
		for (i = 0; i < pACB->vector_count; i++) {
			msix_data = &pACB->msix_data[i];
			if (msix_data->valid)
				free_irq(pACB->entries[i].vector, msix_data);
		}
		kfree(pACB->msix_data);
		arcsas_disable_msix(pACB);
		arcsas_disable_msix_SM(pACB);
	} else {
		for (i = 0; i < pACB->vector_count; i++)
			free_irq(pACB->entries[i].vector, pACB);
		if (pACB->acb_flag & ACB_FLAG_MSI_ENABLED)
			arcsas_disable_msi(pACB);
	}
	pci_free_irq_vectors(pdev);
	pACB->acb_flag &= ~(ACB_FLAG_MSIX_ENABLED | ACB_FLAG_MSI_ENABLED);
//#else
#endif
	if (pACB->acb_flag & ACB_FLAG_MSI_ENABLED) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
		free_irq(pci_irq_vector(pdev, 0), pACB);
		pci_free_irq_vectors(pdev);
#else
		free_irq(pdev->irq, pACB);
		pci_disable_msi(pdev);
#endif
		arcsas_disable_msi(pACB);
		pACB->acb_flag &= ~ACB_FLAG_MSI_ENABLED;
		for (i = 0; i < pACB->core_num; i++)
			tasklet_kill(&pACB->pCore[i]->isr_tasklet);
	} else if (pACB->acb_flag & ACB_FLAG_MSIX_ENABLED) {
		for (i = 0; i < pACB->vector_count; i++) {
			msix_data = &pACB->msix_data[i];
			if (msix_data->valid) {
				tasklet_kill(&pACB->isr_tasklet[i]);
				free_irq(pACB->entries[i].vector, msix_data);
			}
		}
		pci_disable_msix(pdev);
		kfree(pACB->msix_data);
		arcsas_disable_msix(pACB);
		arcsas_disable_msix_SM(pACB);
		pACB->acb_flag &= ~ACB_FLAG_MSIX_ENABLED;
	} else {
		free_irq(pdev->irq, pACB);
		for (i = 0; i < pACB->core_num; i++)
			tasklet_kill(&pACB->pCore[i]->isr_tasklet);
	}
//#endif
}

static int arcsas_probe(struct pci_dev *pcidev, const struct pci_device_id *id)
{
	struct Scsi_Host *host;
	PACB pACB;
	PCore_Extension pCore = 0;
	unsigned int retval = -ENXIO;
	unsigned int core_index;
	int i;
	u32 cnt = 0;
	struct dv_event_work *dv_event;

#if ARCSAS_DBG_INIT
	osm_printk("%s:\n", __func__);
#endif
	retval = pci_enable_device(pcidev);
	if (retval) {
		osm_printk("pci_enable_device returned fail........\n");
		goto out;
	}
	if (aspm == 0) {
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
			pci_disable_link_state(pcidev, PCIE_LINK_STATE_L0S | PCIE_LINK_STATE_L1 | PCIE_LINK_STATE_CLKPM);
		#else
			arcsas_disable_aspm(pcidev, PCIE_LINK_STATE_L0S | PCIE_LINK_STATE_L1 | PCIE_LINK_STATE_CLKPM);
		#endif
	}
#if	!ARCSAS_DBG_MARVELL
	if ((pcidev->subsystem_vendor != PCI_VENDOR_ID_ARECA)) {
		osm_printk("It's not the HBA from Areca. Contact the vendor.\n");
		retval = -ENODEV;
		goto out_disable_device;
	}
#endif
	pci_set_master(pcidev);
	retval = pci_request_regions(pcidev, "arcsas");
	if (retval) {
		osm_printk("pci_request_regions failed\n");
		goto out_disable_device;
	}
	if (arcsas_set_dma_mask(pcidev))
		goto out_pci_release;
	host = scsi_host_alloc(&arcsas_scsi_host_template, sizeof(struct AdapterControlBlock));
	if (!host) {
		retval = -ENOMEM;
		osm_printk("scsi_host_alloc failed\n");
		goto out_pci_release;
	}
	pACB = (struct AdapterControlBlock *)host->hostdata;
	pACB->vendor = pcidev->vendor;
	pACB->system_device_id = pcidev->device;
	pACB->subsystem_vendor = pcidev->subsystem_vendor;
	pACB->subsystem_device = pcidev->subsystem_device;
	pACB->adapter_type = id->driver_data;
	osm_pci_read_config_byte(pcidev, PCI_REVISION_ID, &pACB->Revision_Id);
#if	ARCSAS_DBG_INIT
	osm_printk("%s: Vendor ID = 0x%x, Device ID = 0x%x, SubVenID = 0x%x, SubDevID = 0x%x, RevisionID = 0x%x\n",
	    __func__, pACB->vendor, pACB->system_device_id, pACB->subsystem_vendor, pACB->subsystem_device, pACB->Revision_Id);
#endif
#if	ARCSAS_DBG_MARVELL
	if ((pACB->vendor == PCI_VENDOR_ID_MARVELL_EXT) && ((pACB->system_device_id == DEVICE_ID_1495) ||
	    (pACB->system_device_id == DEVICE_ID_1485))) {
		pACB->vendor = PCI_VENDOR_ID_ARECA;
		pACB->system_device_id = DEVICE_ID_1330;
		pACB->subsystem_vendor = PCI_VENDOR_ID_ARECA;
		pACB->subsystem_device = DEVICE_ID_1330;
	}
#endif
	max_lun = (max_lun <= ARCSAS_LUN_SUPPORTED) ? max_lun : ARCSAS_LUN_SUPPORTED;
	pACB->host = host;
	pACB->pcidev = pcidev;
	/*the parameters for "struct Scsi_Host" can't be omitted, or something troublesome will happen*/
	host->max_id =  ARCSAS_VIRTUAL_DEVICE_ID + 1;
	host->max_sectors = ARCSAS_MAX_CMDXFER_LENGTH / ARCSAS_SECTOR_SIZE;
	host->max_cmd_len = ARCSAS_CMD_LEN;    /*this is issue of 64bit LBA, over 2T byte*/
	host->sg_tablesize = ARCSAS_OS_SG_ENTRIES;
	if ((host_can_queue < ARCSAS_MIN_OUTSTANDING_CMD) || (host_can_queue > ARCSAS_OUTSTANDING_CMD))
		host_can_queue = ARCSAS_OUTSTANDING_CMD;
	host->can_queue = host_can_queue;
	if ((cmd_per_lun < ARCSAS_MIN_CMD_PERDEV) || (cmd_per_lun > ARCSAS_MAX_CMD_PERDEV))
		cmd_per_lun = ARCSAS_DEFAULT_CMD_PERDEV;
	host->cmd_per_lun = cmd_per_lun;
	host->this_id = -1;
	host->max_lun = max_lun;
	pci_set_drvdata(pcidev, host);
	retval = arcsas_initialize_hba(pACB, pcidev);
	if (!retval) {
		retval = -ENOMEM;
		osm_printk("arcsas_initialize_hba failed\n");
		goto out_host_put;
	}
	arcsas_init_virtual_ses_device(pACB);
	arcsas_spin_lock_init(&pACB->doneq_lock);
	arcsas_spin_lock_init(&pACB->discoveryQ_lock);
	arcsas_spin_lock_init(&pACB->host_lock);
	arcsas_spin_lock_init(&pACB->event_element_lock);
	arcsas_spin_lock_init(&pACB->dv_event_lock);
	arcsas_spin_lock_init(&pACB->Device_Tag_Pool.lock);
	arcsas_spin_lock_init(&pACB->Expander_Tag_Pool.lock);
	arcsas_spin_lock_init(&pACB->PM_Tag_Pool.lock);
	arcsas_spin_lock_init(&pACB->ehandler_lock);
	OSM_INIT_LIST_HEAD(&pACB->timeout_list);
	OSM_INIT_LIST_HEAD(&pACB->dv_event_list);
	dv_event = (struct dv_event_work *)kzalloc((sizeof(struct dv_event_work) * DV_EVENT_NUM), GFP_ATOMIC);
	pACB->dv_event = dv_event;
	arcsas_spin_lock(&pACB->dv_event_lock);
	for (i = 0; i < DV_EVENT_NUM; i++) {
		OSM_INIT_LIST_HEAD(&dv_event->list);
		osm_list_add_tail(&dv_event->list, &pACB->dv_event_list);
		dv_event++;
	}
	arcsas_spin_unlock(&pACB->dv_event_lock);
	pACB->acb_flag &= ~ACB_FLAG_ADAPTER_REMOVED;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
	pACB->scsi_lookup = alloc_workqueue("arcsas_%d", (WQ_UNBOUND | WQ_MEM_RECLAIM), 1, host->host_no);
#else
	pACB->scsi_lookup = create_singlethread_workqueue("arcsas");
#endif
	if (!pACB->scsi_lookup) {
		osm_printk("create_singlethread_workqueue failed\n");
		goto out_release_regions;
	}
	arcsas_disable_main_int(pACB);
	retval = arcsas_request_irq(pcidev, pACB);
	if (retval == ARC_FAILURE) {
		osm_printk("request_irq failed\n");
		goto out_kthread_stop;
	}
	arcsas_clear_all_irq_status(pACB);
	host->irq = pcidev->irq;
	retval = scsi_add_host(host, &pcidev->dev);
	if (retval) {
		osm_printk("scsi_add_host failed\n");
		goto out_free_irq;
	}
	for (i = 0; i < ARCSAS_MAX_SCAN_DELAY; i++)
		osm_msleep(1000);
	for (core_index = 0; core_index < pACB->core_num; core_index++) {
		pCore = pACB->pCore[core_index];
		pCore->State |= CORE_STATE_IDLE;
		arcsas_port_state_machine(pCore, NULL);
		do {
			osm_msleep(25);
			arcsas_interrupt_check(pACB);
			if (++cnt >= 800)
				break;
		} while (!(pCore->State & CORE_STATE_STARTED));
	}
#if ARCSAS_DBG_REGS
	if ((cnt >= 800) && (pACB->acb_flag & ACB_FLAG_MSIX_ENABLED))
		arcsas_dump_reg_1330(pCore);
#endif
	for (i = 0; i < ARCSAS_MAX_DEVICE_ID; i++) {
		PDomain_Device pDevice;
		if (pACB->DevicesExpandersPMs_EntryIndex[i] == ID_NOT_MAPPED) {
			continue;
		}
		pDevice = &pACB->Devices[Get_DevicesExpandersPMs_EntryIndex(pACB, i)];/*(0..135)*/
		if ((pDevice->State == DEVICE_STATE_INIT_DONE) || (pDevice->State == DEVICE_STATE_INIT_FAILED)) {
			continue;
		} else {
			//BUG();
			break;
		}
	}
//	arcsas_check_update_flash_device_map(pACB, true);
	arcsas_enable_interrupt(pACB);
	scsi_scan_host(host);
	return 0;
out_free_irq:
	arcsas_disable_interrupt(pACB);
	arcsas_free_irq(pcidev, pACB);
out_kthread_stop:
	kthread_stop(pACB->eh_handler);
	destroy_workqueue(pACB->scsi_lookup);
out_release_regions:
	kfree(pACB->dv_event);
	arcsas_iounmap(pACB);
	arcsas_teardown_dma_pool(pACB);
	osm_vfree(pACB->DevicesLun);
out_host_put:
	pci_set_drvdata(pcidev, NULL);
	scsi_host_put(host);
out_pci_release:
	pci_release_regions(pcidev);
out_disable_device:
	pci_disable_device(pcidev);
out:
	return retval;
}
/*
************************************************************************
************************************************************************
*/
static struct scsi_host_template arcsas_scsi_host_template = {
	.module			= THIS_MODULE,
	.proc_name		= "arcsas",
	.name			= "arcsas",
	.info			= arcsas_info,
	.queuecommand		= arcsas_queuecommand,
	.bios_param		= arcsas_bios_param,
	.eh_abort_handler	= arcsas_eh_abort_handler,
	.eh_device_reset_handler= arcsas_eh_device_reset_handler,
	.can_queue		= ARCSAS_OUTSTANDING_CMD -1,/*mvsas in kernel source*/
	.this_id		= -1, /* or ARCSAS_SCSI_INITIATOR_ID = 255*/
	.sg_tablesize		= ARCSAS_OS_SG_ENTRIES,
	.max_sectors		= ARCSAS_MAX_CMDXFER_LENGTH / ARCSAS_SECTOR_SIZE,
	.cmd_per_lun		= ARCSAS_DEFAULT_CMD_PERDEV,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11)
	.change_queue_depth	= arcsas_change_queue_depth,
#else
	.sdev_attrs		= arcsas_scsi_device_attr,
#endif
#if LINUX_VERSION_CODE <= KERNEL_VERSION(4, 20, 0)
	.use_clustering		= ENABLE_CLUSTERING,
#endif
};
/*
************************************************************************
************************************************************************
*/
static struct pci_device_id arcsas_pci_id_table[] = {
	{PCI_DEVICE(PCI_VENDOR_ID_ARECA, DEVICE_ID_1300),
		.driver_data = ARC_1300},
	{PCI_DEVICE(PCI_VENDOR_ID_ARECA, DEVICE_ID_1320),
		.driver_data = ARC_1320},
	{PCI_DEVICE(PCI_VENDOR_ID_ARECA, DEVICE_ID_1330),
		.driver_data = ARC_1330},
#if	ARCSAS_DBG_MARVELL
	{PCI_DEVICE(PCI_VENDOR_ID_MARVELL_EXT, DEVICE_ID_1495),
		.driver_data = ARC_1330},
	{PCI_DEVICE(PCI_VENDOR_ID_MARVELL_EXT, DEVICE_ID_1485),
		.driver_data = ARC_1330},
#endif
	{} /* Terminating entry */
};
MODULE_DEVICE_TABLE(pci, arcsas_pci_id_table);
/*
************************************************************************
************************************************************************
*/
static struct pci_driver arcsas_pci_driver = {
	.name		= "arcsas",
	.id_table	= arcsas_pci_id_table,
	.probe		= arcsas_probe,
	.remove		= arcsas_remove,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 13)
	.shutdown	= arcsas_shutdown, 
#endif
#ifdef CONFIG_PM
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
	.suspend	= arcsas_suspend,
	.resume		= arcsas_resume,
	#endif
#endif
};
/*
************************************************************************
************************************************************************
*/
static int arcsas_module_init(void)
{
	return pci_register_driver(&arcsas_pci_driver);
}
/*
************************************************************************
************************************************************************
*/
static void  arcsas_module_exit(void)
{
	pci_unregister_driver(&arcsas_pci_driver);
}
module_init(arcsas_module_init);
module_exit(arcsas_module_exit);

